#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#include "debug.h"
#include "mtadel.h"

extern int r_answer;
extern struct symbol *symbol_table_start;
extern struct symbol *next_iterator(struct symbol *);
extern char *replace(char *, char *, char *);
extern char *mtadel_itoa(int n);
extern int resolve_lower_bound(char *);
extern int resolve_upper_bound(char *);
extern int mtadel_eval(char *);

int iterations;

void recursive_iterate(char *, struct symbol *);

/* ============================
evaluate(s)

This function is responsible for actually
	generating the numeric answer for
	the nested summations represented in 
	the linked list whose head is 
	pointed to by _s.

	evaluate() actually calls recursive_
	iterate() to do most of the work. That 
	is, unless the summation to be evaluated
	is already solved. This can be decided
	by checking to see if the sum_is_sym 
	flag of the _s structure is not 
	set.
   ==========================*/
int evaluate(struct summation *s)
{
	if (s->sum_is_sym)
	{
		iterations = 0;
		recursive_iterate(s->sum_sym, NULL);
	}
	else
	{
		r_answer = s->sum;
	}
	return 0;
}

/* ============================
recursive_iterate(eq, symbol)

This function modifies a symbolic equation for
	each of variables listed as 'iterate' from
	within the symbol table. 

	Specifically, the algorithm is as follows:
	1. Find the first symbol whose iterate flag is set.
	2. Ask the symbol table for more information about
		the symbol. This includes updating upper
		and lower bounds on the variable's value 
		based upon information and relationships
		garnered during the parsing phase.
	3. Iterate through each of the possible values 
		for the current variable.
	4. At each iteration, set the current field of
		the variable's structure to the value
		of the current iteration.
	5. Replace the variable representation of the 
		current symbol within the _eq with the
		numeric value given by the value 
		of the current iteration.
	5. Recursively invoke recursive_iterate(). This
		will cause this process to repeat for the next
		variable marked 'iterate' within the symbol
		table.
	6. The process will end when there are no further
		symbols marked as iterate within the symbol table.
   ==========================*/
void recursive_iterate(char *eq, struct symbol *symbol)
{
	/* on the first iteration, symbol == NULL. This is our indication to the point to the
		start of the symbol table to begin the recursion. */
	if (symbol == NULL)
		symbol = symbol_table_start;
	
	if (symbol != (struct symbol*)0xf0000000 && more_iterators(symbol)) 
		/* should return true if _symbol, or subsequent symbol table entries have iterate==1*/
	{
		struct symbol *n = next_iterator(symbol); /* returns either symbol, or a subsequent 
							symbol table entry that has iterate == 1. By
							calling more_iterators() before this, we are
							garaunteed that next_symbol() will not return NULL */
		int lo = 0, hi = 0;
		lo = n->lo;
		hi = n->hi;
		if (n->hybrid == 1)
		{
			EVAL_VERBOSE("we are updating our lower bound dynamically for %s\n", n->name);
			lo = resolve_lower_bound(n->name);
			EVAL_VERBOSE("lo: %d\n", lo);
			if (lo < 0)
				lo = 0;

			EVAL_VERBOSE("we are updating our upper bound dynamically for %s\n", n->name);
			hi = resolve_upper_bound(n->name);
			EVAL_VERBOSE("hi: %d\n", hi);
			if (hi < 0)
				hi = 0;

		}

		/* this is for shifting our limits of summation based on lower shifting! */
		EVAL_VERBOSE("n->is_lo_shifted: %d\n", n->is_lo_shifted);
		if (n->is_lo_shifted == 1)
		{	
			char *t_lo = mtadel_itoa(lo);
			int new_lo_len = strlen(t_lo) + strlen(n->lo_shift) + 1;

			char *new_lo = (char *)malloc(new_lo_len);
			memset(new_lo, 0, new_lo_len);

			strcpy(new_lo, t_lo);
			strcat(new_lo, n->lo_shift);
			EVAL_VERBOSE("replacing lo with an eval(%s) answer!\n", new_lo);
			EVAL_VERBOSE("lo: %d\n", lo);
			lo = mtadel_eval(new_lo);
			EVAL_VERBOSE("new lo: %d\n", lo);
		}

		/* this is for shifting our limits of summation based on upper shifting! */
		EVAL_VERBOSE("n->is_hi_shifted: %d\n", n->is_hi_shifted);
		if (n->is_hi_shifted == 1)
		{
			char *t_hi = mtadel_itoa(hi);
			int new_hi_len = strlen(t_hi) + strlen(n->hi_shift) + 1;

			char *new_hi = (char *)malloc(new_hi_len);
			memset(new_hi, 0, new_hi_len);

			strcpy(new_hi, t_hi);
			strcat(new_hi, n->hi_shift);
			EVAL_VERBOSE("replacing hi with an eval(%s) answer!\n", new_hi);
			hi = mtadel_eval(new_hi);
			EVAL_VERBOSE("new hi: %d\n", hi);
		}
		/* after having established upper and lower bounds, iterate 
			through each of those values. */
		for (lo; lo<=hi; lo++)
		{
			char *new_eq = NULL;

			iterations++;
			EVAL_VERBOSE("n->name: -%s-\n", n->name);
			EVAL_VERBOSE("eq: -%s-\n", eq);
			/* replace the variable with its current value within the
				equation to be evaluated. By calling this function
				recursively, the _eq is never permanently changed
				and therefore we can replace the variables with
				each of its possible values. */
			new_eq = replace(eq, n->name, mtadel_itoa(lo));
			EVAL_VERBOSE("new_eq : %s\n", new_eq);
		
			/* make the current value of lo the current value
				of the symbol within the symbol table. This is
				important in order for subsequent calls to recursive_iterate()
				to see the current value of the symbol. This "current" value 
				is used by resolve_(upper|lower)_bound(). */
			update_current(n->name, lo);
			if (n->next == NULL)
				recursive_iterate(new_eq, (struct symbol*)0xf0000000);
			else
				recursive_iterate(new_eq, n->next);
		}
	}
	else
	{
		/* we end up here if there are not further variables
			in the symbol table marked as "iterate". */

		/* eval_ret is the global variable used by the evaluation code (found in eval/). 
			Since it is global, it is safer to reset it here and be gart'd to have 
			correct answers. */
		int eval_ret = 0;
		EVAL_VERBOSE("this is the base case!\n");
		EVAL_VERBOSE("evaluating: %s\n", eq);

		eval_ret = mtadel_eval(eq);
		if (eval_ret >= 0)
			r_answer += eval_ret;
	}
	return;
}
