#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../debug.h"
#include "../mtadel.h"

extern int is_numeric( char* );

struct symbol *symbol_table_start;

/* ============================
init_symbol_table()

This function should be called before the
	symbol table is first used (ie. any 
	of the other functions in this file
	are called. 

   ==========================*/
void init_symbol_table()
{
	symbol_table_start = NULL;
}


/* ============================
ptr_to_last_symbol()

Return a pointer to the last symbol table
	entry.
   ==========================*/
struct symbol *ptr_to_last_symbol()
{
	struct symbol *iterator = symbol_table_start;
	for (; iterator->next != NULL; iterator = iterator->next);
	return iterator;
}

/* ============================
add_symbol(name, lo, hi, hybrid)

Create and add a new summation structure to the 
	symbol table. Initialize the new structure's
	fields to lo, hi and hybrid approrpiately.
   ==========================*/
void add_symbol(char *name, int lo, int hi, int hybrid)
{
	/* we need to build the table backwards so that items that are added later are closer to the front
		of the list */
	SYMBOL_VERBOSE("adding a symbol named: %s\n", name);
	SYMBOL_VERBOSE("add_symbol says hybrid is %d\n", hybrid);
	if (symbol_table_start == NULL)
	{
		/* adding the first node! */
		symbol_table_start = (struct symbol *)malloc(sizeof(struct symbol));
		memset(symbol_table_start, 0, sizeof(struct symbol));
		symbol_table_start->name = name;
		symbol_table_start->lo = lo;
		symbol_table_start->hi = hi;
		symbol_table_start->hybrid = hybrid;
	}
	else
	{
		struct symbol *new_symbol = (struct symbol*)malloc(sizeof(struct symbol));
		memset(new_symbol, 0, sizeof(struct symbol));

		new_symbol->next = symbol_table_start;
		symbol_table_start = new_symbol;

		//(ptr_to_last_symbol())->next = new_symbol;
		new_symbol->name = name;
		new_symbol->lo = lo;
		new_symbol->hi = hi;
		new_symbol->hybrid = hybrid;
	}
}

/* ============================
get_symbol(name)

Returns a pointer to the symbol structure with name _name. 
	Returns NULL if none were found that match.
   ==========================*/
struct symbol *get_symbol(char *name)
{
	struct symbol *iterator = symbol_table_start;
	for (; iterator != NULL; iterator = iterator->next)
	{
		if (!strcmp(iterator->name, name))
		{
			/* we have a match */
			return iterator;
		}
	}
	return NULL;
}

/* ============================
get_symbol_by_halfs(first, second)

Returns a pointer to the symbol whose first and second
	halves match _first and _second, respectively.
   ==========================*/
struct symbol *get_symbol_by_halfs(char *first, char *second)
{
	struct symbol *iterator = symbol_table_start;
	for (; iterator != NULL; iterator = iterator->next)
	{
		if (iterator->hybrid == 1 && !strcmp(iterator->first, first) && !strcmp(iterator->second, second))
		{
			return iterator;
		}
	}
	return NULL;
}

/* ============================
print_symbol_table()

Print (to the screen) all the fields of
	each of the symbol structures in the
	symbol table (represented by a linked list).
   ==========================*/
void print_symbol_table()
{
	struct symbol *iterator = symbol_table_start;
	ERROR("================================\n");
	for (; iterator != NULL; iterator = iterator->next)
	{
		ERROR("=============\n");
		ERROR("name: %s\n", iterator->name);
		ERROR("lo: %d\n", iterator->lo);
		ERROR("hi: %d\n", iterator->hi);
		ERROR("first: %s\n", iterator->first);
		ERROR("second: %s\n", iterator->second);
		ERROR("upper: %s\n", iterator->upper);
		ERROR("hybrid: %d\n", iterator->hybrid);
		ERROR("iterate: %d\n", iterator->iterate);
		ERROR("current: %d\n", iterator->current);
		ERROR("lo_shift: %s\n", iterator->lo_shift);
		ERROR("hi_shift: %s\n", iterator->hi_shift);
		ERROR("=============\n");
	}
	ERROR("================================\n");
}
/* ============================
normalize_individual()

This function is called by normalize(). This function
	uses _to_normalize->first and _to_normalize->second
	to determine the upper and lower bounds for the
	symbol represented by _to_normalize. Only a symbol
	whose hybrid flag is set to 1 can be normalized. 

	Algorithm:
		_to_normalize->lo = max(_to_normalize->first->lo, 
					_to_normalize->second->lo).
					*Note that _to_normalize->first->lo is not
					technically possible, but semantically
					what actually happens is equivalent.
		_to_normalize->hi = max(_to_normalize->first->lo,
					_to_normalize->second->lo).
   ==========================*/
void normalize_individual(struct symbol *to_normalize)
{
	/* we can assume it is a hybrid. */

	int a_lo, a_hi, b_lo, b_hi;

	struct symbol *t = get_symbol(to_normalize->first);
	if (t==NULL)
	{
		SYMBOL_VERBOSE("get_symbol(1) returned NULL in normalize_individual\n");
		return;
	}

	a_lo = t->lo;
	a_hi = t->hi;

	t = get_symbol(to_normalize->second);
	if (t == NULL)
	{
		SYMBOL_VERBOSE("get_symbol(2) returned NULL in normalize_individual\n");
		return;
	}

	b_lo = t->lo;
	b_hi = t->hi;

	SYMBOL_VERBOSE("a_lo: %d\n", a_lo);
	SYMBOL_VERBOSE("a_hi: %d\n", a_hi);
	SYMBOL_VERBOSE("b_lo: %d\n", b_lo);
	SYMBOL_VERBOSE("b_hi: %d\n", b_hi);

	/* set lo and hi according to the algorithm preceeding this function. */
	to_normalize->lo = (a_lo > b_lo) ? a_lo : b_lo;
	to_normalize->hi = (a_hi <= b_hi) ? b_hi : a_hi;
	return;
}

/* ============================
update_halves(name, first, second)

This function updates the first and second
	fields of _name by setting them to _first
	and _second, respectively.
   ==========================*/
void update_halves(char *name, char *first, char *second)
{
	struct symbol *t = get_symbol(name);
	if (t == NULL)
		return;
	
	t->first = first;
	t->second = second;
	return;
}

/* ============================
normalize()

This function will normalize each of the 
	symbols in the symbol table whose 
	hybrid flag is set to 1. This is 
	accomplished by looping through all
	symbols and calling normalize_individual()
	on any symbol whose hybrid flag is
	set to 1.

	Note: normalize means update the symbol's
	upper and lower bounds based up the
	variables stored within _first and _second. 
	_first and _second represent bindings 
	between different variables within
	the nested summation being solved. 
   ==========================*/
void normalize()
{
	struct symbol *iterator = symbol_table_start;
	SYMBOL_VERBOSE("start normalizing\n");
	for (; iterator != NULL; iterator = iterator->next)
	{
		/* if the induction symbol has a real upper bound, then use it here. */
		if (iterator->hybrid == 1)
		{
			SYMBOL_VERBOSE("we are normalizing ...\n");
			/* we have something to normalize! */
			normalize_individual(iterator);
		}
	}
	SYMBOL_VERBOSE("end normalizing\n");
	return;
}

/* ============================
mark_iterator_variable(name)

Set the iterate flag to 1 of the symbol
	structure corresponding to _name.
   ==========================*/
void mark_iterator_variable(char *name)
{
	struct symbol *s = get_symbol(name);
	if (!s)
		return;
	s->iterate = 1;
}

/* ============================
more_iterators(s)

Given a pointer into the linked list of symbols, s,
	this function determines if subsequent symbols
	have their iterate flag set to 1. If at least 
	1 does, then return 1. Otherwise, return 0. This
	is used in evaluate().
   ==========================*/
int more_iterators(struct symbol *s)
{
	SYMBOL_VERBOSE("start more_iterators()\n");
	if (s == NULL)
		SYMBOL_VERBOSE("in more_iterators the parameter == NULL");

	for (; s != NULL; s = s->next)
	{
		SYMBOL_VERBOSE("%s->iterate: %d\n", s->name, s->iterate);
		if (s->iterate == 1)
			return 1;
	}
	SYMBOL_VERBOSE("end more_iterators()\n");
	return 0;
}

/* ============================
next_iterator(s)

This function returns a pointer to the next
	symbol structure in the linked list
	whose iterate flag is set to 1. This
	function may return s if it is an 
	iterate symbol. If none are found, 
	NULL is returned. However, this should
	not happen because next_iterator() is 
	called from within evaluate.c only after
	more_iterators() tells the loop that 
	another iterate symbol exists. 
   ==========================*/
struct symbol *next_iterator(struct symbol *s)
{
	SYMBOL_VERBOSE("start next_iterator()\n");
	if (s == NULL)
		SYMBOL_VERBOSE("in more_iterators the parameter == NULL");

	for (; s != NULL; s = s->next)
	{
		SYMBOL_VERBOSE("%s->iterate: %d\n", s->name, s->iterate);
		if (s->iterate == 1)
			return s;
	}
	SYMBOL_VERBOSE("end next_iterator()\n");
	return NULL;
}

/* ============================
update_current(name, current)

This function sets the current field of the
	symbol structure associated with _name
	to _current. This is used within the
	evaluate() function (evaluate.c).
   ==========================*/
void update_current ( char *name, int current)
{
	struct symbol *s = get_symbol(name);
	if (s == NULL) return;
	s->current = current;
	SYMBOL_VERBOSE("%s->current: %d\n", s->name, s->current);
}
/* ============================
resolve_lower_bound(name)

This function returns the current
	lower bound of the symbol associated
	with _name. This is done by trying to
	find the current value of the symbol
	associated with _name->second. (This works because
	the second field of symbol is equal to "i" where
	j=i represents the lower 1/2 of a summation.) 

	If there is no symbol found in the table 
	by the name of _name->second, then search
	through all of the hybrid symbols (symbols 
	whose hybrid flag is set to 1) until a
	hybrid symbol i has _i->first == _name->second.
	In this case, return _i->lo.
   ==========================*/
int resolve_lower_bound(char *name)
{
	struct symbol *s, *n;
	s = get_symbol(name);
	if (s == NULL || s->hybrid != 1 ) return -1;

	if (is_numeric(s->second)) return atoi(s->second);

	SYMBOL_VERBOSE("s->second: %s\n", s->second);
	n = get_symbol(s->second);
	if (n == NULL || n->iterate != 1) 
	{
		/* let's try to search all the hybrids for a "first" that matches. */
		struct symbol *iterator = symbol_table_start;
		for (; iterator != NULL; iterator = iterator->next)
		{
			SYMBOL_VERBOSE("we are here!\n");
			if (iterator->hybrid != 1)
				continue;
			if (!strcmp(iterator->first, s->second))
			{
				SYMBOL_VERBOSE("CRAP!\n");
				return iterator->current;
			}
		}
		/* if we reach here, we really can fail */
		return -1;
	}
	else
	{
		SYMBOL_VERBOSE("resolve_lower_bound is returning %d\n", n->lo);
		return n->current;
	}
}

/* ============================
dependents(check)


This function will return true if a summation more deeply
	nested within the summation to be solved depends 
	on _check to be calculated.

	Algorithm: Loop through all the hybrid variables 
		and check to see if their 'second' matches the parameter.
   ==========================*/
int dependents(char *check)
{

	struct symbol *iterator = symbol_table_start;
	for (; iterator!=NULL; iterator = iterator->next)
	{
		if (iterator->hybrid != 1)
			continue;
		/* NOTE: we do not have to check iterator->iterate == 1 here because
			when this function is called within calculate() (../calculate.c)
			not all of the symbols who will end up with iterate==1 are actually
			set that way. If all the symbols who will eventually receive iterate 
			status were set at this point, then yes, a check of iterate==1 would
			be necessary. */
		SYMBOL_VERBOSE("iterator->upper: %s\n", iterator->upper);
		SYMBOL_VERBOSE("check: %s\n", check);
		if (!strcmp(iterator->second, check) || (iterator->upper && !strcmp(iterator->upper, check)))
		{
			return 1;
		}
	}
	return 0;
}

/* ============================
update_upper(name, hi)

This function sets the hi field of the symbol structure
	associated with _name to _hi.
   ==========================*/
void update_upper(char *name, char *hi)
{
	struct symbol *n = get_symbol(name);
	if (n == NULL)
		return;

	n->upper = hi;
}

/* ============================
resolve_upper_bound(name)

This function will return the current value of the 
	upper value associated with the _name symbol.
	The algorithm for calculating this upper bound
	is identical to the algorithm used to calculate
	the lower bound (found in resolve_lower_bound()).
   ==========================*/
int resolve_upper_bound(char *name)
{
	struct symbol *s, *n;
	s = get_symbol(name);
	if (s == NULL || s->hybrid != 1 ) return -1;
	
	SYMBOL_VERBOSE("s->second: %s\n", s->second);
	if (s->upper == NULL)
		return s->hi;

	n = get_symbol(s->upper);
	if (n == NULL || n->iterate != 1) 
	{
		/* let's try to search all the hybrids for a "first" that matches. */
		struct symbol *iterator = symbol_table_start;
		for (; iterator != NULL; iterator = iterator->next)
		{
			SYMBOL_VERBOSE("we are here!\n");
			if (iterator->hybrid != 1)
				continue;
			if (!strcmp(iterator->first, s->upper))
			{
				SYMBOL_VERBOSE("CRAP!\n");
				return iterator->current;
			}
		}
		/* if we reach here, we really can fail */
		return -1;
	}
	else
	{
		SYMBOL_VERBOSE("resolve_upper_bound is returning %d\n", n->lo);
		return n->current;
	}
}

/* ============================
update_hi_shift(name, hi_shift)

This function will set the hi_shift field of the
	symbol structure associated with _name
	to _hi_shift, given that _hi_shift is not
	NULL and that there actually exists a 
	symbol named _name).
   ==========================*/
void update_hi_shift(char *name, char *hi_shift)
{
	struct symbol *s = get_symbol(name);
	SYMBOL_VERBOSE("we are setting %s->hi_shift = %s\n", name, hi_shift);
	if (s == NULL || hi_shift == NULL)
		return;
	
	s->hi_shift = hi_shift;
	s->is_hi_shifted = 1;
}

/* ============================
update_lo_shift(name, lo_shift)

Identical to update_hi_shift, except it is 
	used for the lo_shift field.
   ==========================*/
void update_lo_shift(char *name, char *lo_shift)
{
	struct symbol *s = get_symbol(name);
	SYMBOL_VERBOSE("we are setting %s->lo_shift = %s\n", name, lo_shift);
	if (s == NULL || lo_shift == NULL)
		return;
	
	s->lo_shift = lo_shift;
	s->is_lo_shifted = 1;
}
