/*
 * CFG.c is used to convert assembly into path,inf files
 */

#include "CFG.h"

#include <cstdlib>
#include <string>

/*
 * error - print error and exit
 */
void error(char *s){
  fprintf(stderr, "%s\n", s);
  exit(-1);
}

/*
 * compress_whitespace - remove extra whitespace from the assembly 
 */
void compress_whitespace(char *s){
  /*
   * Whitespace at start
   */
  while (*s && isspace(*s))
    strcpy(s, s + 1);

  for (; *s; s++)
    if (isspace(*s)) {
      /*
       * in middle
       */
      while (*(s + 1) && isspace(*(s + 1)))
	strcpy(s, s + 1);
      /*
       * at end
       */
      if (*(s + 1))
	*s = ' ';
    }
}


/*
 * write_rtls - prinnt lines to output file (for test print assembly again)
 */
void write_rtls(FILE *out, struct line *rtl){
  for (; rtl; rtl = rtl->next)
    fprintf(out,"%s", rtl->str);
}

/*
 * write_block - bblock info and lines to stdout
 * in: 1st line of bblock
 */
void write_block(FILE *out, struct bblock *cblk) {
  struct blist *list;
  void print_blist(FILE *out, struct blist *list);
  struct blist *reverse_blist(struct blist *plist, struct blist *list);
  /*
   * bblock info: block #, pred, succ
   */
  fprintf(out,"\n!new block # %d, preds", cblk->nr);
  print_blist(out, cblk->preds);
  fprintf(out,", succs");
  print_blist(out, cblk->succs);
  fprintf(out,", doms");
  print_blist(out, cblk->doms);
  if (cblk->loop)
    fprintf(out,", loop");
  print_blist(out, cblk->loop);
  fprintf(out,"\n");
  fprintf(out,"!LABEL=%s, TARGET=%s, NUM_INSTS=%d, LOOPNO=%d\n",
	  cblk->label,cblk->target,cblk->num_inst,cblk->loopno);

  write_rtls(out, cblk->rtls);
}

/*
 * add_blist - add new block nblk to list, return head of new list
 */
struct blist *add_blist(struct blist *list, struct bblock *nblk){
  struct blist *root;

  root = (struct blist *) malloc(sizeof(struct blist));
  if (root == NULL)
    error("out of heap memeory for block lists");
  root->ptr = nblk;
  root->next = list;
  return(root);
}

/*
 * add_sorted_blist - insert sorted new nblk in list, return head of new list
 */
struct blist *add_sorted_blist(struct blist *root, struct bblock *nblk){
  struct blist *nlist, *plist, *clist;

  nlist = (struct blist *) malloc(sizeof(struct blist));
  if (nlist == NULL)
    error("out of heap memeory for block lists");
  nlist->ptr = nblk;

  for (plist = NULL, clist = root; clist; plist = clist, clist = clist->next)
    if (nlist->ptr->nr > clist->ptr->nr)
      break;

  if (plist) {
    plist->next = nlist;
    nlist->next = clist;
  }
  else {
    root = nlist;
    nlist->next = clist;
  }
  return(root);
}

/*
 * is_in_blist - check if cblk in list, return TRUE/FALSE
 */
int is_in_blist(struct blist *list, struct bblock *cblk){
  for (; list; list = list->next)
    if (list->ptr == cblk)
      return(TRUE);
  return(FALSE);
}

/*
 * is_in_sorted_blist - block cblk in sorted list
 * (higher block # first) ? return TRUE, sonst FALSE
 */
int is_in_sorted_blist(struct blist *list, struct bblock *cblk){
  for (; list && list->ptr->nr > cblk->nr; list = list->next);
  return(list && list->ptr == cblk);
}

/*
 * reverse_blist - reverse ordering of block list
 */
struct blist *reverse_blist(struct blist *plist, struct blist *list){
  struct blist *root;

  if (list) {
    root = reverse_blist(list, list->next);
    list->next = plist;
    return(root);
  }
  return(plist);
}

/*
 * print_blist - print block list to stdout
 */
void print_blist(FILE *out, struct blist *list){
  struct blist *newlist;
  void free_blist();

#ifdef TRASH
  newlist = NULL;
  for (; list; list = list->next)
,    newlist = add_sorted_blist(newlist, list->ptr);
#else
  newlist = list;
#endif

  for (list = newlist; list; list = list->next)
    fprintf(out," %d", list->ptr->nr);
  fflush(stdout);

  free_blist(newlist);
}

/*
 * free_blist - release heap space of block list
 */
void free_blist(struct blist *list){
  if (list) {
    free_blist(list->next);
    free(list);
  }
}

/*
 * find_target - search for target label of jump in cblk in block list
 * starting from tblk and register pred/succ w/ cblk
 */
void find_target(struct bblock *cblk, struct bblock *tblk, char *label){
  for (; tblk; tblk = tblk->next) {
    if (strcmp(tblk->label, label) == 0) {
      cblk->succs = add_blist(cblk->succs, tblk);
      tblk->preds = add_blist(tblk->preds, cblk);
    }}
}





/*
 * modified_doms - line (5) of algorithm 10.16 Dragon book for block cblk
 * and change set iff dominators of cblk midified
 */
struct blist *modified_doms(struct bblock *cblk, int *change,
			    struct bblock *top) {
  struct blist *start, *preds; /* predessors */
  struct blist *pdoms; /* dom list of 1st pred */
  struct blist *doms; /* new dom list */
  struct blist *odoms, *ndoms; /* old/new doms */

  doms = NULL;
  for (start = cblk->preds;
       start && !is_in_sorted_blist(start->ptr->doms, top);
       start = start->next); /* skip unreachable code */
  if (start)
    for (pdoms = start->ptr->doms; pdoms; pdoms = pdoms->next) {
      for (preds = start->next; preds; preds = preds->next)
	if (is_in_sorted_blist(preds->ptr->doms, top) && /* skip unreachable */
	    !is_in_sorted_blist(preds->ptr->doms, pdoms->ptr))
	  break; /* not in D(preds->ptr) */
      if (!preds)
	doms = add_blist(doms, pdoms->ptr); /* in intersection, add to D(n) */
    }

  /*
   * Assume: add_blist has dominators in ascending block #,
   * let's reverse it now
   */
  doms = reverse_blist(NULL, doms);

  if (!is_in_sorted_blist(doms, cblk))
    doms = add_sorted_blist(doms, cblk); /* D(n) := {n} */

  /*
   * Compare new doms w/ old ones
   * Except: both sorted in descending oder of block #
   */
  for (odoms = cblk->doms, ndoms = doms; odoms && ndoms;
       odoms = odoms->next, ndoms = ndoms->next)
    if (odoms->ptr != ndoms->ptr)
      break;
  if (odoms != ndoms) {
    *change = TRUE;
    free_blist(cblk->doms);
    return(doms);
  }
  free_blist(doms);
  return(cblk->doms);
}

/*
 * find_doms - finds doms in func w/ 1st block top
 * Algorithm 10.16 Dragon book
 */
void find_doms(struct bblock *top){
  struct bblock *cblk, *dblk;
  int change;

  top->doms = add_blist(top->doms, top); /* line (1) */
  for (cblk = top->next; cblk; cblk = cblk->next) /* line (2) */
    for (dblk = top; dblk; dblk = dblk->next)
      cblk->doms = add_blist(cblk->doms, dblk);

  /*
   * Assume: add_blist has doms in desc. oder
   */

  do {
    change = FALSE;
    for (cblk = top->next; cblk; cblk = cblk->next) /* line (4) */
      cblk->doms = modified_doms(cblk, &change, top); /* line (5) */
  } while (change); /* line (3) */
}

/*
 * dominates - return TRUE if cblk dominates dblk
 */
int dominates(struct bblock *dblk, struct bblock *cblk){
  return(is_in_sorted_blist(cblk->doms, dblk));
}

/*
 * insert - cf. algorithm 10.1 Dragon book
 */
void insert(struct blist **loop, struct bblock *cblk, struct blist **stack){
  if (!is_in_blist(*loop, cblk)) {
    *loop = add_blist(*loop, cblk);
    *stack = add_blist(*stack, cblk);
  }
}

/*
 * ReachUnder - cblk --> dblk: Find all blocks between head dblk and
 * source of backedge cblk in CFG
 * and add these blocks (incl. cblk and dblk) to
 * new loop block list of cblk, cf. algorithm 10.1 Dragon book
 */
void ReachUnder(struct bblock *cblk, struct bblock *dblk){
  struct blist *preds;
  struct blist *stack, *old;
  struct blist *loop;
  
  stack = NULL;
  loop = dblk->loop;

  if (!is_in_blist(loop, dblk)) {
    loop = add_blist(loop, dblk);
  }
  insert(&loop, cblk, &stack);

  while (stack != NULL) {
    /*
     * pop top of stack
     */
    old = stack;
    preds = stack->ptr->preds;
    stack = stack->next;
    free(old);
    for (; preds; preds = preds->next){
      insert(&loop, preds->ptr, &stack);
    }
  }
  dblk->loop = loop;
}

/*subset -  check if one array is inside the other */
int subset(int *inner, int size1, int *outer, int size2) {
  int i, j, found;

  if (size1>size2) return FALSE;

  for (i=0;i<size1;i++) {
    found = FALSE;
    for(j=0;j<size2;j++) if (inner[i] == outer[j]) found = TRUE;
    if (!found) return FALSE;
  }
  return TRUE;
}


/*sort_list - sort the array.... */
void sort_list(int *list, int size) {
int i, j, k,temp;

  for (k=0; k<size; k++) {
    temp = list[k];
    j = k;
    for (i=k; i<size; i++) {
      if (temp > list[i]) {
	j = i;
	temp = list[i];
      }
    }
    list[j] = list[k];
    list[k] = temp;
  } 
}


/*
 * find_loops - Find natural loops in function w/ 1st block top
 */
void find_loops(struct bblock *top){
  struct bblock *cblk;
  struct blist *succs, *loop;
  int num_blocks, i, j, k, temp, *array, *temp_list;
  cblk = NULL;
  succs = loop = NULL;
  temp_list = array = NULL;

  for (cblk = top; cblk; cblk = cblk->next)
    for (succs = cblk->succs; succs; succs = succs->next)
      if (dominates(succs->ptr, cblk))
	ReachUnder(cblk, succs->ptr); /* backedge cblk --> succs->ptr */

  /*after finding all the blocks in the loop fill the loop_blocklist array*/
  for (cblk = top; cblk; cblk = cblk->next){
    num_blocks = 0;
    if (cblk->loop) {
      for (loop = cblk->loop;loop ; loop = loop->next) num_blocks++;
      cblk->loop_blocklist = (int *) malloc(num_blocks * sizeof(int));
      cblk->loop_bno = num_blocks;
      array = cblk->loop_blocklist;
      for (loop = cblk->loop, i = 0;loop ; loop = loop->next, i++) array[i] = loop->ptr->nr;

      /*now sort the array*/
      /*sort_list(array, num_blocks);*/

      /*reverse the array... in case thats the way to go */
      if (temp_list !=NULL) free(temp_list);
      temp_list = (int *) malloc(num_blocks * sizeof(int));
      
      for (i=0; i<num_blocks; i++) temp_list[i] = array[i];

      for (i=0, j=num_blocks-1; i <num_blocks; i++,j--) array[i] = temp_list[j];
      
    }
  }
}

/*calculate loopfreq, which actually is the nesting */
void find_nesting(struct bblock *top){
  struct bblock *cblk;
  struct blist *looplist, *looptop, *clist, *templist;
  int loopcnt = 0, nest, i, j;
  int exits[100], numexits;
  looplist = NULL;
  looptop = NULL;

  /*get a ptr to all the blocks that have loop info*/
  for (cblk=top;cblk;cblk = cblk->next){
    if (cblk->loop) { 
      cblk->loopno = ++loopcnt;
      if (looptop) {
	looplist->next = (struct blist *) malloc(sizeof(struct blist));
	looplist = looplist->next;
      }
      else looptop = looplist =  (struct blist *) malloc(sizeof(struct blist));
            
      looplist->ptr = cblk;
      looplist->next = NULL;
    }
  }

  /*find the nesting of the loops... if loop_blocklist of one loop is
    inside the loop_blocklist of another loop, then it is a nested loop*/
  looplist = NULL;
  for (clist = looptop;clist;clist = clist->next){
    nest = 1;
    for (looplist = looptop; looplist; looplist = looplist->next) {
      if (looplist == clist) continue;
      if (subset(clist->ptr->loop_blocklist,clist->ptr->loop_bno,
		 looplist->ptr->loop_blocklist,looplist->ptr->loop_bno)) nest++;

    }
    clist->ptr->loopfreq = nest;
  }

  /*spread the nesting information to all the blocks in the loop */
  /*REMOVE THIS LATER.....*/
  /*  for (clist = looptop; clist; clist = clist->next) {
    for(i=0; i<clist->ptr->loop_bno; i++) {
      for(cblk = top; cblk; cblk = cblk->next) {
	if (clist->ptr->loop_blocklist[i] == cblk->nr) {
	  if (clist->ptr->loopfreq > cblk->loopfreq) {
	    cblk->loopfreq = clist->ptr->loopfreq;
	    cblk->loopno = clist->ptr->loopno;
	    cblk->loop_blocklist = clist->ptr->loop_blocklist;
	  }
	}
      }
    }
    }*/
  /*a better way yo do the above */

  for (looplist = looptop; looplist ; looplist = looplist->next) {
    for (clist = looplist->ptr->loop;clist;clist=clist->next) {
      if (looplist->ptr->loopfreq > clist->ptr->loopfreq ) {
	clist->ptr->loopfreq = looplist->ptr->loopfreq;
	clist->ptr->loopno = looplist->ptr->loopno;
	clist->ptr->loop_blocklist = looplist->ptr->loop_blocklist;
	clist->ptr->loop_bno = looplist->ptr->loop_bno;
      }
    }
  }

  numexits =0;
  /*while we are here ... lets calculate the exit paths in the loops */
  /*ASSUMPTION - NO LOOP MEMBER BB CAN HACE 0 SUCCS */
  for (looplist = looptop; looplist; looplist = looplist->next) {
    for (clist = looplist->ptr->loop; clist; clist= clist->next){
      for (templist = clist->ptr->succs; templist; templist = templist->next) {
	if (!subset(&templist->ptr->nr, 1, looplist->ptr->loop_blocklist, looplist->ptr->loop_bno)) {
	  exits[numexits] = clist->ptr->nr;
	  numexits++;
	}
      }
    }
    
    looplist->ptr->exitpaths = (int *) malloc(numexits * sizeof(int));
    for (i=0; i<numexits; i++) {
      looplist->ptr->exitpaths[i] = exits[i];
    }
    looplist->ptr->numexits = numexits;

    sort_list(looplist->ptr->exitpaths, looplist->ptr->numexits);

    numexits =0;
  }
  numexits = 0;
  /*finding the exit paths for the function */
  for (cblk = top; cblk; cblk=cblk->next) {
    if (cblk->succs == NULL) {
      exits[numexits] = cblk->nr;
      numexits++;
    } 
  }
  top->exitpaths = (int *) malloc(numexits * sizeof(int));
  for (i=0; i<numexits; i++) {
    top->exitpaths[i] = exits[i];
  }
  top->numexits = numexits;
  sort_list(top->exitpaths, top->numexits);
}


void remove_dead_blocks(struct bblock *top) {
  struct bblock * currptr, *tempptr;
  struct line *end;

  currptr = NULL;

  for (currptr = top; currptr; currptr = currptr->next) {
    if (currptr->num_inst == 0) {
      for (end = currptr->rtls; end->next; end = end->next) ;

    printf ("%p %p %p %p %p %p %p\n",currptr->preds, currptr->succs, currptr->doms, currptr->loop
	  ,currptr->instrs, currptr ->loop_blocklist, currptr->exitpaths);
    

    }
  }
  printf ("%p %p %p %p %p %p %p\n",top->preds, top->succs, top->doms, top->loop
	  ,top->instrs, top ->loop_blocklist, top->exitpaths);

}




/***********************************************************************
 * main - use: CFG <ass-file-in> <ass-file-out>
 * 
 *
 ***********************************************************************/

main(int argc, char **argv) {
  struct line *rtls, *kiran; /* pointer to 1st line */
  struct bblock *top, *cblk;
  struct temp_inst *curptr, *head;
  FILE *newf, *in, *out, *loopin, *outpath, *outinf;
  FILE *outdata, *datain;
  char c, str[MAXSTR], *temp, pathname[20], infname[20], loopname[20], tempstr[MAXSTR];
  char datainname[MAXSTR], dataoutname[MAXSTR];
  char mybuffy[MAXSTR] = "";
  char mylabel[MAXSTR] = "";
  char mytarget[MAXSTR] = "";
  int myinst = 0;
  int NEWFUNC=FALSE, DO_INF = TRUE, DO_PATH = TRUE, DO_DEBUG = TRUE;  
  int nr, nextOff=0, tempop=0, i;

  head = curptr =NULL;
  rtls = NULL;
  top = cblk = NULL;
  newf = in = out = loopin = outpath = outinf = NULL;
  temp = NULL;
  /*Setup the FILE pointers */


  if (argc < 3)     error("use: patchmaker [flags]  <ass-file-in> <ass-file-out>\n");

  i = 1;
  while (*argv[i]=='-') {
    if (*(argv[i]+1)=='i') {
      DO_INF = FALSE;
      printf("No inf file will be created\n");
    }
    else if (*(argv[i]+1)=='p') {
      DO_PATH = FALSE;
      printf("No path file will be created\n");
    }
    else if (*(argv[i]+1)=='d') {
      DO_DEBUG = FALSE;
      printf("No debug file will be created\n");
    }
    i ++;
  }
  
  if ((in = fopen(argv[i],"r"))==NULL)  error("Error reading input file \n");
  strcpy(loopname,argv[i]);
  i++;

  if (DO_DEBUG) {
    if ((out = fopen(argv[i],"w"))==NULL)  error("Error writing output file \n");
  }
  else out = NULL;
  
  strcpy(pathname,argv[i]);
  strcpy(infname,argv[i]);
  strcat(pathname,".path");
  strcat(infname,".inf");
  strcpy(strstr(loopname,".s"),".loop");
  strcpy(datainname,argv[i]);
  strcpy(dataoutname,argv[i]);
  strcat(datainname,".dat");
  strcat(dataoutname,".dst");

  //  printf("%s %s\n",dataoutname, datainname);
  if (DO_PATH) 
    if ((outpath = fopen(pathname,"w"))==NULL)  error("Error writing output file \n");  
  

  if (DO_INF) {
    if ((loopin = fopen(loopname,"r"))==NULL) error("Error opening loop file\n");
    if ((outinf = fopen(infname,"w"))==NULL)  error("Error writing output file \n");  
    /*    if ((outdata = fopen(dataoutname,"w"))==NULL)  error("Error writing output data file \n");  
	  if ((datain = fopen(datainname,"r"))==NULL)  error("Error writing input data file \n");  */
    
  }

  newf = fopen("instset.curr","r");
  if (newf == NULL) error("error reading instset.curr");

  while (tempop!=EOF && tempop!=-1) {
    fscanf(newf," %d %s ",&tempop,tempstr);

    if (head) {
      curptr->next = (struct temp_inst *) malloc(sizeof(struct temp_inst));
      if (curptr->next == NULL) error("out of heap memory for temp inst");
      curptr = curptr->next;
    }
    else {
      head = curptr = (struct temp_inst *) malloc(sizeof(struct temp_inst));
      if (curptr == NULL) error("out of heap memory for temp inst");
    }
    
    curptr->next = NULL;
    strcpy(curptr->str, tempstr);
    curptr->opcode = tempop;
  }


  /*BEFORE STARTING WITH THE NEW READ_BLOCK, TAKE CARE OF INITIAL
    COMMENTS AND DECLARATIONS IN THE SOURCE */

  /*ASSUMPTION FIRST BASIC BLOCK STARTS WITH  A FUNCTION */

  read_initialize(in, out);

  while ((rtls = read_block(in, mybuffy, mylabel, mytarget, &myinst, &NEWFUNC)) != NULL) {
    top = (struct bblock *) NULL;
    nr = 1;
    while (rtls != NULL ) {

      if (top == NULL) {
	top = cblk = (struct bblock *) malloc(sizeof(struct bblock));
	if (cblk == NULL)
	  error("out of heap memory for blocks");
      }
      else {
	cblk->next = (struct bblock *) malloc(sizeof(struct bblock));
	if (cblk->next == NULL)
	  error("out of heap memory for blocks");
	cblk = cblk->next;
      }
      cblk->rtls = rtls;
      cblk->preds = cblk->succs = NULL;
      cblk->next = NULL;
      cblk->doms = NULL;
      cblk->loop = NULL; 
      cblk->offset = nextOff; 
      cblk->num_inst = 0; 
      cblk->loopno = 0;
      cblk->loopfreq = 0;
      cblk->loop_blocklist = NULL;
      cblk->exitpaths = NULL;
      cblk->numexits = 0;
      cblk->loop_bno = 0;
      cblk->nr = nr++;
 
      if (mylabel[0]!='\0') {
	temp = index(mylabel,LABEL_TERM);
	if (temp) *temp = '\0';
      }
      
      if (mytarget[0]!='\0') {
	temp = index(mytarget,'\n');
	if (temp) *temp = '\0';
      }

      strcpy(cblk->label, mylabel);
      strcpy(cblk->target,mytarget);
      cblk->num_inst = myinst;
      nextOff += myinst*4;
      cblk->instrs = read_instrs(head, cblk->rtls, cblk->nr, &cblk->num_inst);

      if (NEWFUNC) {
	NEWFUNC= FALSE;
	break;
      }

      rtls = read_block(in, mybuffy, mylabel, mytarget, &myinst, &NEWFUNC);

    }

    if (top) {
      /*      printf("Removing dead blocks... \n");
      remove_dead_blocks(top);
      printf("...done\n");*/
      printf("Making CFG...\n");
      printf("\tfunction name is %s\n",top->label);
      make_cfg(top);
      printf("...done\n");
      printf("Finding dominators");
      find_doms(top);
      printf("...done\n");
      printf("Finding loops");
      find_loops(top);
      find_nesting(top);
      printf("...done\n");

      /*OPTIMIZE THE CFG... */
      /*BEFORE WRITNG... MAKE PATH INFO  */
      if (DO_PATH) {
	printf("Writing PATH file");
	path_info(outpath, top);
	printf("...done\n");
      }

      

      /*AND THE INF INFO */
      if (DO_INF) {
	printf("Data analysis \n");
	/*	data_info(outdata,datain,top);*/
	printf("..done\n");

	printf("Writing INF file using input");
	inf_info(outinf, loopin, top);
	printf("...done\n");
      }


      if (DO_DEBUG) {
	for (cblk = top; cblk; cblk = cblk->next)
	  write_block(out, cblk);

      }
    }
    //    write_rtls(rtls); /* write declaration */
  }

  if (DO_DEBUG) 
    fclose(out);

  fclose(in);

  if (DO_INF) {
  fclose(loopin);
  fclose(outinf);
  }

  if (DO_PATH)
    fclose(outpath);

  fclose(newf);

}

/*
data_info(FILE * outdata, FILE *datain, struct bblock *top) {

  char str[5000];
  int i;
  
  printf("HELLO\n\n\n\n");
  i =0;
  while ((fgets(str, 5000, datain)) != NULL) {
    printf("%d %s \n",i++, str);
  }

  



}
*/
