/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: category.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Instruction categorization
 */

/*
 * #define LINEAR
 * for old categorization w/ linear cache state for set-associative caches,
 * has some problems as documented in ex_linear[2] with set size 2
 */

#include "scs.h"

struct f_statinfo_s {
  float hits;
  float misses;
  float firstmisses;
  float firsthits;
};

static struct f_statinfo_s best_statinfo = {0.0, 0.0, 0.0};
static struct f_statinfo_s worst_statinfo = {0.0, 0.0, 0.0};

/*
 * infmlist - determine if line already in first miss list
 */
int infmlist(fmptr, line)
struct fmline_t *fmptr;
int line;
{
  while (fmptr)
    if (fmptr->line == line)
      return(TRUE);
    else
      fmptr = fmptr->next;
  return(FALSE);
}

/*
 * addfmlist - add to fmlist
 */
void *addfmlist(root, line)
struct fmline_t **root;
int line;
{
  struct fmline_t *fmptr;
  fmptr = (struct fmline_t *) MALLOC(sizeof(struct fmline_t));
  fmptr->line = line;
  fmptr->next = *root;
  *root = fmptr;
}

#ifdef BIT_ENCODE
/*
 * addcostlist - add to costlist
 */
void *addcostlist(root, pinfo, cost)
struct costlist_t **root;
struct pathinfo *pinfo;
int cost;
{
  struct costlist_t *costptr;
  costptr = (struct costlist_t *) MALLOC(sizeof(struct costlist_t));
  costptr->cost = cost;
  costptr->path = pinfo;
  costptr->next = *root;
  *root = costptr;
}
#endif /*BIT_ENCODE*/

/*
 * effected_paths - count the conflicting paths whose states need to be changed
 */
int effected_paths(line, instance, pinfo, conf_line, thisset)
int line, instance, conf_line;
struct pathinfo *pinfo;
int thisset;
{
  int k, max;
  int effect = 0;
  int set = line % numsets;

  struct pathlistlist *lptr;
  struct pathlist *pptr;
  struct instinfo *iptr;
  struct callerlist *cptr;
  struct conf_t *rconf;
  vector in;

  for (lptr = setpath[set]; lptr; lptr = lptr->next)
    for (pptr = lptr->list; pptr; pptr = pptr->next)
      for (iptr = pptr->path->insts; iptr; iptr = iptr->next)
	if (iptr->firstline <= conf_line - numsets &&
	    conf_line - numsets <= iptr->lastline)
	  for (k = 0, cptr = pptr->path->curfunc->caller; cptr;
	       k++, cptr = cptr->next) {
	    if (pptr->path == pinfo && k == instance)
	      continue; /* exclude source of conflict */
	    in = vinit();
	    cache_rpath(pptr->path, pptr->path->rout[k].v, &in, thisset);
	    if (vinset(in, line)) {
	      effect++; /* for each conflicting line */
	      rconf = (struct conf_t *) MALLOC(sizeof(struct conf_t));
	      rconf->lline = conf_line;   /* local line */
	      rconf->rline = line;        /* remote line */
	      rconf->instance = instance; /* remote instance */
	      rconf->path = pinfo;        /* remote path */
	      rconf->next = pptr->path->conf[k];
	      pptr->path->conf[k] = rconf;
	    }
	    vfree(in);
	  }

  return(effect);
}

/*
 * category - determine the category for cache prediction of a program line
 * i_in_w: current line i in input cache state
 * i_in_l: current line i in linear cache state
 * l_in_w: number of lines conflicting with i in input cache state
 * l_in_r: number of lines conflicting with i in reaching state
 */
int category(i_in_w, i_in_l, l_in_w, l_in_r)
     int i_in_w, i_in_l, l_in_w, l_in_r;
{
  if (!i_in_w)
    return(C_MISS);
  if (l_in_w < setsize && i_in_l)
    return(C_HIT);
  if (l_in_r < setsize) /* check category of overlaps later */
    return(C_FIRSTMISS);
  return(C_CONFLICT);
}

/*
 * is_line_in_linear - return true if line "l" in linear state
 * of all predecessors (w/o backedges) NOT sharing the same line in instance "k"
 */
int is_line_in_linear(pinfo, k, l, cptr, set)
struct pathinfo *pinfo;
int k, l;
struct callerlist *cptr;
int set;
{
  struct pathlist *pptr, *pptr2;
  struct callerlist *cptr2;
  struct instinfo *iptr;
  int i, s, result = TRUE, or_result;
  vector v, out, diff[MAXSETSIZE];

  if (!pinfo->preds && pinfo->curfunc != mainfuncptr) {
    i = cptr->instance;
    for (iptr = cptr->caller->insts; iptr; iptr = iptr->next)
      if (iptr->firstline <= l - numsets && l - numsets <= iptr->lastline)
	break;
    if (!iptr) {
      v = vinit();
      init_set(diff);
      or_result = FALSE;
      for (s = 0; s < setsize; s++) {
	out = vinit();
	cache_path(cptr->caller, cptr->caller->lin[i][s].v, &out, diff, s, v, set);

	or_result |= vinset(out, l);
	
	vfree(out);
	if (s + 1 < setsize)
	  v = diff[s];
      }
      result &= or_result;
      free_set(diff);
    }
  }
  else for (pptr = pinfo->preds; pptr; pptr = pptr->next) {
    if (pptr->path->func)
      for (i = 0, cptr2 = pptr->path->func->caller; cptr2;
	   i++, cptr2 = cptr2->next) {
	if (cptr2->succ_pathlabel == pptr->path->succs->path->pathlabel
	    && cptr2->func == pptr->path->curfunc
	    && cptr2->instance == k)
	  for (pptr2 = pptr->path->func->exitpaths; pptr2; pptr2 = pptr2->next){
	    for (iptr = pptr2->path->insts; iptr; iptr = iptr->next)
	      if (iptr->firstline <= l - numsets &&
		  l - numsets <= iptr->lastline)
		break;
	    if (!iptr) {
	      v = vinit();
	      init_set(diff);
	      or_result = FALSE;
	      for (s = 0; s < setsize; s++) {
		out = vinit();
		cache_path(pptr2->path, pptr2->path->lin[i][s].v,
			   &out, diff, s, v, set);

		or_result |= vinset(out, l);

		vfree(out);
		if (s + 1 < setsize)
		  v = diff[s];
	      }
	      result &= or_result;
	      free_set(diff);
	    }
	  }
      }
    else if (!is_backedge(pinfo, pptr->path)) {
      for (iptr = pptr->path->insts; iptr; iptr = iptr->next)
	if (iptr->firstline <= l - numsets && l - numsets <= iptr->lastline)
	  break;
      if (!iptr) {
	v = vinit();
	init_set(diff);
	or_result = FALSE;
	for (s = 0; s < setsize; s++) {
	  out = vinit();
	  cache_path(pptr->path, pptr->path->lin[k][s].v,
		     &out, diff, s, v, set);

	  or_result |= vinset(out, l);

	  vfree(out);
	  if (s + 1 < setsize)
	    v = diff[s];
	}
	result &= or_result;
	free_set(diff);
      }
    }
  }

  return(result);
}

/*
 * is_other_first_miss - return false if one overlapping line is classified
 * as a conflict or miss.
 * The problem we try to address here is that a line can only be a firstmiss
 * for a given path if all other paths sharing this line also classify it as
 * a firstmiss. Thus, if miss + conflict > 0, we cannot classify the line as
 * firstmiss.
 */
int is_other_first_miss(line, thisset)
int line;
int thisset;
{
  int i = line + numsets, k, l, s;
  int set = line % numsets;
  int tag = line / numsets;
  struct pathlistlist *lptr;
  struct pathlist *pptr;
  struct pathinfo *pinfo;
  struct instinfo *iptr;
  struct callerlist *cptr;
  vector w, diff[MAXSETSIZE], prev_diff, lin, reach, ws[MAXSETSIZE];
  int firstmiss = 0, miss = 0, conflict = 0;
  int i_in_w, i_in_l, l_in_w, l_in_r;

  for (lptr = setpath[set]; lptr; lptr = lptr->next)
    if (lptr->tag == tag)
      break;

  for (pptr = lptr->list; pptr; pptr = pptr->next) {
    pinfo = pptr->path;
    for (k = 0, cptr = pinfo->curfunc->caller; cptr; k++, cptr = cptr->next) {
      w = lin = reach = vinit();
      init_set(diff);
      for (s = 0; s < setsize; s++) {
	ws[s] = vinit();
	vcpy(&ws[s], pinfo->in[k][0].v);
      }
      l = 0;
      for (s = 0; s < setsize; s++) {
	prev_diff = (s > 0 ? diff[s-1] : vinit());
	for (iptr = pinfo->insts; iptr && l < i; iptr = iptr->next)
	  for (l = iptr->firstline + numsets; l < i; l++)
	    if (l % numsets == thisset)
	      cache_line(&ws[s], l, diff, s, prev_diff);
      }
      free_set(diff);
      for (s = 0; s < setsize; s++)
	vunion(&w, ws[s]);
#ifdef LINEAR
      vcpy(&lin, pinfo->lin[k][0].v);
#else
      vcpy(&lin, pinfo->din[k]);
#endif
      vcpy(&reach, pinfo->rout[k].v);
	    
      i_in_w = vinset(w, i);
#ifdef LINEAR
      i_in_l = vinset(lin, i) && is_line_in_linear(pinfo, k, i, cptr, thisset);
#else
      i_in_l = vinset(lin, i);
#endif
      l_in_w = l_in_r = 0;
      for (l = set + numsets; l < vecbits; l += numsets)
	if (i != l) {
	  if (vinset(w, l))
	    l_in_w++;
	  if (vinset(reach, l))
	    l_in_r++;
	}
      
      switch (category(i_in_w, i_in_l, l_in_w, l_in_r)) {
      case C_FIRSTMISS:
	firstmiss++;
	break;
      case C_MISS:
	miss++;
	break;
      case C_CONFLICT:
	conflict++;
	break;
      default:
	break;
      }
      vfree(w);
      vfree(lin);
      vfree(reach);
      for (s = 0; s < setsize; s++)
	vfree(ws[s]);
    }
  }

  /*
   * kluge: force firstmiss if no conflicting program lines,
   * i.e. when program fits in cache.
   */
  if (firstmiss > 0 && miss > 0 && conflict == 0 && setpath[set]->next == NULL)
    miss = 0;

  return(miss + conflict == 0);
}

/*
 * output_line - print out line information for prediction
 */
void output_line(pinfo, iptr, inst, line, cat, v, set)
     struct pathinfo *pinfo;
     struct instinfo *iptr;
     int inst, line, cat;
     vector v;
     int set;
{
  struct callerlist *cptr;
  int i, k, s, t, succtype, numinst, max;
  int nesting = 1;
  struct statinfo_s best_nesting, worst_nesting;
  int inst_offset = iptr->addr / INSTSIZE;
  int func_offset = pinfo->curfunc->entrypaths->path->insts->addr / INSTSIZE;
  char str[MAXSTRING];
  extern char *strcpy();

  str[0] = '\0';

  switch (cat) {
  case C_MISS:
    strcpy(str, "m/m ");
    best_statinfo.misses++;
    worst_statinfo.misses++;
    break;
  case C_HIT:
    strcpy(str, "h/h ");
    best_statinfo.hits++;
    worst_statinfo.hits++;
    break;
  case C_FIRSTMISS:
  case C_CONFLICT:
    best_nesting.hits =
      best_nesting.misses =
	best_nesting.firstmisses =
	  best_nesting.firsthits =
	    best_nesting.conflicts =
	      worst_nesting.hits =
		worst_nesting.misses =
		  worst_nesting.firstmisses =
		    worst_nesting.firsthits =
		      worst_nesting.conflicts = 0;

    /*
     * for first misses and conflicts,
     * iterate through all loop levels and predict for each level
     */
    {
      struct pathinfo *loopheader;
      struct pathlist *loopheader_preds;
      vector w;
      int l, i_in_w, l_in_w, i_in_p, i_in_l, l_in_lu;
#ifdef TRASH
      int l_in_l;
#endif
      int last_i_in_p = TRUE, loopheader_preds_i;
      int set = line % numsets;
      char worst = 'f', best = 'h';
      
      i = inst;
      for (loopheader = pinfo->loopheader; loopheader; nesting++) {
	/*
	 * allow transitions only
	 * worst: f --> m
	 * best:  h --> f
	 * after transition, no more change
	 */
	if (worst == 'f' || best != 'm') {
	  w = vinit();
	  vcpy(&w, loopheader->loopmask);
	  vinter(&w, v);
	  i_in_w = vinset(w, line + numsets);
	  l_in_w = 0;
	  for (l = set + numsets; l < vecbits; l += numsets)
	    if (line + numsets != l && vinset(w, l)) {
	      l_in_w++;
#ifdef TRASH
	      if (loopheader->freq > 0)
		for (s = 0; s < setsize; s++) {
		  if (vinset(loopheader->in[i][s].v, l)) {
		    struct pathlist *pptr2;
		
		    for (pptr2 = loopheader->preds; pptr2; pptr2 = pptr2->next)
		      if (pptr2->path->loopheader == loopheader) {
			vector v, out, diff[MAXSETSIZE];
			v = vinit();
			init_set(diff);
			for (t = 0; t < setsize; t++) {
			  out = vinit();
			  cache_path(pptr2->path,
				     pptr2->path->in[i][t].v,
				     &out, diff, t, v, set);
			  vunion(&v, out);
			  vfree(out);
			  if (vinset(v, l)) {
			    free_set(diff);
			    break; /* line retained along backedge */
			  }
			  if (t + 1 < setsize)
			    v = diff[t];
			  else
			    free_set(diff);
			}
			if (t < setsize)
			  break;
		      }
		    if (pptr2)
		      break;
		    else
		      l_in_w--;
		  }
		  if (s < setsize)
		    break;
		}
#endif
	    }
	  vfree(w);

	  i_in_p = vinset(loopheader->preheaders[i], line + numsets);

	  w = vinit();
	  for (s = 0; s < setsize; s++)
	    vunion(&w, pinfo->lin[i][s].v);

#ifdef LINEAR      
	  i_in_l = vinset(w, line + numsets);
#else
	  i_in_l = vinset(pinfo->din[i].v, line + numsets);
	  /* make l_in_lu for $(s(p) /cup linear) /cap u$ */
	  vunion(&w, loopheader->preheaders[i]);
#endif
#ifdef TRASH
	  l_in_l = 0;
	  for (l = set + numsets; l < vecbits; l += numsets)
	    if (line + numsets != l && vinset(w, l))
	      l_in_l++;
#endif

	  vinter(&w, loopheader->loopmask);
	  l_in_lu = 0;
	  for (l = set + numsets; l < vecbits; l += numsets)
	    if (line + numsets != l && vinset(w, l))
	      l_in_lu++;
	  vfree(w);
	  
	  if (worst == 'f')
	    worst = (l_in_w >= setsize/* || !loopheader->preds*/ ? 'm' : 'f');
	  if (best != 'm')
            best = (i_in_l ? 'h' : (last_i_in_p ? 'f' : 'm'));
	  if (!l_in_w && !i_in_w && !vinset(v, set)) {
	    fprintf(stderr, "impossible output state combination, exit\n");
	    exit(-1);
	    /* worst = best = 'm';*/
	  }
	}

#ifdef TRASH
	/*
	 * commented out again, see ex_rtss / Fig. 8 dissertation, inst 13
	 *
	 * Correct best-case hits, if there is a conflicting line.
	 * Make it a first-miss if the current line was in the prev. preheader,
	 * otherwise make it a miss (permanent transition fm -> m).
	 */
	if (best == 'h' && l_in_l)
	    best = (last_i_in_p ? 'f' : 'm');
#endif

	/*
	 * always try to make the transition to a first hit;
	 * once a first hit, always a first hit (both best and worst)
	 */
	if (best != 'i' || worst != 'i') {
	  char old_best = best, old_worst = worst;
	  int conf_line;
	  int l_in_p, i_in_b, i_in_pd;
	  struct pathlist *pptr;
	  
	  w = vinit();
	  vcpy(&w, loopheader->loopmask);
	  vinter(&w, v);
#ifdef TRASH
	  i_in_w = vinset(w, line + numsets);
#endif
	  /*
	   * only one conflicting line?
	   */
	  l_in_w = 0;
	  for (l = set + numsets; l < vecbits; l += numsets)
	    if (line + numsets != l && vinset(w, l))
	      l_in_w++;
	  vfree(w);
	  
	  /*
	   * i_in_p checks for abstract cache state of preheaders within loop,
	   * i.e. intersect lines within the loop nesting level
	   */
	  w = vinit();
	  vcpy(&w, loopheader->preheaders[i]);
	  vinter(&w, v);
	  i_in_p = vinset(w, line + numsets);
	  l_in_p = 0;
#ifdef LINEAR
	  for (l = set + numsets; l < vecbits; l += numsets)
	    if (line + numsets != l &&
		vinset(w, l))
	      l_in_p++;
#endif
	  vfree(w);
	  
	  if (!(loopheader_preds = loopheader->preds) &&
	      loopheader->curfunc != mainfuncptr) {
	    for (k = 0, cptr = loopheader->curfunc->caller; cptr;
		 k++, cptr = cptr->next)
	      if (i == k)
		break;
	    loopheader_preds =
	      (struct pathlist *) MALLOC(sizeof(struct pathlist));
	    loopheader_preds->path = cptr->caller;
	    loopheader_preds->next = (struct pathlist *) NULL;
	    loopheader_preds_i = cptr->instance;
	  }
	  else
	    loopheader_preds_i = i;
			
	  /*
	   * The condition for a first hit is rather restrictive here:
	   * b/w (1) line in abstract cache states of current path
	   * b/w (2) exists a conf. line in abstract cache state of curr. p
	   * b/w (3) line in abstract cache state of ALL preheaders
	   *   w (4) conf lines NOT in abstract cache state of preheader
	   * b/w (5) line in post-dom of header (line is always reached in loop)
	   *   w (6) conf lines NOT in linear cache state of curr. p
	   *         (conf line CANNOT precede curr. p when entering loop)
	   * b   (7) line NOT in abstract cache state of backedge
	   *         (always replaced by conf line during one loop iteration)
	   *     (8) conf lines in abstract cache state of backedge
	   *         (implied by (2), (4), and (6))
	   *   w (9) line in linear cache state
	   */
	  if (loopheader_preds &&
	      i_in_w && l_in_w >= setsize && i_in_p) {     /* (1), (2), (3) */
	    if (l_in_p < setsize && l_in_lu < setsize &&   /* (4), (6) */
		i_in_l)                                    /* (9) */
              worst = 'i';

	    i_in_p = TRUE;
	    i_in_b = FALSE;

	    w = vinit();
	    cache_rpath(loopheader, loopheader->pout[i].v, &w, set);
	    i_in_pd = vinset(w, line + numsets);
	    vfree(w);

	    /*
	     * for each predecessor of the header (either backedge or preheader)
	     */
	    for (pptr = loopheader_preds; pptr; pptr = pptr->next) {
	      vector v, out, diff[MAXSETSIZE];
	      w = v = vinit();
	      init_set(diff);
	      for (s = 0; s < setsize; s++) {
		out = vinit();
		cache_path(pptr->path,
			   pptr->path->in[loopheader_preds_i][s].v,
			   &out, diff, s, v, set);
		vunion(&v, out);
		vunion(&w, v);
		vfree(out);
		if (s + 1 < setsize)
		  v = diff[s];
	      }
	      free_set(diff);

	      if (is_backedge(loopheader, pptr->path)) {
		vinter(&w, loopheader->loopmask);
		i_in_b |= vinset(w, line + numsets);
		vfree(w);
	      }
	      else if (i_in_p) { /* preheader */
		i_in_p = vinset(w, line + numsets);
		vfree(w);
	      }
	    }
	    
	    /*
	     * This condition (3) is stronger than the one above
	     */
	    if (worst == 'i' && i_in_p && i_in_pd)         /* (3), (5) */
	      /* worst = 'i'*/;
	    else
	      worst = old_worst;

	    if (i_in_p && i_in_pd && !i_in_b)              /* (3), (5), (7) */
	      best = 'i';
	  }
	  if (!loopheader->preds)
	    FREE(loopheader_preds);
	}

	switch (best) {
	case 'h': best_nesting.hits++;        break;
	case 'f': best_nesting.firstmisses++; break;
	case 'i': best_nesting.firsthits++;   break;
	}
	switch (worst) {
	case 'm': worst_nesting.misses++;      break;
	case 'f': worst_nesting.firstmisses++; break;
	case 'i': worst_nesting.firsthits++;   break;
	}

	sprintf(&str[strlen(str)], "%c/%c ", worst, best);

#ifdef TRASH
	/*
	 * adjust miss back to first miss if m->f transition due
	 * to loop frequency 1, i.e. function level. This forces
	 * a reevaluation of the classification at the next loop level.
	 */
	if (worst == 'm' && !loopheader->preds)
	  worst = 'f';
#endif

	/*
	 * find next outer loop header
	 */
	if (loopheader->prevheader) { /* next outer loop */
	  last_i_in_p = vinset(loopheader->preheaders[i], line + numsets);
	  loopheader = loopheader->prevheader;
	}
	else if (loopheader->curfunc == mainfuncptr) /* terminate at main */
	  break;
	else { /* next loop is caller of function */
	  for (k = 0, cptr = loopheader->curfunc->caller; cptr;
	       k++, cptr = cptr->next)
	    if (i == k)
	      break;
	  loopheader = cptr->succs->path->loopheader;
	  i = cptr->instance;
	}
      }
      best_statinfo.hits +=
	(float) best_nesting.hits         / (float) nesting;
      best_statinfo.firstmisses +=
	(float) best_nesting.firstmisses  / (float) nesting;
      best_statinfo.firsthits +=
	(float) best_nesting.firsthits    / (float) nesting;
      worst_statinfo.misses  +=
	(float) worst_nesting.misses      / (float) nesting;
      worst_statinfo.firstmisses +=
	(float) worst_nesting.firstmisses / (float) nesting;
      worst_statinfo.firsthits +=
	(float) worst_nesting.firsthits   / (float) nesting;
    }
  }

  if (line == iptr->firstline)
    if (line == iptr->lastline)
      numinst = iptr->num;
    else
      numinst = instonline - inst_offset % instonline;
  else if (line == iptr->lastline)
    numinst = iptr->num - (iptr->numlines - 1) * instonline +
      inst_offset % instonline;
  else
    numinst = instonline;
  
  if (line != iptr->firstline)
    inst_offset = (inst_offset / instonline) * instonline;

  max = inst_offset + numinst +
    (line - iptr->firstline) * instonline;

  for (i = inst_offset + (line - iptr->firstline) * instonline; i < max; i++) {
    succtype = (line == iptr->lastline && i + 1 == max && !pinfo->calltoexit ?
		pinfo->succtype : NOCALL);
    fprintf(output_file, "0 inst %d %d %d %d %s%d\n",
	    i - func_offset, pinfo->loopno, line, nesting, str, succtype);
    strcpy(str, "h/h "); /* autohits */
    nesting = 1;
  }
  if (succtype == CALL) {
    {
    for (k = 0, cptr = pinfo->func->caller; cptr; k++, cptr = cptr->next)
      if (cptr->succ_pathlabel == pinfo->succs->path->pathlabel &&
	  cptr->func == pinfo->curfunc &&
	  cptr->instance == inst)
	break;
    }
    fprintf(output_file, "call %s instance %d\n", pinfo->func->funcname, k);
  }
}

/*
 * output_simple_conflicts_body - print out cache behavior according to simple
 * scheme below
 */
void output_simple_conflicts_body(thisset)
  int thisset;
{
  struct funclist *fptr;
  struct callerlist *cptr;
  struct pathlist *pptr;
  struct pathinfo *pinfo;
  struct instinfo *iptr;
  int i, k, l, max, s;
  int hit, miss, firstmiss, conflict, effect, autohit, cat;
  int i_in_w, i_in_l, l_in_w, l_in_r;
  int set, line;
  vector v, w, diff[MAXSETSIZE], prev_diff, lin, reach, ws[MAXSETSIZE];

  static struct statinfo_s statinfo = {0, 0, 0, 0};
  int refs;

  for (fptr = funcs; fptr; fptr = fptr->next)
    for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next) {
#ifdef DEBUG
      printf("func %s instance %d\n", fptr->func->funcname, k);
#endif /*DEBUG*/
      if (predict)
	fprintf(output_file, "-1 func %s instance %d parent %s\n",
		fptr->func->funcname, k,
		(cptr->succs == NULL ?
		 "0" : cptr->succs->path->curfunc->funcname));
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	pinfo = pptr->path;
	hit = miss = firstmiss = conflict = effect = autohit = 0;
	v = w = lin = reach = vinit();	
	for (s = 0; s < setsize; s++) {
	  ws[s] = vinit();
	  vcpy(&ws[s], pinfo->in[k][s].v);
	  vunion(&w, pinfo->in[k][s].v);
	}
#ifdef LINEAR
	vcpy(&lin, pinfo->lin[k][0].v);
#else
	vcpy(&lin, pinfo->din[k].v);
#endif
	vcpy(&reach, pinfo->rout[k].v);
	for (iptr = pinfo->insts; iptr; iptr = iptr->next) {
	  max = iptr->lastline + numsets;
	  for (i = iptr->firstline + numsets; i <= max; i++) {
	    set = i % numsets;
	    line = i - numsets;

	    if (set != thisset)
	      continue;

	    /*
	     * (a) line not yet seen in path
	     * (b) line seen before but has been replaced by another
	     *     conflicting line of the same path since then
	     */
	    if (!vinset(v, i) || !vinset(w, i)) {
	      vinsertset(&v, i);
	      
	      i_in_w = vinset(w, i);

#ifdef LINEAR
	      i_in_l = vinset(lin, i) && is_line_in_linear(pinfo, k, i, cptr, thisset);
#else
	      i_in_l = vinset(lin, i);
#endif

	      l_in_w = l_in_r = 0;
	      for (l = set; l < vecbits; l += numsets)
		if (i != l) {
		  if (vinset(w, l)) {
		    l_in_w++;
		    if (vinset(reach, l))
		      l_in_r++;
		  }
		}

	      if (i_in_w && l_in_w >= setsize)
		for (s = 0; s < setsize; s++)
		  if (vinset(ws[s], i)) {
		    for (l = set; l < vecbits; l += numsets)
		      if (i != l && vinset(ws[s], l))
			break;
		    if (l >= vecbits) {
		      l_in_w = 0;
		      break;
		    }
		  }

	      switch (cat = category(i_in_w, i_in_l, l_in_w, l_in_r)) {
	      case C_MISS:  
		miss++;
#ifdef BIT_ENCODE
		if (!infmlist(pinfo->mmptr, line))
		  addfmlist(&pinfo->mmptr, line);
		if (!infmlist(pinfo->missptr, line))
		  addfmlist(&pinfo->missptr, line);
#endif /*BIT_ENCODE*/
		break;
	      case C_FIRSTMISS:
		if (is_other_first_miss(line, thisset)) {
		  firstmiss++;
		  if (!infmlist(pinfo->fmptr[k], line))
		    addfmlist(&pinfo->fmptr[k], line);
#ifdef BIT_ENCODE
		  if (!infmlist(pinfo->firstmissptr, line))
		    addfmlist(&pinfo->firstmissptr, line);
#endif /*BIT_ENCODE*/
		  break;
		}
		/*
		 * otherwise: some overlapping line is conflict =>
		 * fall thru to make this line conflict as well
		 */
	      case C_CONFLICT:
		conflict++;
		for (l = set + numsets; l < vecbits; l += numsets)
		  if (i != l && vinset(w, l))
		    effect += effected_paths(i, k, pinfo, l, thisset);

		if (!infmlist(pinfo->cfptr[k], i))
		    addfmlist(&pinfo->cfptr[k], i);
#ifdef BIT_ENCODE
		if (!infmlist(pinfo->mmptr, line))
		  addfmlist(&pinfo->mmptr, line);
		if (!infmlist(pinfo->conflictptr, line))
		  addfmlist(&pinfo->conflictptr, line);
#endif /*BIT_ENCODE*/
		break;
	      case C_HIT:
		hit++;
		break;
	      }
	      if (predict)
		output_line(pinfo, iptr, k, line, cat, w, thisset);
	      vfree(w);
	      init_set(diff);
	      for (s = 0; s < setsize; s++) {
		prev_diff = (s > 0 ? diff[s-1] : vinit());
		if (i % numsets == thisset)
		  cache_line(&ws[s], i, diff, s, prev_diff);
	      }
	      free_set(diff);
	      w = vinit();
	      for (s = 0; s < setsize; s++)
		vunion(&w, ws[s]);
	    }
	  }
	  autohit += iptr->num;
	}
	vfree(v);
	vfree(w);
	vfree(lin);
	vfree(reach);
	for (s = 0; s < setsize; s++)
	  vfree(ws[s]);

	autohit -= hit + miss + firstmiss + conflict;
	pinfo->hit[k]       = hit + autohit;
	pinfo->miss[k]      = miss;
	pinfo->firstmiss[k] = firstmiss;
	pinfo->conflict[k]  = conflict;

	statinfo.hits        += pinfo->hit[k];
	statinfo.misses      += pinfo->miss[k];
	statinfo.firstmisses += pinfo->firstmiss[k];
	statinfo.conflicts   += pinfo->conflict[k];

	best_statinfo.hits  += (float) autohit;
	worst_statinfo.hits += (float) autohit;

#ifdef DEBUG
	printf("path %2d insts %2d hits %2d misses %2d first misses %2d conflicts %2d,%2d autohits %2d\n",
	       pinfo->pathnum, hit+miss+firstmiss+autohit+conflict,
	       hit, miss, firstmiss,
	       conflict, effect, autohit);
#endif
      }
    }

#ifdef DEBUG
  refs = statinfo.hits + statinfo.misses +
    statinfo.firstmisses + statinfo.conflicts;
  if (predict) {
    printf("STATIC CACHE PREDICTION for nesting in function instance graph:\n");
    
    printf("BEST case:\n");
    printf("\t%2.2f  hits\n", best_statinfo.hits);
    printf("\t%2.2f%% hits\n",
	   100.0 * best_statinfo.hits        / (float) refs);
    printf("\t%2.2f%% misses\n",
	   100.0 * best_statinfo.misses      / (float) refs);
    printf("\t%2.2f%% first misses\n",
	   100.0 * best_statinfo.firstmisses / (float) refs);
    printf("\t%2.2f%% first hits\n",
	   100.0 * best_statinfo.firsthits / (float) refs);
    
    printf("WORST case:\n");
    printf("\t%2.2f  hits\n", worst_statinfo.hits);
    printf("\t%2.2f%% hits\n",
	   100.0 * worst_statinfo.hits        / (float) refs);
    printf("\t%2.2f%% misses\n",
	   100.0 * worst_statinfo.misses      / (float) refs);
    printf("\t%2.2f%% first misses\n",
	   100.0 * worst_statinfo.firstmisses / (float) refs);
    printf("\t%2.2f%% first hits\n",
	   100.0 * worst_statinfo.firsthits / (float) refs);
  }
  else {
    printf("STATIC CACHE PREDICTION for function instance graph:\n");
    printf("\t%d hits\n", statinfo.hits);
    printf("\t%2.2f%% hits\n",
	   100.0 * (float) statinfo.hits        / (float) refs);
    printf("\t%2.2f%% misses\n",
	   100.0 * (float) statinfo.misses      / (float) refs);
    printf("\t%2.2f%% first misses\n",
	   100.0 * (float) statinfo.firstmisses / (float) refs);
    printf("\t%2.2f%% conflicts\n",
	   100.0 * (float) statinfo.conflicts   / (float) refs);
  }
#else /*!DEBUG*/
  refs = statinfo.hits + statinfo.misses +
    statinfo.firstmisses + statinfo.conflicts;
  if (predict) {
    printf("STATIC CACHE PREDICTION for nesting in function instance graph:\n");
    
    printf("BEST case:\n");
    printf("\t%2.2f  hits\n", best_statinfo.hits);
    printf("\t%2.2f%% hits\n",
	   100.0 * best_statinfo.hits        / (float) refs);
    printf("\t%2.2f%% misses\n",
	   100.0 * best_statinfo.misses      / (float) refs);
    printf("\t%2.2f%% first misses\n",
	   100.0 * best_statinfo.firstmisses / (float) refs);
    printf("\t%2.2f%% first hits\n",
	   100.0 * best_statinfo.firsthits / (float) refs);
    
    printf("WORST case:\n");
    printf("\t%2.2f  hits\n", worst_statinfo.hits);
    printf("\t%2.2f%% hits\n",
	   100.0 * worst_statinfo.hits        / (float) refs);
    printf("\t%2.2f%% misses\n",
	   100.0 * worst_statinfo.misses      / (float) refs);
    printf("\t%2.2f%% first misses\n",
	   100.0 * worst_statinfo.firstmisses / (float) refs);
    printf("\t%2.2f%% first hits\n",
	   100.0 * worst_statinfo.firsthits / (float) refs);
  } else {
    refs = statinfo.hits + statinfo.misses +
      statinfo.firstmisses + statinfo.conflicts;
    printf("%10d", statinfo.hits + statinfo.misses +
	   statinfo.firstmisses + statinfo.conflicts);
    printf(" %6.2f%%",
	   100.0 * (float) statinfo.hits        / (float) refs);
    printf(" %6.2f%%",
	   100.0 * (float) statinfo.misses      / (float) refs);
    printf(" %6.2f%%",
	   100.0 * (float) statinfo.firstmisses / (float) refs);
    printf(" %6.2f%%",
	   100.0 * (float) statinfo.conflicts   / (float) refs);
  }
#endif /*DEBUG*/
}

/*
 * output_simple_conflicts - print out cache behavior according to simple
 * scheme below
 */
void output_simple_conflicts()
{
  int set;

  for (set = 0; set < numsets; set++)
    output_simple_conflicts_body(set);
}
