/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: dfa.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Data-flow analysis for
 * (1) abstract cache state
 * (2) reaching state
 * (3) post-dominator
 * (4) linear cache state
 * (5) dominator
 */

#include "scs.h"

/*#define DEBUG2*/
/* Modified by Sibin M on May 3, 2004.
 * Commented out the following line - to prevent unnecessary outputs.
 */
                                                                                                       
// #define DEBUG_STATES

void vinsertset(sptr, r)
vector *sptr;
unsigned int r;
{
  vinsert(sptr, r / numsets);
}

void vdeleteset(sptr, r)
vector *sptr;
unsigned int r;
{
  vdelete(sptr, r / numsets);
}

int vinset(ptr, r)
vector ptr;
unsigned int r;
{
  vin(ptr, r / numsets);
}

/*
 * init_counters - initialize counters for dfa
 */
void init_counters()
{
  struct funclist *fptr;
  struct pathlist *pptr, *pptr2;
  int i;

  for (fptr = funcs; fptr; fptr = fptr->next) {
    i = 0;
    for (pptr = fptr->func->exitpaths; pptr; pptr = pptr->next)
      i++;
    fptr->func->numexitpaths = i;
    if (i > MAXPREDS) {
      fprintf(stderr, "ERROR in %s: too many exits: %d > MAXPREDS\n",
	      fptr->func->funcname, i);
      exit(1);
    }

    for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
      i = 0;
      for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
	i++;
      pptr->path->numpreds = i;
      if (i > MAXPREDS) {
	fprintf(stderr, "ERROR in %s path %d: too many preds: %d > MAXPREDS\n",
		fptr->func->funcname, pptr->path->pathnum, i);
	exit(1);
      }
    }
  }
}

/*
 * init_set - initialize list of bit vectors
 */
void init_set(diff)
     vector diff[];
{
  int s;

  for (s = 0; s < setsize; s++)
    diff[s] = vinit();
}

/*
 * free_set - free list of bit vectors
 */
void free_set(diff)
     vector diff[];
{
  int s;

  for (s = 0; s < setsize; s++)
    vfree(diff[s]);
}

/*
 * shift - cascade conflicting lines within difference vector, i.e.
 * from set[s] of the list of bit vectors to set[s+1] if conflicting
 * lines exist
 */
void shift(diff, i, s)
vector diff[];
int i, s;
{
  int k;

  if (s == setsize)
    return;

  if (s + 1 < setsize) {
    for (k = i; k < vecbits; k += numsets)
      if (vinset(diff[s + 1], k))
	break;
    if (k < vecbits)
      shift(diff, i, s + 1);
  }

  for (k = i; k < vecbits; k += numsets)
    if (vinset(diff[s], k)) {
      vdeleteset(&diff[s], k);
      if (s + 1 < setsize)
	vinsertset(&diff[s + 1], k);
    }
}

/*
 * cache_line - add a line to a set and delete the conflicting lines (union the
 *              them instead to diff)
 */
void cache_line(v, i, diff, s, prev_diff)
vector *v, diff[], prev_diff;
int i, s;
{
  int k, last_k = vecbits;
  int first = (s == 0), next = !first;
  int i_in_w = vinset(*v, i);

  if (s == 0)
    vinsertset(v, i);
  else {
    vdeleteset(v, i);
#ifndef TRASH
    if (vinset(prev_diff, i))
      vdeleteset(&prev_diff, i);
#else
    if (vinset(prev_diff, i))
      last_k = vecbits;
    else
#endif
      for (k = i % numsets; k < vecbits; k += numsets)
	if (vinset(prev_diff, k)) {
	  last_k = k;
	  vinsertset(v, k);
	}
  }
  if (s == 0 || last_k < vecbits && vinset(prev_diff, last_k))
    for (k = i % numsets; k < vecbits; k += numsets)
      if (k != i && vinset(*v, k) && !vinset(prev_diff, k)) {
	/*
	 * if a path has two lines mapping into the same cache line,
	 * the first line (in conflict w/ the second) needs to be
	 * shifted into the diff-vector. Also, if there are already
	 * conflicting lines in the diff-vector, shift them further down.
	 */
	if (first) {
	  for (last_k = i % numsets; last_k < vecbits; last_k += numsets)
	    if (vinset(diff[s], last_k))
	      break;
	  if (last_k < vecbits)
	    shift(diff, i % numsets, s);
	  first = FALSE;
	}
	else if (next) {
	  /*
	   * if we have already dealt with two lines mapping into the same
	   * cache line, the diff vectors may cause conflicting lines in
	   * the in-vector to be replaced. These lines can simply be shifted
	   * to the first diff-vector not already containing a conflicting
	   * line, i.e. they are appended to the diff-vector list. This
	   * only works since only the first line (of the two conflicting lines)
	   * will find other conflicting lines in the in-vector. The second
	   * one then does not have any work to do anymore. (If it did,
	   * simple append wouldn't work but rather an insert in the middle
	   * would be required -- but the middle is hard to determine.)
	   */
	  do {
	    for (last_k = i % numsets; last_k < vecbits; last_k += numsets)
	      if (vinset(diff[s], last_k))
		break;
	    if (last_k < vecbits)
	      s++;
	  } while (last_k < vecbits && s < setsize);
	  next = FALSE;
	}
	vdeleteset(v, k);
	if (s < setsize)
	  vinsertset(&diff[s], k);
      }
  if (i_in_w && 0 < s && s < setsize)
    vinsertset(&diff[s], i);
}

/*
 * cache_path - add the lines in the current path and delete conflicting lines
 * set-assoc caches: return diff vector, i.e. add to diff the lines replaced
 *                   by the path's lines (used as init. vec. of next-higher set)
 */
void cache_path(path, in, out, diff, s, prev_diff, set)
struct pathinfo *path;
vector in, *out, diff[], prev_diff;
int s, set;
{
  struct instinfo *iptr;
  register int i, max;

  vcpy(out, in);

  for (iptr = path->insts; iptr; iptr = iptr->next) {
    max = iptr->lastline + numsets;
    for (i = iptr->firstline + numsets; i <= max; i++)
      if (i % numsets == set)
	cache_line(out, i, diff, s, prev_diff);
  }
}
      
/*
 * change_path - check if state has changed; if so, set change flag, ret TRUE
 */
int change_path(vptr, change, in)
struct veclist *vptr;
int *change;
vector *in;
{
  int ret;
  vector old_v = vptr->v;

  if (ret = !vcmp(old_v, *in)) {
    vptr->v = *in;
    *change = TRUE;
    if (old_v != vallptr)
      vfree(old_v);
  }
  else
    vfree(*in);

  return(ret);
}

/*
 * cache_rpath - add the lines in the current path
 * but do NOT delete conflicting lines
 */
void cache_rpath(path, in, out, set)
struct pathinfo *path;
vector in, *out;
int set;
{
  struct instinfo *iptr;
  register int i, max;
  
  vcpy(out, in);

  for (iptr = path->insts; iptr; iptr = iptr->next) {
    max = iptr->lastline + numsets;
    for (i = iptr->firstline + numsets; i <= max; i++)
      if (i % numsets == set)
	vinsertset(out, i);
  }
}

/*
 * cache_loop - add the lines in the current path
 * but do NOT delete conflicting lines and do NOT copy
 */
void cache_loop(path, out, set)
struct pathinfo *path;
vector *out;
int set;
{
  struct instinfo *iptr;
  register int i, max;
  
  for (iptr = path->insts; iptr; iptr = iptr->next) {
    max = iptr->lastline + numsets;
    for (i = iptr->firstline + numsets; i <= max; i++)
      if (i % numsets == set)
	vinsertset(out, i);
  }
}

/*
 * invalidate_union_cache - invalidate cache for incoming vector of entry paths
 * of main for union vectors
 */
void invalidate_union_cache(pptr, numsets, set)
struct pathlist *pptr;
int numsets;
{
  vector first = vinit();
  int i;

  for (i = 0; i < numsets; i++)
    if (i % numsets == set)
      vinsertset(&first, i);
  for (; pptr; pptr = pptr->next) {
    vunion(&pptr->path->in[0][0].v, first);
    pptr->path->in[0][0].changed = TRUE;
  }
  vfree(first);
}

/*
 * invalidate_reaching_cache - invalidate cache for reaching vector
 * of entry paths of main (for reaching vectors)
 */
void invalidate_reaching_cache(pptr, numsets, set)
struct pathlist *pptr;
int numsets;
int set;
{
  int i;
  
  for (; pptr; pptr = pptr->next)
    pptr->path->rout[0].changed = TRUE;
}

/*
 * invalidate_post_dominator_cache - invalidate cache for post-dominating vector
 * of entry paths of main (for reaching vectors)
 */
void invalidate_post_dominator_cache(pptr, numsets, set)
struct pathlist *pptr;
int numsets;
int set;
{
  int i;
  
  for (; pptr; pptr = pptr->next) {
    pptr->path->pout[0].v = vinit();
    pptr->path->pout[0].changed = TRUE;
  }
}

/*
 * invalidate_linear_cache - invalidate cache for linear incoming vector
 * of entry paths of main for union vectors
 */
void invalidate_linear_cache(pptr, numsets, set)
struct pathlist *pptr;
int numsets;
int set;
{
  vector first = vinit();
  int i;
  
  for (i = 0; i < numsets; i++)
    if (i % numsets == set)
      vinsertset(&first, i);
  for (; pptr; pptr = pptr->next) {
    vunion(&pptr->path->lin[0][0].v, first);
    pptr->path->lin[0][0].changed = TRUE;
  }
  vfree(first);
}

/*
 * invalidate_dominator_cache - invalidate cache for dominating vector
 * of entry paths of main (for reaching vectors)
 */
void invalidate_dominator_cache(pptr, numsets, set)
struct pathlist *pptr;
int numsets;
int set;
{
  int i;
  vector v;
  
  for (; pptr; pptr = pptr->next) {
#ifdef TRASH
    if (set == 0) {
      v = vinit();
      vinsertset(&v, 0);
      vdeleteset(&v, 0);
      pptr->path->din[0].v = v;
    }
#else
    v = vinit();
    vinsertset(&v, 0);
    vdeleteset(&v, 0);
    pptr->path->din[0].v = v;
#endif
    pptr->path->din[0].changed = TRUE;
  }
}

/*
 * simple_union_body - calculate union of ins from predecessors
 * (for lines mapping into block set)
 * invariant: set members monotonically increasing
 */
void simple_union_body(set)
     int set;
{
  struct funclist *fptr;
  struct pathlist *pptr, *pptr2;
  struct callerlist *cptr, *cptr2;
  register int i, j, k, s, t, change, local_change;
  vector v, out, diff[MAXPREDS][MAXSETSIZE], prev_diff;

  /*
   * propagate the cache information through all paths
   */
#ifdef DEBUG_STATES
  char str[MAXSTRING];

  fprintf(stdout, "UNION");
#endif
  do {
    change = FALSE;
    /* for each function */
    for (fptr = funcs; fptr; fptr = fptr->next) {
      /* for each function instance */
#ifdef DEBUG_STATES
      fprintf(stdout, "\nscanning func %s instance ", fptr->func->funcname);
#endif
#ifdef DEBUG2
      fprintf(stdout, "u"); fflush(stdout);
#endif
      for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next) {
	do {
	  local_change = FALSE;
	  
	  /* predecessors changed, no calls */
	  for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
	      if (pptr2->path->func)
		break;
	    if (pptr2)
	      continue;
#ifdef TRASH
	    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
	      if (pptr2->path->in[k][0].changed)
		break;
	    if (pptr2) {
#else
	    {
#endif
	      for (j = 0; j < pptr->path->numpreds; j++)
		init_set(diff[j]);
	      for (s = 0; s < setsize; s++) {
		v = vinit();
		vcpy(&v, pptr->path->in[k][s].v);
		for (j = 0, pptr2 = pptr->path->preds; pptr2;
		     j++, pptr2 = pptr2->next) {
		  out = vinit();
		  prev_diff = (s > 0 ? diff[j][s-1] : vinit());
		  vunion(&v, prev_diff);
		  cache_path(pptr2->path, pptr2->path->in[k][s].v,
			     &out, diff[j], s, prev_diff, set);
		  vunion(&v, out);
		  vfree(out);
		}
		local_change |= change_path(&pptr->path->in[k][s],
					    &pptr->path->in[k][0].changed, &v);
	      }
	      for (j = 0; j < pptr->path->numpreds; j++)
		free_set(diff[j]);
	    }
	  }
#ifdef DEBUG_STATES
	  if (local_change)
	    fprintf(stdout, "%d ", k);
#endif
	  if (local_change)
	    change = TRUE;
	} while (local_change);
	
	/*
	 * if the sibling of a caller or an exit path of the callee changed
	 * then propagate outs from union of exit paths and
	 * siblings of caller (predecessors of caller's successors w/o calls)
	 * to ins of caller's successors
	 */
	if (fptr->func != mainfuncptr) {
	  i = cptr->instance;
	  for (pptr2 = cptr->succs; pptr2; pptr2 = pptr2->next) {
#ifdef TRASH
	    for (pptr = fptr->func->exitpaths; pptr; pptr = pptr->next)
	      if (pptr->path->in[k][0].changed)
		break;
	    if (!pptr)
	      for (pptr = pptr2->path->preds; pptr; pptr = pptr->next)
		if (pptr->path->in[i][0].changed)
		  break;
	    if (pptr) {	
#else
	    {
#endif
	      for (j = 0; j < pptr2->path->numpreds + fptr->func->numexitpaths;
		   j++)
		init_set(diff[j]);
	      for (s = 0; s < setsize; s++) {
		v = vinit();
		vcpy(&v, pptr2->path->in[i][s].v);
		for (j = pptr2->path->numpreds, pptr = fptr->func->exitpaths;
		     pptr; j++, pptr = pptr->next) {
		  out = vinit();
		  prev_diff = (s > 0 ? diff[j][s-1] : vinit());
		  vunion(&v, prev_diff);
		  cache_path(pptr->path, pptr->path->in[k][s].v,
			     &out, diff[j], s, prev_diff, set);
		  vunion(&v, out);
		  vfree(out);
		}
		for (j = 0, pptr = pptr2->path->preds; pptr;
		     j++, pptr = pptr->next)
		  if (!pptr->path->func) {
		    out = vinit();
		    prev_diff = (s > 0 ? diff[j][s-1] : vinit());
		    vunion(&v, prev_diff);
		    cache_path(pptr->path, pptr->path->in[i][s].v,
			       &out, diff[j], s, prev_diff, set);
		    vunion(&v, out);
		    vfree(out);
		  }
		change |= change_path(&pptr2->path->in[i][s],
				      &pptr2->path->in[i][0].changed, &v);
#ifdef DEBUG_STATES
		if (s == 0 && pptr2->path->in[i][0].changed)
		  fprintf(stdout, "%du%d ", k, pptr2->path->pathnum);
#endif
	      }
	      for (j = 0; j < pptr2->path->numpreds + fptr->func->numexitpaths;
		   j++)
		free_set(diff[j]);
	    }
	  }
	}
	
	/*
	 * propagate outs at calls into instance of callee, reset changed flag
	 */
	for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
#ifndef TRASH
	  {
#else
	  if (pptr->path->in[k][0].changed) {
#endif
	    pptr->path->in[k][0].changed = FALSE;
	    if (pptr->path->func)
	      for (i = 0, cptr2 = pptr->path->func->caller; cptr2;
		   i++, cptr2 = cptr2->next)
		if (cptr2->succ_pathlabel == pptr->path->succs->path->pathlabel
		    && cptr2->func == pptr->path->curfunc
		    && cptr2->instance == k) {
		  for (pptr2 = pptr->path->func->entrypaths; pptr2;
		       pptr2 = pptr2->next) {
		    init_set(diff[0]);
		    for (s = 0; s < setsize; s++) {
		      v = vinit();
		      vcpy(&v, pptr2->path->in[i][s].v);
		      prev_diff = (s > 0 ? diff[0][s-1] : vinit());
		      vunion(&v, prev_diff);
		      out = vinit();
		      cache_path(pptr->path, pptr->path->in[k][s].v,
				 &out, diff[0], s, prev_diff, set);
		      vunion(&v, out);
		      vfree(out);
		      
		      change |= change_path(&pptr2->path->in[i][s],
					    &pptr2->path->in[i][0].changed, &v);
#ifdef DEBUG_STATES
		      if (s == 0 && pptr2->path->in[i][0].changed)
			fprintf(stdout, "%dd%d ", k, pptr->path->pathnum);
#endif /*DEBUG_STATES*/
		    }
		    free_set(diff[0]);
		  }
		}
	  }
      }
    }
  } while (change);
#ifdef DEBUG_STATES
  printf("\n");

  fprintf(stdout, "input/output states:\n");
  if (verbose) {
    fprintf(stdout, "%26s ", "cache line");
    for (s = 0; s < 2 * setsize; s++) {
      for (i = 0; i < vecbits; i++) {
	for (t = i; t > 10; t %= 10)
	  fprintf(stdout, " ");
	fprintf(stdout, "%d ", i % numsets);
      }
      fprintf(stdout, "      ");
    }
    fprintf(stdout, "\n");

    fprintf(stdout, "%26s ", "program line");
    for (s = 0; s < 2 * setsize; s++) {
      for (i = 0; i < vecbits; i++)
	if (i < numsets)
	  fprintf(stdout, "I ");
	else
	  fprintf(stdout, "%d ", i - numsets);
      fprintf(stdout, "      ");
    }
    fprintf(stdout, "\n");
  }
	
  for (fptr = funcs; fptr; fptr = fptr->next)
    for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	v = vinit();
	init_set(diff[0]);
	for (s = 0; s < setsize; s++) {
	  out = vinit();
	  cache_path(pptr->path, pptr->path->in[k][s].v,
		     &out, diff[0], s, v, set);
	  vunion(&v, out);
	  vget(pptr->path->in[k][s].v, str);
	  if (!verbose || !s)
	    fprintf(stdout, "%10s.%2d[%2d][%2d].",
		    fptr->func->funcname, pptr->path->pathnum, k, s);
	  fprintf(stdout, "in =[%s]", str);
	  if (!verbose)
	    fprintf(stdout, "\n");
	  vget(v, str);
	  if (!verbose)
	    fprintf(stdout, "%10s.%2d[%2d][%2d].",
		    fptr->func->funcname, pptr->path->pathnum, k, s);
	  fprintf(stdout, "out=[%s]", str);
	  if (!verbose || s + 1 == setsize)
	    fprintf(stdout, "\n");
	  vfree(out);
	  if (s + 1 < setsize)
	    v = diff[0][s];
	}
	free_set(diff[0]);
      }
#endif /*DEBUG_STATES*/
}

/*
 * simple_union - calculate union of ins from predecessors
 * invariant: set members monotonically increasing
 */
void simple_union()
{
  int set;

  for (set = 0; set < numsets; set++)
    simple_union_body(set);
}

/*
 * simple_reaching_body - calculate union of reaching from successors,
 * (for lines mapping into block set)
 *                   do NOT cache path but simply ADD the path's lines
 * invariant: set members monotonically increasing
 */
void simple_reaching_body(set)
     int set;
{
  struct funclist *fptr;
  struct pathlist *pptr, *pptr2, *pptr3;
  struct callerlist *cptr, *cptr2;
  register int i, k, change, local_change, path_change;
  vector v, in;

  /*
   * propagate the cache information through all paths
   */
#ifdef DEBUG_STATES
  char str[MAXSTRING];
  int s, t;

  fprintf(stdout, "REACHING");
#endif
  do {
    change = FALSE;
    /* for each function */
    for (fptr = funcs; fptr; fptr = fptr->next) {
      /* for each function instance */
#ifdef DEBUG_STATES
      fprintf(stdout, "\nscanning func %s instance ", fptr->func->funcname);
#endif
#ifdef DEBUG2
      fprintf(stdout, "r"); fflush(stdout);
#endif
      for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next) {
	do {
	  local_change = FALSE;
	  
	  /* successors changed */
	  for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
	    if (!pptr->path->func) {
	      path_change = FALSE;
	      for (pptr2 = pptr->path->succs; pptr2; pptr2 = pptr2->next)
#ifndef TRASH
		{
#else
		if (pptr2->path->rout[k].changed) {
#endif
		  if (!path_change) {
		    v = vinit();
		    vcpy(&v, pptr->path->rout[k].v);
		    path_change = TRUE;
		  }
		  in = vinit();
		  cache_rpath(pptr2->path, pptr2->path->rout[k].v, &in, set);
		  vunion(&v, in);
		  vfree(in);
		}
	      if (path_change)
		local_change |= change_path(&pptr->path->rout[k],
					    &pptr->path->rout[k].changed, &v);
	    }
	  
#ifdef DEBUG_STATES
	  if (local_change)
	    fprintf(stdout, "%d ", k);
#endif
	  if (local_change)
	    change = TRUE;
	} while (local_change);
	
	/*
	 * propagate ins of entrypaths to outs of caller
	 */
	if (fptr->func != mainfuncptr) {
	  i = cptr->instance;
	  path_change = FALSE;
	  for (pptr = fptr->func->entrypaths; pptr; pptr = pptr->next)
#ifndef TRASH
	    {
#else
	    if (pptr->path->rout[k].changed) {
#endif
	      if (!path_change) {
		v = vinit();
		vcpy(&v, cptr->caller->rout[i].v);
		path_change = TRUE;
	      }
	      in = vinit();
	      cache_rpath(pptr->path, pptr->path->rout[k].v, &in, set);
	      vunion(&v, in);
	      vfree(in);
	    }
	  if (path_change)
	    change |= change_path(&cptr->caller->rout[i],
				  &cptr->caller->rout[i].changed, &v);
#ifdef DEBUG_STATES
	  if (cptr->caller->rout[i].changed)
	    fprintf(stdout, "%du%d ", k, cptr->caller->pathnum);
#endif
	}
	
	/*
	 * propagate ins of successors of calls into
	 * exitpaths of instance of callee, reset changed flag
	 */
	for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
#ifndef TRASH
	  {
#else
	  if (pptr->path->rout[k].changed) {
#endif
	    pptr->path->rout[k].changed = FALSE;
	    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
	      if (pptr2->path->func)
		for (i = 0, cptr2 = pptr2->path->func->caller; cptr2;
		     i++, cptr2 = cptr2->next)
		  if (cptr2->succ_pathlabel == pptr->path->pathlabel
		      && cptr2->func == pptr->path->curfunc
		      && cptr2->instance == k)
		    for (pptr3 = pptr2->path->func->exitpaths; pptr3;
			 pptr3 = pptr3->next) {
		      v = vinit();
		      vcpy(&v, pptr3->path->rout[i].v);
		      in = vinit();
		      cache_rpath(pptr->path, pptr->path->rout[k].v, &in, set);
		      vunion(&v, in);
		      vfree(in);
		      
		      change |= change_path(&pptr3->path->rout[i],
					    &pptr3->path->rout[i].changed, &v);
#ifdef DEBUG_STATES
		      if (pptr3->path->rout[i].changed)
			fprintf(stdout, "%dd%d ", k, pptr->path->pathnum);
#endif /*DEBUG_STATES*/
		    }
	  }
      }
    }
  } while (change);
#ifdef DEBUG_STATES
  printf("\n");

  fprintf(stdout, "input/output states:\n");
  if (verbose) {
    fprintf(stdout, "%22s ", "cache line");
    for (s = 0; s < 2; s++) {
      for (i = 0; i < vecbits; i++) {
	for (t = i; t > 10; t %= 10)
	  fprintf(stdout, " ");
	fprintf(stdout, "%d ", i % numsets);
      }
      fprintf(stdout, "      ");
    }
    fprintf(stdout, "\n");

    fprintf(stdout, "%22s ", "program line");
    for (s = 0; s < 2; s++) {
      for (i = 0; i < vecbits; i++)
	if (i < numsets)
	  fprintf(stdout, "I ");
	else
	  fprintf(stdout, "%d ", i - numsets);
      fprintf(stdout, "      ");
    }
    fprintf(stdout, "\n");
  }
	
  for (fptr = funcs; fptr; fptr = fptr->next)
    for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	in = vinit();
	cache_rpath(pptr->path, pptr->path->rout[k].v, &in, set);
	vget(in, str);
	fprintf(stdout, "%10s.%2d[%2d].",
		fptr->func->funcname, pptr->path->pathnum, k);
	fprintf(stdout, "in =[%s]", str);
	if (!verbose)
	  fprintf(stdout, "\n");
	vget(pptr->path->rout[k].v, str);
	if (!verbose)
	  fprintf(stdout, "%10s.%2d[%2d].",
		  fptr->func->funcname, pptr->path->pathnum, k);
	fprintf(stdout, "out=[%s]", str);
	fprintf(stdout, "\n");
	vfree(in);
      }
#endif /*DEBUG_STATES*/
}

/*
 * simple_reaching - calculate union of reaching from successors,
 *                   do NOT cache path but simply ADD the path's lines
 * invariant: set members monotonically increasing
 */
void simple_reaching()
{
  int set;

  for (set = 0; set < numsets; set++)
    simple_reaching_body(set);
}

/*
 * simple_post_dominator_body - calculate intersection of reaching from successos,
 * (for lines mapping into block set)
 *                         do NOT cache path but simply ADD the path's lines
 * invariant: set members monotonically increasing
 */
void simple_post_dominator_body(set)
     int set;
{
  struct funclist *fptr;
  struct pathlist *pptr, *pptr2, *pptr3;
  struct callerlist *cptr, *cptr2;
  register int i, k, change, local_change, path_change;
  vector v, in;

  /*
   * propagate the cache information through all paths
   */
#ifdef DEBUG_STATES
  char str[MAXSTRING];
  int s, t;

  fprintf(stdout, "POST DOMINATING");
#endif
  do {
    change = FALSE;
    /* for each function */
    for (fptr = funcs; fptr; fptr = fptr->next) {
      /* for each function instance */
#ifdef DEBUG_STATES
      fprintf(stdout, "\nscanning func %s instance ", fptr->func->funcname);
#endif
#ifdef DEBUG2
      fprintf(stdout, "p"); fflush(stdout);
#endif
      for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next) {
	do {
	  local_change = FALSE;
	  
	  /* successors changed */
	  for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
	    if (!pptr->path->func) {
	      path_change = FALSE;
	      for (pptr2 = pptr->path->succs; pptr2; pptr2 = pptr2->next)
#ifndef TRASH
		{
#else
		if (pptr2->path->pout[k].changed) {
#endif
		  if (!path_change) {
		    v = vinit();
#ifdef TRASH
		    vcpy(&v, pptr->path->pout[k].v);
#else
		    vcpy(&v, vallptr);
#endif
		    path_change = TRUE;
		  }
		  in = vinit();
		  cache_rpath(pptr2->path, pptr2->path->pout[k].v, &in, set);
		  if (in)
		    vinter(&v, in);
		  else
		    vcpy(&v, in);
		  vfree(in);
		}
	      if (path_change)
		local_change |= change_path(&pptr->path->pout[k],
					    &pptr->path->pout[k].changed, &v);
	    }
	  
#ifdef DEBUG_STATES
	  if (local_change)
	    fprintf(stdout, "%d ", k);
#endif
	  if (local_change)
	    change = TRUE;
	} while (local_change);
	
	/*
	 * propagate ins of entrypaths to outs of caller
	 */
	if (fptr->func != mainfuncptr) {
	  i = cptr->instance;
	  path_change = FALSE;
	  for (pptr = fptr->func->entrypaths; pptr; pptr = pptr->next)
#ifndef TRASH
	    {
#else	      
	    if (pptr->path->pout[k].changed) {
#endif
	      if (!path_change) {
		v = vinit();
#ifdef TRASH
		vcpy(&v, cptr->caller->pout[i].v);
#else
		vcpy(&v, vallptr);
#endif
		path_change = TRUE;
	      }
	      in = vinit();
	      cache_rpath(pptr->path, pptr->path->pout[k].v, &in, set);
	      if (in)
		vinter(&v, in);
	      else
		vcpy(&v, in);
	      vfree(in);
	    }
	  if (path_change)
	    change |= change_path(&cptr->caller->pout[i],
				  &cptr->caller->pout[i].changed, &v);
#ifdef DEBUG_STATES
	  if (cptr->caller->pout[i].changed)
	    fprintf(stdout, "%du%d ", k, cptr->caller->pathnum);
#endif
	}
	
	/*
	 * propagate ins of successors of calls into
	 * exitpaths of instance of callee, reset changed flag
	 */
	for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
#ifndef TRASH
	  {
#else
	  if (pptr->path->pout[k].changed) {
#endif
	    pptr->path->pout[k].changed = FALSE;
	    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
	      if (pptr2->path->func)
		for (i = 0, cptr2 = pptr2->path->func->caller; cptr2;
		     i++, cptr2 = cptr2->next)
		  if (cptr2->succ_pathlabel == pptr->path->pathlabel
		      && cptr2->func == pptr->path->curfunc
		      && cptr2->instance == k)
		    for (pptr3 = pptr2->path->func->exitpaths; pptr3;
			 pptr3 = pptr3->next) {
		      v = vinit();
#ifdef TRASH
		      vcpy(&v, pptr3->path->pout[i].v);
#else
		      vcpy(&v, vallptr);
#endif
		      in = vinit();
		      cache_rpath(pptr->path, pptr->path->pout[k].v, &in, set);
		      if (in)
			vinter(&v, in);
		      else
			vcpy(&v, in);
		      vfree(in);
		      
		      change |= change_path(&pptr3->path->pout[i],
					    &pptr3->path->pout[i].changed, &v);
#ifdef DEBUG_STATES
		      if (pptr3->path->pout[i].changed)
			fprintf(stdout, "%dd%d ", k, pptr->path->pathnum);
#endif /*DEBUG_STATES*/
		    }
	  }
      }
    }
  } while (change);
#ifdef DEBUG_STATES
  printf("\n");

  fprintf(stdout, "input/output states:\n");
  if (verbose) {
    fprintf(stdout, "%22s ", "cache line");
    for (s = 0; s < 2; s++) {
      for (i = 0; i < vecbits; i++) {
	for (t = i; t > 10; t %= 10)
	  fprintf(stdout, " ");
	fprintf(stdout, "%d ", i % numsets);
      }
      fprintf(stdout, "      ");
    }
    fprintf(stdout, "\n");

    fprintf(stdout, "%22s ", "program line");
    for (s = 0; s < 2; s++) {
      for (i = 0; i < vecbits; i++)
	if (i < numsets)
	  fprintf(stdout, "I ");
	else
	  fprintf(stdout, "%d ", i - numsets);
      fprintf(stdout, "      ");
    }
    fprintf(stdout, "\n");
  }
	
  for (fptr = funcs; fptr; fptr = fptr->next)
    for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	in = vinit();
	cache_rpath(pptr->path, pptr->path->pout[k].v, &in, set);
	vget(in, str);
	fprintf(stdout, "%10s.%2d[%2d].",
		fptr->func->funcname, pptr->path->pathnum, k);
	fprintf(stdout, "in =[%s]", str);
	if (!verbose)
	  fprintf(stdout, "\n");
	vget(pptr->path->pout[k].v, str);
	if (!verbose)
	  fprintf(stdout, "%10s.%2d[%2d].",
		  fptr->func->funcname, pptr->path->pathnum, k);
	fprintf(stdout, "out=[%s]", str);
	fprintf(stdout, "\n");
	vfree(in);
      }
#endif /*DEBUG_STATES*/
}

/*
 * simple_post_dominator - calculate intersection of reaching from successos,
 *                         do NOT cache path but simply ADD the path's lines
 * invariant: set members monotonically increasing
 */
void simple_post_dominator()
{
  int set;

  for (set = 0; set < numsets; set++)
    simple_post_dominator_body(set);
}

/*
 * simple_linear_body - calculate union of ins from predecessors, excl. backedges
 * (for lines mapping into block set)
 * invariant: set members monotonically increasing
 */
void simple_linear_body(set)
     int set;
{
  struct funclist *fptr;
  struct pathlist *pptr, *pptr2;
  struct callerlist *cptr, *cptr2;
  register int i, j, k, s, t, change, local_change;
  vector v, out, diff[MAXPREDS][MAXSETSIZE], prev_diff;

  /*
   * propagate the cache information through all paths
   */
#ifdef DEBUG_STATES
  char str[MAXSTRING];
  vector dummy[MAXSETSIZE];

  fprintf(stdout, "LINEAR STATES");
#endif
  do {
    change = FALSE;
    /* for each function */
    for (fptr = funcs; fptr; fptr = fptr->next) {
      /* for each function instance */
#ifdef DEBUG_STATES
      fprintf(stdout, "\nscanning func %s instance ", fptr->func->funcname);
#endif
#ifdef DEBUG2
      fprintf(stdout, "l"); fflush(stdout);
#endif
      for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next) {
	do {
	  local_change = FALSE;

	  /* predecessors changed, no calls */
	  for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
	      if (pptr2->path->func)
		break;
	    if (pptr2)
	      continue;
	    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
	      if (
#ifdef TRASH
		  pptr2->path->lin[k][0].changed &&
#endif
		  !is_backedge(pptr->path, pptr2->path)
		  )
		break;
	    if (pptr2) {
	      for (j = 0; j < pptr->path->numpreds; j++)
		init_set(diff[j]);
	      for (s = 0; s < setsize; s++) {
		v = vinit();
		vcpy(&v, pptr->path->lin[k][s].v);
		for (j = 0, pptr2 = pptr->path->preds; pptr2;
		     j++, pptr2 = pptr2->next) {
		  if (!is_backedge(pptr->path, pptr2->path)) {
		    out = vinit();
		    prev_diff = (s > 0 ? diff[j][s-1] : vinit());
		    vunion(&v, prev_diff);
		    cache_path(pptr2->path, pptr2->path->lin[k][s].v,
			       &out, diff[j], s, prev_diff, set);
		    vunion(&v, out);
		    vfree(out);
		  }
		}
		local_change |= change_path(&pptr->path->lin[k][s],
					    &pptr->path->lin[k][0].changed, &v);
	      }
	      for (j = 0; j < pptr->path->numpreds; j++)
		free_set(diff[j]);
	    }
	  }
#ifdef DEBUG_STATES
	  if (local_change)
	    fprintf(stdout, "%d ", k);
#endif
	  if (local_change)
	    change = TRUE;
	} while (local_change);

	/*
	 * if the sibling of a caller or an exit path of the callee changed
	 * then propagate outs from union of exit paths and
	 * siblings of caller (predecessors of caller's successors w/o calls)
	 * to ins of caller's successors
	 */
	if (fptr->func != mainfuncptr) {
	  i = cptr->instance;
	  for (pptr2 = cptr->succs; pptr2; pptr2 = pptr2->next) {
#ifdef TRASH
	    for (pptr = fptr->func->exitpaths; pptr; pptr = pptr->next)
	      if (pptr->path->lin[k][0].changed)
		break;
	    if (!pptr)
	      for (pptr = pptr2->path->preds; pptr; pptr = pptr->next)
		if (pptr->path->lin[i][0].changed)
		  break;
	    if (pptr) {
#else
	    {
#endif
	      for (j = 0; j < pptr2->path->numpreds + fptr->func->numexitpaths;
		   j++)
		init_set(diff[j]);
	      for (s = 0; s < setsize; s++) {
		v = vinit();
		vcpy(&v, pptr2->path->lin[i][s].v);
		for (j = pptr2->path->numpreds, pptr = fptr->func->exitpaths;
		     pptr; j++, pptr = pptr->next) {
		  out = vinit();
		  prev_diff = (s > 0 ? diff[j][s-1] : vinit());
		  vunion(&v, prev_diff);
		  cache_path(pptr->path, pptr->path->lin[k][s].v,
			     &out, diff[j], s, prev_diff, set);
		  vunion(&v, out);
		  vfree(out);
		}
		for (j = 0, pptr = pptr2->path->preds; pptr;
		     j++, pptr = pptr->next)
		  if (!pptr->path->func
		      && !is_backedge(pptr2->path, pptr->path)
		      ) {
		    out = vinit();
		    prev_diff = (s > 0 ? diff[j][s-1] : vinit());
		    vunion(&v, prev_diff);
		    cache_path(pptr->path, pptr->path->lin[i][s].v,
			       &out, diff[j], s, prev_diff, set);
		    vunion(&v, out);
		    vfree(out);
		  }
		change |= change_path(&pptr2->path->lin[i][s],
				      &pptr2->path->lin[i][0].changed, &v);
#ifdef DEBUG_STATES
		if (s == 0 && pptr2->path->lin[i][0].changed)
		  fprintf(stdout, "%du%d ", k, pptr2->path->pathnum);
#endif
	      }
	      for (j = 0; j < pptr2->path->numpreds + fptr->func->numexitpaths;
		   j++)
		free_set(diff[j]);
	    }
	  }
	}
	
	/*
	 * propagate outs at calls into instance of callee, reset changed flag
	 */
	for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
#ifndef TRASH
	  {
#else
	  if (pptr->path->lin[k][0].changed) {
#endif
	    pptr->path->lin[k][0].changed = FALSE;
	    if (pptr->path->func)
	      for (i = 0, cptr2 = pptr->path->func->caller; cptr2;
		   i++, cptr2 = cptr2->next)
		if (cptr2->succ_pathlabel == pptr->path->succs->path->pathlabel
		    && cptr2->func == pptr->path->curfunc
		    && cptr2->instance == k) {
		  for (pptr2 = pptr->path->func->entrypaths; pptr2;
		       pptr2 = pptr2->next) {
		    init_set(diff[0]);
		    for (s = 0; s < setsize; s++) {
		      v = vinit();
		      vcpy(&v, pptr2->path->lin[i][s].v);
		      prev_diff = (s > 0 ? diff[0][s-1] : vinit());
		      vunion(&v, prev_diff);
		      out = vinit();
		      cache_path(pptr->path, pptr->path->lin[k][s].v,
				 &out, diff[0], s, prev_diff, set);
		      vunion(&v, out);
		      vfree(out);

		      change |= change_path(&pptr2->path->lin[i][s],
					    &pptr2->path->lin[i][0].changed, &v);
#ifdef DEBUG_STATES
		      if (s == 0 && pptr2->path->lin[i][0].changed)
			fprintf(stdout, "%dd%d ", k, pptr->path->pathnum);
#endif /*DEBUG_STATES*/
		    }
		    free_set(diff[0]);
		  }
		}
	  }
      }
    }
  } while (change);
#ifdef DEBUG_STATES
  printf("\n");

  fprintf(stdout, "input/output states:\n");
  if (verbose) {
    fprintf(stdout, "%26s ", "cache line");
    for (s = 0; s < 2 * setsize; s++) {
      for (i = 0; i < vecbits; i++) {
	for (t = i; t > 10; t %= 10)
	  fprintf(stdout, " ");
	fprintf(stdout, "%d ", i % numsets);
      }
      fprintf(stdout, "      ");
    }
    fprintf(stdout, "\n");

    fprintf(stdout, "%26s ", "program line");
    for (s = 0; s < 2 * setsize; s++) {
      for (i = 0; i < vecbits; i++)
	if (i < numsets)
	  fprintf(stdout, "I ");
	else
	  fprintf(stdout, "%d ", i - numsets);
      fprintf(stdout, "      ");
    }
    fprintf(stdout, "\n");
  }
	
  for (fptr = funcs; fptr; fptr = fptr->next)
    for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	v = vinit();
	init_set(diff[0]);
	for (s = 0; s < setsize; s++) {
	  out = vinit();
	  cache_path(pptr->path, pptr->path->lin[k][s].v,
		     &out, diff[0], s, v, set);
	  vunion(&v, out);
	  vget(pptr->path->lin[k][s].v, str);
	  if (!verbose || !s)
	    fprintf(stdout, "%10s.%2d[%2d][%2d].",
		    fptr->func->funcname, pptr->path->pathnum, k, s);
	  fprintf(stdout, "in =[%s]", str);
	  if (!verbose)
	    fprintf(stdout, "\n");
	  vget(v, str);
	  if (!verbose)
	    fprintf(stdout, "%10s.%2d[%2d][%2d].",
		    fptr->func->funcname, pptr->path->pathnum, k, s);
	  fprintf(stdout, "out=[%s]", str);
	  if (!verbose || s + 1 == setsize)
	    fprintf(stdout, "\n");
	  vfree(out);
	  if (s + 1 < setsize)
	    v = diff[0][s];
	}
	free_set(diff[0]);
      }
#endif /*DEBUG_STATES*/

  /*
   * union sets together, free sets > 0
   */
  for (fptr = funcs; fptr; fptr = fptr->next)
    for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	v = vinit();
	for (s = 0; s < setsize; s++) {
	  vunion(&v, pptr->path->lin[k][s].v);
	  vfree(pptr->path->lin[k][s].v);
	  pptr->path->lin[k][s].v = vinit();
	}
	pptr->path->lin[k][0].v = v;
      }

#ifdef DEBUG_STATES
  printf("\n");

  fprintf(stdout, "input/output states:\n");
  if (verbose) {
    fprintf(stdout, "%22s ", "cache line");
    for (s = 0; s < 2; s++) {
      for (i = 0; i < vecbits; i++) {
	for (t = i; t > 10; t %= 10)
	  fprintf(stdout, " ");
	fprintf(stdout, "%d ", i % numsets);
      }
      fprintf(stdout, "      ");
    }
    fprintf(stdout, "\n");

    fprintf(stdout, "%22s ", "program line");
    for (s = 0; s < 2; s++) {
      for (i = 0; i < vecbits; i++)
	if (i < numsets)
	  fprintf(stdout, "I ");
	else
	  fprintf(stdout, "%d ", i - numsets);
      fprintf(stdout, "      ");
    }
    fprintf(stdout, "\n");
  }
	
  for (fptr = funcs; fptr; fptr = fptr->next)
    for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	out = vinit();
	init_set(dummy);
	cache_path(pptr->path, pptr->path->lin[k][0].v,
		   &out, dummy, 0, vinit(), set);
	vget(pptr->path->lin[k][0].v, str);
	fprintf(stdout, "%10s.%2d[%2d].",
		fptr->func->funcname, pptr->path->pathnum, k);
	fprintf(stdout, "in =[%s]", str);
	if (!verbose)
	  fprintf(stdout, "\n");
	vget(out, str);
	if (!verbose)
	  fprintf(stdout, "%10s.%2d[%2d].",
		  fptr->func->funcname, pptr->path->pathnum, k);
	fprintf(stdout, "out=[%s]", str);
	fprintf(stdout, "\n");
	vfree(out);
	free_set(dummy);
      }
#endif /*DEBUG_STATES*/
}

/*
 * simple_linear - calculate union of ins from predecessors, excl. backedges
 * invariant: set members monotonically increasing
 */
void simple_linear()
{
  int set;

  for (set = 0; set < numsets; set++)
    simple_linear_body(set);
}

/*
 * simple_dominator_body - calculate intersection of reaching from predecessors,
 * (for lines mapping into block set)
 *                         do NOT cache path but simply ADD the path's lines
 * invariant: set members monotonically increasing
 */
void simple_dominator_body(set)
     int set;
{
  struct funclist *fptr;
  struct pathlist *pptr, *pptr2, *pptr3;
  struct callerlist *cptr, *cptr2;
  register int i, j, k, change, local_change, path_change;
  vector v, out;

  /*
   * propagate the cache information through all paths
   */
#ifdef DEBUG_STATES
  char str[MAXSTRING];
  int t;

  fprintf(stdout, "DOMINATOR");
#endif
  do {
    change = FALSE;
    /* for each function */
    for (fptr = funcs; fptr; fptr = fptr->next) {
      /* for each function instance */
#ifdef DEBUG_STATES
      fprintf(stdout, "\nscanning func %s instance ", fptr->func->funcname);
#endif
#ifdef DEBUG2
      fprintf(stdout, "d"); fflush(stdout);
#endif
      for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next) {
	do {
	  local_change = FALSE;
	  
	  /* predecessors changed, no calls */
	  for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
	      if (pptr2->path->func)
		break;
	    if (pptr2)
	      continue;
#ifdef TRASH
	    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
	      if (pptr2->path->din[k].changed)
		break;
	    if (pptr2) {
#else
	    {
#endif
	      v = vinit();
#ifndef TRASH
	      vcpy(&v, vallptr);
#endif 
	      path_change = FALSE;
	      for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
#ifndef TRASH
		{
#else
		if (pptr2->path->din[k].changed || pptr2->path->din[k].v) {
#endif
		  path_change = TRUE;
		  out = vinit();
		  cache_rpath(pptr2->path, pptr2->path->din[k].v, &out, set);
		  if (v)
		    vinter(&v, out);
		  else
		    vcpy(&v, out);
		  vfree(out);
		}
	      if (path_change)
		local_change |= change_path(&pptr->path->din[k],
					    &pptr->path->din[k].changed, &v);
	    }
	  }
	  
#ifdef DEBUG_STATES
	  if (local_change)
	    fprintf(stdout, "%d ", k);
#endif
	  if (local_change)
	    change = TRUE;
	} while (local_change);
	
	/*
	 * if the sibling of a caller or an exit path of the callee changed
	 * then propagate outs from intersection of exit paths and
	 * siblings of caller (predecessors of caller's successors w/o calls)
	 * to ins of caller's successors
	 */
	if (fptr->func != mainfuncptr) {
	  i = cptr->instance;
	  for (pptr2 = cptr->succs; pptr2; pptr2 = pptr2->next) {
#ifdef TRASH
	    for (pptr = fptr->func->exitpaths; pptr; pptr = pptr->next)
	      if (pptr->path->din[k].changed)
		break;
	    if (!pptr)
	      for (pptr = pptr2->path->preds; pptr; pptr = pptr->next)
		if (pptr->path->din[i].changed)
		  break;
	    if (pptr) {	
#else
	    {
#endif
	      v = vinit();
#ifndef TRASH
	      vcpy(&v, vallptr);
#endif 
	      for (j = pptr2->path->numpreds, pptr = fptr->func->exitpaths;
		   pptr; j++, pptr = pptr->next)
#ifndef TRASH
		{
#else
		if (pptr->path->din[i].changed || pptr->path->din[k].v) {
#endif
		  out = vinit();
		  cache_rpath(pptr->path, pptr->path->din[k].v, &out, set);
		  if (v)
		    vinter(&v, out);
		  else
		    vcpy(&v, out);
		  vfree(out);
		}
	      for (j = 0, pptr = pptr2->path->preds; pptr;
		   j++, pptr = pptr->next)
		if (!pptr->path->func
#ifndef TRASH
		    ) {
#else
		    && (pptr->path->din[i].changed || pptr->path->din[i].v)) {
#endif
		  out = vinit();
		  cache_rpath(pptr->path, pptr->path->din[i].v, &out, set);
		  if (v)
		    vinter(&v, out);
		  else
		    vcpy(&v, out);
		  vfree(out);
		}
	      change |= change_path(&pptr2->path->din[i],
				    &pptr2->path->din[i].changed, &v);
#ifdef DEBUG_STATES
	      if (pptr2->path->din[i].changed)
		fprintf(stdout, "%du%d ", k, pptr2->path->pathnum);
#endif
	    }
	  }
	}
		
	/*
	 * propagate outs at calls into instance of callee, reset changed flag
	 */
	for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
#ifndef TRASH
	  {
#else
	  if (pptr->path->din[k].changed) {
#endif
	    pptr->path->din[k].changed = FALSE;
	    if (pptr->path->func)
	      for (i = 0, cptr2 = pptr->path->func->caller; cptr2;
		   i++, cptr2 = cptr2->next)
		if (cptr2->succ_pathlabel == pptr->path->succs->path->pathlabel
		    && cptr2->func == pptr->path->curfunc
		    && cptr2->instance == k) {
		  for (pptr2 = pptr->path->func->entrypaths; pptr2;
		       pptr2 = pptr2->next) {
		    v = vinit();
		    cache_rpath(pptr->path, pptr->path->din[k].v, &v, set);
		    change |= change_path(&pptr2->path->din[i],
					  &pptr2->path->din[i].changed, &v);
#ifdef DEBUG_STATES
		    if (pptr2->path->din[i].changed)
		      fprintf(stdout, "%dd%d ", k, pptr->path->pathnum);
#endif /*DEBUG_STATES*/
		  }
		}
	  }
      }
    }
  } while (change);
#ifdef DEBUG_STATES
  printf("\n");

  fprintf(stdout, "input/output states:\n");
  if (verbose) {
    fprintf(stdout, "%22s ", "cache line");
    for (i = 0; i < vecbits; i++) {
      for (t = i; t > 10; t %= 10)
	fprintf(stdout, " ");
      fprintf(stdout, "%d ", i % numsets);
    }
    fprintf(stdout, "\n");

    fprintf(stdout, "%22s ", "program line");
    for (i = 0; i < vecbits; i++)
      if (i < numsets)
	fprintf(stdout, "I ");
      else
	fprintf(stdout, "%d ", i - numsets);
    fprintf(stdout, "\n");
  }
	
  for (fptr = funcs; fptr; fptr = fptr->next)
    for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	out = vinit();
	cache_rpath(pptr->path, pptr->path->din[k].v, &out, set);
	vget(pptr->path->din[k].v, str);
	fprintf(stdout, "%10s.%2d[%2d].",
		fptr->func->funcname, pptr->path->pathnum, k);
	fprintf(stdout, "in =[%s]", str);
	if (!verbose)
	  fprintf(stdout, "\n");
	vget(out, str);
	if (!verbose)
	  fprintf(stdout, "%10s.%2d[%2d].",
		  fptr->func->funcname, pptr->path->pathnum, k);
	fprintf(stdout, "out =[%s]", str);
	fprintf(stdout, "\n");
	vfree(out);
      }
#endif /*DEBUG_STATES*/
}

/*
 * simple_dominator - calculate intersection of reaching from predecessors,
 *                         do NOT cache path but simply ADD the path's lines
 * invariant: set members monotonically increasing
 */
void simple_dominator()
{
  int set;

  for (set = 0; set < numsets; set++)
    simple_dominator_body(set);
}

/*
 * optimize_vec - free duplicate bit vecvtors in for paths in func
 */
void optimize_vec(func, num)
struct funcinfo *func;   /* for this function */
int num;                 /* allocate so many new instances */
{
  struct pathlist *pptr, *pptr2, *pptr3;
  int i, s, t, u;
#ifdef DEBUG_STATES
  int savings = 0;
#endif

  for (pptr = func->paths; pptr; pptr = pptr->next) {
    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
      if (pptr2->path->curfunc == pptr->path->curfunc)
	for (i = 0; i < num; i++)
	  for (s = 0; s < setsize; s++)
	    for (t = 0; t < setsize; t++)
	      if (pptr->path->in[i][s].v != pptr2->path->in[i][t].v &&
		  vcmp(pptr->path->in[i][s].v, pptr2->path->in[i][t].v)) {
#ifdef DEBUG_STATES
		savings++;
#endif
		for (pptr3 = func->paths; pptr3; pptr3 = pptr3->next)
		  if (pptr->path != pptr3->path)
		    for (u = 0; u < setsize; u++)
		      if (pptr->path->in[i][s].v == pptr3->path->in[i][u].v) {
			pptr3->path->in[i][u].v = pptr2->path->in[i][t].v;
#ifdef DEBUG_STATES
			savings++;
#endif
		      }
		FREE(pptr->path->in[i][s].v);
		pptr->path->in[i][s].v = pptr2->path->in[i][t].v;
	      }
  }
#ifdef DEBUG_STATES
  printf("vec savings in %s: %d\n", func->funcname, savings);
#endif
}

/*
 * optimize_rvec - free duplicate bit vecvtors rout for paths in func
 */
void optimize_rvec(func, num)
struct funcinfo *func;   /* for this function */
int num;                 /* allocate so many new instances */
{
  struct pathlist *pptr, *pptr2, *pptr3;
  int i;
#ifdef DEBUG_STATES
  int savings = 0;
#endif

  for (pptr = func->paths; pptr; pptr = pptr->next) {
    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
      if (pptr2->path->curfunc == pptr->path->curfunc)
	for (i = 0; i < num; i++)
	  if (pptr->path->rout[i].v != pptr2->path->rout[i].v &&
	      vcmp(pptr->path->rout[i].v, pptr2->path->rout[i].v)) {
#ifdef DEBUG_STATES
	    savings++;
#endif
	    for (pptr3 = func->paths; pptr3; pptr3 = pptr3->next)
	      if (pptr->path != pptr3->path &&
		  pptr->path->rout[i].v == pptr3->path->rout[i].v) {
		pptr3->path->rout[i].v = pptr2->path->rout[i].v;
#ifdef DEBUG_STATES
		savings++;
#endif
	      }
	    FREE(pptr->path->rout[i].v);
	    pptr->path->rout[i].v = pptr2->path->rout[i].v;
	  }
  }
#ifdef DEBUG_STATES
  printf("rvec savings in %s: %d\n", func->funcname, savings);
#endif
}

/*
 * optimize_pvec - free duplicate bit vecvtors pout for paths in func
 */
void optimize_pvec(func, num)
struct funcinfo *func;   /* for this function */
int num;                 /* allocate so many new instances */
{
  struct pathlist *pptr, *pptr2, *pptr3;
  int i;
#ifdef DEBUG_STATES
  int savings = 0;
#endif

  for (pptr = func->paths; pptr; pptr = pptr->next) {
    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
      if (pptr2->path->curfunc == pptr->path->curfunc)
	for (i = 0; i < num; i++)
	  if (pptr->path->pout[i].v != pptr2->path->pout[i].v &&
	      vcmp(pptr->path->pout[i].v, pptr2->path->pout[i].v)) {
#ifdef DEBUG_STATES
	    savings++;
#endif
	    for (pptr3 = func->paths; pptr3; pptr3 = pptr3->next)
	      if (pptr->path != pptr3->path &&
		  pptr->path->pout[i].v == pptr3->path->pout[i].v) {
		pptr3->path->pout[i].v = pptr2->path->pout[i].v;
#ifdef DEBUG_STATES
		savings++;
#endif
	      }
	    FREE(pptr->path->pout[i].v);
	    pptr->path->pout[i].v = pptr2->path->pout[i].v;
	  }
  }
#ifdef DEBUG_STATES
  printf("pvec savings in %s: %d\n", func->funcname, savings);
#endif
}

/*
 * optimize_lvec - free duplicate bit vecvtors lin for paths in func
 */
void optimize_lvec(func, num)
struct funcinfo *func;   /* for this function */
int num;                 /* allocate so many new instances */
{
  struct pathlist *pptr, *pptr2, *pptr3;
  int i;
#ifdef DEBUG_STATES
  int savings = 0;
#endif

  for (pptr = func->paths; pptr; pptr = pptr->next) {
    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
      if (pptr2->path->curfunc == pptr->path->curfunc)
	for (i = 0; i < num; i++)
	  if (pptr->path->lin[i][0].v != pptr2->path->lin[i][0].v &&
	      vcmp(pptr->path->lin[i][0].v, pptr2->path->lin[i][0].v)) {
#ifdef DEBUG_STATES
	    savings++;
#endif
	    for (pptr3 = func->paths; pptr3; pptr3 = pptr3->next)
	      if (pptr->path != pptr3->path &&
		  pptr->path->lin[i][0].v == pptr3->path->lin[i][0].v) {
		pptr3->path->lin[i][0].v = pptr2->path->lin[i][0].v;
#ifdef DEBUG_STATES
		savings++;
#endif
	      }
	    FREE(pptr->path->lin[i][0].v);
	    pptr->path->lin[i][0].v = pptr2->path->lin[i][0].v;
	  }
  }
#ifdef DEBUG_STATES
  printf("lvec savings in %s: %d\n", func->funcname, savings);
#endif
}

/*
 * optimize_dvec - free duplicate bit vecvtors din for paths in func
 */
void optimize_dvec(func, num)
struct funcinfo *func;   /* for this function */
int num;                 /* allocate so many new instances */
{
  struct pathlist *pptr, *pptr2, *pptr3;
  int i;
#ifdef DEBUG_STATES
  int savings = 0;
#endif

  for (pptr = func->paths; pptr; pptr = pptr->next) {
    for (pptr2 = pptr->path->preds; pptr2; pptr2 = pptr2->next)
      if (pptr2->path->curfunc == pptr->path->curfunc)
	for (i = 0; i < num; i++)
	  if (pptr->path->din[i].v != pptr2->path->din[i].v &&
	      vcmp(pptr->path->din[i].v, pptr2->path->din[i].v)) {
#ifdef DEBUG_STATES
	    savings++;
#endif
	    for (pptr3 = func->paths; pptr3; pptr3 = pptr3->next)
	      if (pptr->path != pptr3->path &&
		  pptr->path->din[i].v == pptr3->path->din[i].v) {
		pptr3->path->din[i].v = pptr2->path->din[i].v;
#ifdef DEBUG_STATES
		savings++;
#endif
	      }
	    FREE(pptr->path->din[i].v);
	    pptr->path->din[i].v = pptr2->path->din[i].v;
	  }
  }
#ifdef DEBUG_STATES
  printf("rvec savings in %s: %d\n", func->funcname, savings);
#endif
}

