/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: fig.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Function instance graph
 */

#include "scs.h"

/*
 * allocate_caller - record path which calls function and instance offset
 */
int allocate_caller(func, caller, instance)
struct funcinfo *func;   /* for this function */
struct pathinfo *caller; /* called by this path */
int instance;            /* instance of the caller */
{
  register struct callerlist *cptr, *cptr2;
  register int ret;

  /*
   * check if already in list
   */
  for (ret = 0, cptr = func->caller; cptr; ret++, cptr = cptr->next)
    if (cptr->succ_pathlabel == caller->succs->path->pathlabel &&
	cptr->func == caller->curfunc &&
	cptr->instance == instance)
      return(ret);

  /*
   * make a new entry
   */
  cptr = (struct callerlist *) MALLOC(sizeof(struct callerlist));
  cptr->caller = caller;
  cptr->succs = caller ? caller->succs : (struct pathlist *) NULL;
  cptr->succ_pathlabel = caller ? caller->succs->path->pathlabel : 0;
  cptr->func = caller ? caller->curfunc : (struct funcinfo *) NULL;
  cptr->instance = instance;
  /*
   * append to caller list
   */
  cptr->next = (struct callerlist *) NULL;
  for (ret = 0, cptr2 = func->caller;
       cptr2 && cptr2->next; 
       ret++, cptr2 = cptr2->next);
  if (cptr2) {
    ret++;
    cptr2->next = cptr;
  }
  else
    func->caller = cptr;

  return(ret);
}

/*
 * allocate_caller_dfs - call graph traversed depth-first order, record callers
 */
void allocate_caller_dfs(pptr, instance)
struct pathlist *pptr;
int instance;
{
  for (; pptr; pptr = pptr->next)
    if (pptr->path->func) {
      pptr->path->instance = allocate_caller(pptr->path->func,
					     pptr->path,
					     instance);
      allocate_caller_dfs(pptr->path->func->paths, pptr->path->instance);
    }
}

/*
 * buildcallgraph - connect the functions together according to calls
 */
void buildcallgraph()
{
  register int change, change_set, i;
  register struct funclist *fptr, *fptr2, *nptr;
  register struct pathlist *pptr;
#ifdef DEBUG2
  vector v;
  char s[MAXSTRING];
#endif
  
  /*
   * initialize vector length
   */
  vecbits = numfuncs;
  veclen = (vecbits >> intscale) + 1;

  /* For each function in the program. */
   for (fptr = funcs; fptr; fptr = fptr->next) {

     /*
      * initialize reaching vector
      */
#ifdef DEBUG2
     v = vinit();
#endif

     /* If is the main function, then save the ptr. */
     if (strcmp(fptr->func->funcname, "_main") == 0)
       mainfuncptr = fptr->func;
     
     /* For each path in the function. */
     for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
       
       /* If the path ends with a call, then find the function within the
	  program that is being called. */
       if (pptr->path->succtype == CALL)
	 for (fptr2 = funcs; fptr2; fptr2 = fptr2->next)
	   if (strcmp(fptr2->func->funcname, pptr->path->funcname) == 0) {
	     pptr->path->func = fptr2->func;
	     FREE(pptr->path->funcname);
	     pptr->path->funcname = (char *) NULL;
	     
	     /*
	      * add direct calls to reaching info across functions
	      */
#ifdef DEBUG2
	     if (!vin(v, fptr2->func->funcnum))
	       vinsert(&v, fptr2->func->funcnum);
#endif

	     break;
	   }
#ifdef DEBUG2
     fptr->func->freach = v;
#endif
   }

#ifdef DEBUG2
  /*
   * set up function table
   */
  functab = (struct funcinfo **) MALLOC(sizeof(struct functinfo *) *
					(numfuncs+1));
  for (fptr = funcs; fptr; fptr = fptr->next)
    functab[fptr->func->funcnum] = fptr->func;

  /*
   * closure over reachable functions
   */
  do {
    change = FALSE;
    for (fptr = funcs; fptr; fptr = fptr->next) {
      change_set = FALSE;
      v = fptr->func->freach;
      for (i = 1; i < numfuncs; i++)
	if (vin(v, i) && vuniond(&v, functab[i]->freach, vinit()))
	  change = change_set = TRUE;
      if (change_set)
	fptr->func->freach = v;
    }
  } while (change);

#ifdef DEBUG
  /*
   * show function reach info
   */
  for (fptr = funcs; fptr; fptr = fptr->next) {
    vget(fptr->func->freach, s);
    printf("func %s %d reaches %s\n",
	   fptr->func->funcname, fptr->func->funcnum, s);
    if (vin(fptr->func->freach, fptr->func->funcnum)) {
      fprintf(stderr, "ERROR: recursive call in %s\n", fptr->func->funcname);
      exit(-1);
    }
  }
  fflush(stdout);
#endif /*DEBUG*/
#endif /*DEBGU2*/

  /* If main function not found, then cannot do the static simulation. */
  if (!mainfuncptr)
    error("main func not found.");
}

