/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: instrument.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Code instrumentation
 */

#include "scs.h"

#define LOAD TRUE
#define STORE FALSE

#ifdef PROFILING
/*
 * dump_ldst - dump a load/store instruction with the fast instruction format
 * (1 instruction) when the offsets do not exceed 4k, otherwise use a slow
 * format (4 instructions for load, one for store).
 */
void dump_ldst(load, ind_reg, off1, off2, reg, offset)
int load, offset;
char *reg, *off1, *off2, *ind_reg;
{
  if (offset < MAX_OFFSET) {
    if (load)
      fprintf(fmacros, "\tld\t[%%%s+%s-%s],%%%s ; \\\n",
	      ind_reg, off1, off2, reg);
    else
      fprintf(fmacros, "\tst\t%%%s,[%%%s+%s-%s] ; \\\n",
	      reg, ind_reg, off1, off2);
    return;
  }

  if (load) {
    fprintf(fmacros, "\tset\t%s-%s,%%%s ; \\\n", off1, off2, reg);
    fprintf(fmacros, "\tadd\t%%%s,%%%s,%%%s ; \\\n", ind_reg, reg, "temp1");
    fprintf(fmacros, "\tld\t[%%%s],%%%s ; \\\n", "temp1", reg);
  }
  else
    fprintf(fmacros, "\tst\t%%%s,[%%%s] ; \\\n", reg, "temp1");
}

/*
 * reload_offset - dump a reload instruction corresponding to the ld/st
 * instruction to reload a lost value (2 instructions).
 */
void reload_offset(reg, ind_reg, s1, s2, offset)
char *reg, *ind_reg, *s1, *s2;
int offset;
{
  if (offset >= MAX_OFFSET) {
    fprintf(fmacros, "\tset\t%s-%s,%%%s ; \\\n", s1, s2, reg);
    fprintf(fmacros, "\tadd\t%%%s,%%%s,%%%s ; \\\n", ind_reg, reg, reg);
  }
}

/*
 * dump_logic - dump a logical instruction with the fast instruction format
 * (1 instruction) when the value do not exceed 4k, otherwise use a slow
 * format (3 instructions + 2 instructions for reload).
 */
void dump_logic(inst, reg, temp, value)
char *inst, *reg, *temp;
unsigned int value;
{
  if (value < MAX_VALUE)
    fprintf(fmacros, "\t%s\t%%%s,%d,%%%s ; \\\n", inst, reg, value, reg);
  else {
    fprintf(fmacros, "\tset\t%u,%%%s ; \\\n", value, temp);
    fprintf(fmacros, "\t%s\t%%%s,%%%s,%%%s ; \\\n", inst, reg, temp, reg);
  }
}

/*
 * reload_logic - dump a reload instruction corresponding to the logic
 * instruction to reload a lost value (2 instructions).
 */
void reload_logic(reg, label, value)
char *reg, *label;
unsigned int value;
{
  if (value >= MAX_VALUE)
    fprintf(fmacros, "\tset\t%s,%%%s ; \\\n", label, reg);
}

/*
 * cache_self - check if a path is only partially cached after execution
 * (happens when |cache| < |path|).
 * This check only applies for conflicting lines.
 */
int cache_self(iptr, v, cfptr, k)
struct instinfo *iptr;
vector *v;
struct fmline_t *cfptr;
int *k;
{
  register int i, l, m;
  register struct fmline_t *cfptr2;

  if (!iptr)
    return(0);

  m = cache_self(iptr->next, v, cfptr, k);

  for (l = iptr->lastline; l >= iptr->firstline; l--) {
    if (!vin(*v, l % numsets)) {
      vinsert(v, l % numsets);
      vinsert(v, l + numsets);
    }
    else
      for (i = 0, cfptr2 = cfptr; cfptr2; i++, cfptr2 = cfptr2->next)
	if (l + numsets == cfptr2->line && !vin(*v, l + numsets) &&
	    !(m & ((1 << i) * sizeof(int)))) {
	  (*k)++;
	  m |= (1 << i) * sizeof(int);
	}
  }

  return(m);
}

/*
 * inpathinsts - test if a line is in the list of instructions of a path
 */
int inpathinsts(iptr, line)
     struct instinfo *iptr;
{
  for (; iptr; iptr = iptr->next)
    if (iptr->firstline <= line && line <= iptr->lastline)
      return(TRUE);
  return(FALSE);
}

/*
 * simple_states - determine states,
 * emit code for state transition and counting (MACRO),
 * emit code for conflicting states (MACRO),
 * allocate space for states,
 * determine # hits and # misses for each state,
 * generate path table
 */
void simple_states()
{
  struct funclist *fptr;
  struct pathlist *pptr, *pptr2;
  struct pathlistlist *lptr;
  struct instinfo *iptr;
  struct callerlist *cptr, *cptr2;
  struct fmline_t *fmptr;
  struct conf_t *lconf, *lconf2;
  char *fn, *last_fn, *reg;
  char last_inst[MAXSTRING];
  char base_out[MAXSTRING];
  char s1[MAXSTRING], s2[MAXSTRING];
  int i, k, l, m, max, states, offset;
  int prev_line, last_conflict, last_k, temp1_set, clear_pending, uncached;
  struct linechange_t *lchange,*lchange2;
  vector v, dummy[MAXSETSIZE];
  struct fmline_t *cfptr, *cfptr2;

  extern void add_pathlist();
  extern void free_pathlist();
  extern char *strcpy();
  extern struct linechange_t *addlchange();

  init_set(dummy);
  /*
   * dump out assembly prologue for path table
   */
  fprintf(ftables, "\t.seg\t\"data\"\n");
  fprintf(ftables, "\t.align\t4\n");
  fprintf(ftables, "\t.global\t___profile\n");
  fprintf(ftables, "___profile:\t.word\t%d\n", profile);
#ifdef MERGE_STATES
  fprintf(ftables, "\t.global\t___state_table\n");
  fprintf(ftables, "___state_table:\n");
#else /* ifndef MERGE_STATES */
  fprintf(ftables, "\t.global\t___path_table\n");
  fprintf(ftables, "___path_table:\n");
#endif /*MERGE_STATES*/

  /*
   * state table
   * determine maximum conflict level over instances of each path
   */
  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (max = -1, cptr = fptr->func->caller; cptr; max++, cptr = cptr->next);
    fn = fptr->func->funcname;
    for (pptr = fptr->func->paths; pptr; pptr = pptr->next)
      /*
       * already visited ?
       */
      if (pptr->path->states < 0 && pptr->path->numinst > 0) {
#ifndef MERGE_STATES
	states = pptr->path->conflict[0];
	for (k = max; k; k--)
	  states = MAX(states, pptr->path->conflict[k]);
	pptr->path->states = states;
#else /* ifdef MERGE_STATES */
	/*
	 * determine the different conflicting lines in the equivalence class
	 * of paths, record the corresponding conflict line number (ordered),
	 * and count the number of conflict lines in this class (-> states).
	 */
	states = 0;
	v = vinit();
	for (pptr2 = pptr->path->equivs; pptr2; pptr2 = pptr2->next)
	  for (k = max; k >= 0; k--)
	    for (cfptr = pptr2->path->cfptr[k]; cfptr; cfptr = cfptr->next)
	      if (!vin(v, cfptr->line)) {
		addfmlist(&pptr->path->eqcfptr, cfptr->line);
		vinsert(&v, cfptr->line);
		states++;
	      }
	for (pptr2 = pptr->path->equivs; pptr2; pptr2 = pptr2->next) {
	  pptr2->path->states = states;
	  pptr2->path->eqcfptr = pptr->path->eqcfptr;
	}
	vfree(v);
	/*
	 * if states are merged, counters which correspond to a set of
	 * conflict lines which cannot occur may be incremented, i.e.
	 * one has to compensate for this by applying a mask to the
	 * counter offset in ease which clears impossible line misses.
	 */
#ifndef BIT_ENCODE
#ifdef DEBUG
	fprintf(ftables, "\t.global\t___path%d%s_curstate\n",
		pptr->path->equivs->path->pathnum, fn);
#else /* ifndef DEBUG */
	if (states > 0) {
#endif /*DEBUG*/
	  fprintf(ftables,
		  "___path%d%s_curstate:\n",
		  pptr->path->equivs->path->pathnum, fn);
	  fprintf(ftables, "\t.word\t%d\n",
		  ((1 << states)-1)*sizeof(int));  /*curstate*/
#ifndef DEBUG
	}
#endif /*DEBUG*/
#endif /*BIT_ENCODE*/
#endif /*MERGE_STATES*/
#ifdef DEBUG
	if (states > MAX_STATES)
	  printf("%d states in %s path %d\n",
		 states, fptr->func->funcname, pptr->path->pathnum);
#endif
	if (states > MAX_MASK_STATES) {
	  fprintf(stderr, "ERROR: %d conflicts in %s path %d\n\tequivs =",
		  states, fptr->func->funcname, pptr->path->pathnum);
	  for (pptr2 = pptr->path->equivs; pptr2; pptr2 = pptr2->next)
	    fprintf(stderr," (%d,%d)",
		    pptr2->path->pathnum, pptr2->path->insts->addr);
	  fprintf(stderr, "\n");
	  exit(1);
	}
      }
#ifdef MERGE_STATES
  }
  
  /*
   * separate state table from counters
   */
  fprintf(ftables, "\t.global\t___path_table\n");
  fprintf(ftables, "___path_table:\n");

  for (fptr = funcs; fptr; fptr = fptr->next) {
    for (max = -1, cptr = fptr->func->caller; cptr; max++, cptr = cptr->next);
    fn = fptr->func->funcname;
#endif /*MERGE_STATES*/
    /*
     * counter table
     * 1st part: address of counter tables for each called function instance
     * 2nd part: counter structure for each path
     */
    for (k = max; k >= 0; k--)
      for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
	if (pptr->path->numinst <= 0)
	  continue;

	if (pptr == fptr->func->paths) {	
	  fprintf(ftables, "\t.global\t___inst_table%s%d\n", fn, k);
	  fprintf(ftables, "___inst_table%s%d:\n", fn, k);
	  if (profile)
	    fprintf(fprofile, "-1 %s %d\n", fn, k);
	  offset = 0;

	  /*
	   * 1st part -- address of counter tables for each called function inst
	   * (1) count callee
	   * (2) dump out 1st part of counter table, but
	   *     only if # callees > 0 and if # caller's instances > 0:
	   *     o -1 (BEGIN_COUNTERTABLE)
	   *     o # callees
	   *     o address of counter table for each callee
	   */
	  i = 0;
	  for (pptr2 = fptr->func->paths; pptr2; pptr2 = pptr2->next)
	    if (pptr2->path->func && !pptr2->path->calltoexit)
	      i++;

	  if (max > 0 && i > 0) {
	    fprintf(ftables, "\t.word\t%d\n", BEGIN_COUNTER_TABLE);
	    fprintf(ftables, "\t.word\t%d\n", i);
	    offset += 2;
	    for (pptr2 = fptr->func->paths; pptr2; pptr2 = pptr2->next)
	      if (pptr2->path->func && !pptr2->path->calltoexit) {
		pptr2->path->ctab_offset = offset * sizeof(int);
		for (i = 0, cptr = pptr2->path->func->caller; cptr;
		     i++, cptr = cptr->next)
		  if (cptr->instance == k &&
		      cptr->succ_pathlabel ==
		      pptr2->path->succs->path->pathlabel &&
		      cptr->func == pptr2->path->curfunc)
		    break;
		fprintf(ftables, "\t.word\t___inst_table%s%d\n",
			pptr2->path->func->funcname, i);
		offset++;
	      }
	  }
	}

	/*
	 * 2nd part -- counter structure for each path
	 * curstate: set to state of highest conflict level for instance
	 *  all hit: number of always hits
	 * all miss: number of always misses
	 * 1st miss: number of first misses, then always hits
	 * # states: maximum number of states for any instance of this path
	 *  Example: (a) instance 0:
	 *           # states: 4 (same for all instances)
	 *           curstate: 4 (means only one hit/miss possible for instance)
	 *            all hit: 4 (+1 hit for state 4 or +2 hits for state 0)
	 *           all miss: 0
	 *           1st miss: 1
	 *           (b) instance 1:
	 *           # states: 4
	 *           curstate: 0 (means no conflicts for this instance)
	 *            all hit: 3 (+2 hits for state 0)
	 *           all miss: 0
	 *           1st miss: 2
	 *
	 * curstate has been replaced by a mask to identify the valid states
	 * when states are merged.
	 *
	 * Observe: When states are merged, it is possible that the number
	 * of states for which space is reserved exceeds the number of lines
	 * in a UP. In this case, the mask provides that we don't interpret
	 * some counters as misses. Yet, the total number of lines is adjusted
	 * by decreasing the number of hits according to the difference in
	 * the number of states and the actual lines in the UP. Thus, the
	 * number of hits may occasionally be a *negative* number (when the
	 * above difference is greater than the number of hits).
	 */
#ifdef DEBUG
	fprintf(ftables, "\t.global\t___path%d%s_%d_curstate\n",
	       pptr->path->pathnum, fn, k);
#endif /*DEBUG*/
	fprintf(ftables,
		"___path%d%s_%d_curstate:\n", pptr->path->pathnum, fn, k);
	if (profile) {
	  i = -1;
	  if (pptr->path->func)
	    for (i = 0, cptr=pptr->path->func->caller; cptr;
		 i++, cptr=cptr->next)
	      if (cptr->instance == k &&
		  cptr->succ_pathlabel == pptr->path->succs->path->pathlabel &&
		  cptr->func == pptr->path->curfunc)
		break;
	  fprintf(fprofile, "0 %d %d %s \n",
		  pptr->path->pathnum,
		  i,
		  i == -1 ? "" : pptr->path->func->funcname);
	}
#ifndef MERGE_STATES
	offset++;
	fprintf(ftables, "\t.word\t%d\n",
		((1 << pptr->path->conflict[k])-1)*sizeof(int));    /*curstate*/
#else /* ifdef MERGE_STATES */
	l = 0;

#ifndef BIT_ENCODE
	for (i = 0, cfptr = pptr->path->eqcfptr;
	     cfptr; i++, cfptr = cfptr->next)
	  for (cfptr2 = pptr->path->cfptr[k]; cfptr2; cfptr2 = cfptr2->next)
	    if (cfptr->line == cfptr2->line)
	      l |= (1 << i);
	if (pptr->path->states > MAX_STATES)
	  l <<= 2;
#endif /*!BIT_ENCODE*/

	offset++;
	fprintf(ftables, "\t.word\t%u\n", (unsigned int) l);        /*AND mask*/
#endif /*MERGE_STATES*/
	offset += 4;
#ifndef BIT_ENCODE
	if (pptr->path->states > MAX_STATES)
	  i = pptr->path->hit[k];
	else
	  i = pptr->path->hit[k] - pptr->path->states + pptr->path->conflict[k];
	fprintf(ftables, "\t.word\t%d\n", i);                       /*all hit */
	fprintf(ftables, "\t.word\t%d\n", pptr->path->miss[k]);     /*all miss*/
	fprintf(ftables, "\t.word\t%d\n", pptr->path->firstmiss[k]);/*1st miss*/
	if (pptr->path->states > MAX_STATES)                        /*# states*/
	  fprintf(ftables, "\t.word\t%d\n", - pptr->path->conflict[k]);
	else
	  fprintf(ftables, "\t.word\t%d\n", 1 << pptr->path->states);
#else /*BIT_ENCODE*/
	/*
	 * count # misses: MM fetches +
	 *                 firstmisses which are not in MM fetch list
	 */

	for (i = 0, fmptr = pptr->path->mmptr; fmptr; i++, fmptr = fmptr->next);

	l = pptr->path->firstmiss[k];
	for (fmptr = pptr->path->fmptr[k]; fmptr; fmptr = fmptr->next)
	  if (infmlist(pptr->path->mmptr, fmptr->line))
	    l--;
	
	fprintf(ftables, "\t.word\t%d\n",
		pptr->path->numinst - (i + l));                     /*all hit */
	fprintf(ftables, "\t.word\t%d\n", i);                       /*all miss*/
	fprintf(ftables, "\t.word\t%d\n", l);                       /*1st miss*/
	fprintf(ftables, "\t.word\t%d\n", 1);                       /*# states*/
#endif /*BIT_ENCODE*/
#ifdef DEBUG
	fprintf(ftables, "\t.global\t___path%d%s%d\n",
		pptr->path->pathnum, fn, k);
#endif /*DEBUG*/
	fprintf(ftables, "___path%d%s%d:\n", pptr->path->pathnum, fn, k);

	pptr->path->inst_offset = offset;
#ifdef BIT_ENCODE
	offset++;
	fprintf(ftables, "\t.word\t0\n"); /* T_COUNT */
#else /*!BIT_ENCODE*/
	if (pptr->path->states > MAX_STATES)
	  i = 2;
	else /* for (all permutations of relevant lines) */
	  i = 1 << pptr->path->states;

	fprintf(ftables, "\t.skip\t%u\n",((unsigned long int) sizeof(int)) * i);

	if (offset < MAX_OFFSET && i > 0)
	  offset += i;
	else
	  offset = MAX_OFFSET + 1;
#endif /*BIT_ENCODE*/
      }

    for (pptr = fptr->func->paths; pptr; pptr = pptr->next) {
      /*
       * MACROS to pass addr of state table for function instance (CALL)
       * (1) caller only has one instance: set reg to callee's counter tab addr
       * (2) caller has multiple instances: load callee's counter tab addr from
       *       beginnign of counter tab of current func (current inst) by
       *       indexing through ctab_offset
       */
      if (pptr->path->func)
	/* find callee instance */
	for (i = 0, cptr=pptr->path->func->caller; cptr; i++, cptr=cptr->next)
	  if (cptr->succ_pathlabel == pptr->path->succs->path->pathlabel &&
	      cptr->func == pptr->path->curfunc)
	    break;

      if (pptr->path->func &&
	  (!cptr || cptr->succs == pptr->path->succs)) {
	strcpy(base_out, (pptr->path->func->base_on_stack ? "g2" : "base_out"));
	fprintf(fmacros, "#define CALL%d%s(base_in, base_out, temp) \\\n",
		pptr->path->callnum, fn);
#ifndef TRASH
	if (max == 0) {
	  fprintf(fmacros, "\tset\t___inst_table%s%d,%%%s\n",
		  pptr->path->func->funcname, i, base_out);
	}
	else
	  fprintf(fmacros, "\tld\t[%%base_in+%d],%%%s\n", 
		  pptr->path->ctab_offset, base_out);
#else
	/* skip the first caller instance */
	for (i = 0, cptr=pptr->path->func->caller; cptr; i++, cptr=cptr->next)
	  if (cptr->succ_pathlabel == pptr->path->succs->path->pathlabel &&
	      cptr->func == pptr->path->curfunc)
	    break;
	/* generate code for all other instances */
	for (i++, cptr = cptr->next; cptr; i++, cptr = cptr->next)
	  if (cptr->succ_pathlabel == pptr->path->succs->path->pathlabel &&
	      cptr->func == pptr->path->curfunc) {
	    fprintf(fmacros, "\tset\t___inst_table%s%d,%%temp ; \\\n",
		    fn, cptr->instance);
	    fprintf(fmacros, "\tcmp\t%%temp,%%base_in ; \\\n");
	    fprintf(fmacros, "\tsethi\t%%hi(___inst_table%s%d),%%%s ; \\\n",
		    pptr->path->func->funcname, i, base_out);
	    fprintf(fmacros, "\tbe\tL___call%s%d_cont ; \\\n",
		    fn, pptr->path->callnum);
	    fprintf(fmacros, "\tor\t%%%s,%%lo(___inst_table%s%d),%%base_out ; \\\n",
		    base_out, pptr->path->func->funcname, i);
	  }

	/* generate code for the first caller instance */
	for (i = 0, cptr=pptr->path->func->caller; cptr; i++, cptr=cptr->next)
	  if (cptr->succ_pathlabel == pptr->path->succs->path->pathlabel &&
	      cptr->func == pptr->path->curfunc)
	    break;
	fprintf(fmacros, "\tset\t___inst_table%s%d,%%%s ; \\\n",
		pptr->path->func->funcname, i, base_out);
	fprintf(fmacros, "L___call%s%d_cont: \n", fn, pptr->path->callnum);
#endif
      }
      
      /*
       * MACRO expansion for measurements, called by UP (transition) (PATH)
       */

      /*
       * (a) increment count for current state and save new (zero) state
       */
      fprintf(fmacros, "#define PATH%d%s(base, temp1, temp2) \\\n",
	      pptr->path->pathnum, fn);
#ifdef DEBUG
      fprintf(fmacros, "\t.global\t___path%d%s_ ; \\\n",
	      pptr->path->pathnum, fn);
      fprintf(fmacros, "___path%d%s_: ; \\\n", pptr->path->pathnum, fn);
#endif /*DEBUG*/

      if (pptr->path->numinst <= 0) {
	fprintf(fmacros, "\n");
	continue;
      }

#ifdef MERGE_STATES
#ifndef BIT_ENCODE
      clear_pending = FALSE;
      if (pptr->path->states > 0) {
	fprintf(fmacros, "\tsethi\t%%hi(___path%d%s_curstate),%%temp1 ; \\\n",
		pptr->path->equivs->path->pathnum, fn);
	fprintf(fmacros,
		"\tld\t[%%temp1+%%lo(___path%d%s_curstate)],%%temp2 ; \\\n",
		pptr->path->equivs->path->pathnum, fn);

	/*
	 * clear state (1)
	 *
	 * (a) Only clear those lines which are part of this path.
	 *     Lines of other paths in the same equivalence class
	 *     should remain untouched.
	 */
	k = l = 0;
	for (i = 0, cfptr = pptr->path->eqcfptr; cfptr;
	     i++, cfptr = cfptr->next)
	  for (iptr = pptr->path->insts; iptr; iptr = iptr->next)
	    if (iptr->firstline + numsets <= cfptr->line &&
		cfptr->line <= iptr->lastline + numsets &&
		!(l & ((1 << i) * sizeof(int)))) {
	      k++;
	      l |= (1 << i) * sizeof(int);
	    }

	/*
	 * (b) Only clear those lines which are actually cached
	 *     at the end of the path. Consider |UP| > |cache|.
	 *     Then only the last |cache| lines are in cache after
	 *     executing the UP. This is done by cache_self().
	 *     In addition, lines in the equivalence class which
	 *     are replaced by executing this path need to be
	 *     invalidated, done in the loop below.
	 */
	i = 0;
	v = vinit();
	uncached = cache_self(pptr->path->insts, &v, pptr->path->eqcfptr, &i);
	for (m = 0, cfptr = pptr->path->eqcfptr; cfptr;
	     m++, cfptr = cfptr->next) {
	  for (iptr = pptr->path->insts; iptr; iptr = iptr->next)
	    if (iptr->firstline + numsets <= cfptr->line &&
		cfptr->line <= iptr->lastline + numsets)
	      break;
	  if (!iptr && vin(v, cfptr->line % numsets))
	    uncached |= sizeof(int) << m;
	}
	vfree(v);

	if (k > 0)
	  if (pptr->path->states == k && !uncached)
	    fprintf(fmacros,
		    "\tst\t%%g0,[%%temp1+%%lo(___path%d%s_curstate)] ; \\\n",
		    pptr->path->equivs->path->pathnum, fn);
	  else
	    clear_pending = TRUE;

	if (pptr->path->states > MAX_STATES) {
	  sprintf(s1, "___path%d%s_0_curstate", pptr->path->pathnum, fn);
	  sprintf(s2, "___inst_table%s0", fn);
	  dump_ldst(LOAD, "base", s1, s2, "temp1", pptr->path->inst_offset - 5);
	  
	  fprintf(fmacros, "\tandcc\t%%temp2,%%temp1,%%temp1 ; \\\n");
	  
	  fprintf(fmacros, "\tbz\tL___path%d%s_nocnt;\\\n",
		  pptr->path->pathnum, fn);
	  fprintf(fmacros, "\tmov\t%%g0,%%temp2 ; \\\n");
	  
	  fprintf(fmacros,"L___path%d%s_cnt: ; \\\n", pptr->path->pathnum, fn);
	  fprintf(fmacros,"\tbz\tL___path%d%s_ccnt;\\\n",pptr->path->pathnum,fn);
	  fprintf(fmacros,"\tbtst\t1,%%temp1 ; \\\n");
	  fprintf(fmacros,"\tsrl\t%%temp1,1,%%temp1 ; \\\n");
	  fprintf(fmacros,"\tbz\tL___path%d%s_cnt ;\\\n",pptr->path->pathnum,fn);
	  fprintf(fmacros,"\ttst\t%%temp1 ; \\\n");
	  fprintf(fmacros,"\tb\tL___path%d%s_cnt ; \\\n",pptr->path->pathnum,fn);
	  fprintf(fmacros,"\tinc\t%%temp2 ; \\\n");
	  
	  fprintf(fmacros, "L___path%d%s_ccnt: ; \\\n", pptr->path->pathnum, fn);
	  reg = "base";

	  /*
	   * increment conflict count
	   */
	  sprintf(s1, "4+___path%d%s0", pptr->path->pathnum, fn);
	  sprintf(s2, "___inst_table%s0", fn);
	  dump_ldst(LOAD, reg, s1, s2, "temp1", pptr->path->inst_offset + 1);
	  fprintf(fmacros, "\tadd\t%%temp1,%%temp2,%%temp2 ; \\\n");
	  reload_offset("temp1", reg, s1, s2, pptr->path->inst_offset + 1);
	  dump_ldst(STORE, reg, s1, s2, "temp2", pptr->path->inst_offset + 1);

	  fprintf(fmacros,"L___path%d%s_nocnt: ; \\\n", pptr->path->pathnum, fn);
	}
	else {
	  fprintf(fmacros, "\tadd\t%%temp2,%%base,%%temp1 ; \\\n");
	  reg = "temp1";
	}
      }
      else
#endif /*!BIT_ENCODE*/
	reg = "base";

      /*
       * increment count
       */
      sprintf(s1, "___path%d%s0", pptr->path->pathnum, fn);
      sprintf(s2, "___inst_table%s0", fn);
      dump_ldst(LOAD, reg, s1, s2, "temp2", pptr->path->inst_offset);
      fprintf(fmacros, "\tinc\t%%temp2 ; \\\n");
      dump_ldst(STORE, reg, s1, s2, "temp2", pptr->path->inst_offset);

#ifndef BIT_ENCODE
      /*
       * clear state (2)
       */
      if (clear_pending) {
	if (l && l != uncached || uncached && pptr->path->states != i) {
	  fprintf(fmacros, "\tsethi\t%%hi(___path%d%s_curstate),%%temp1 ; \\\n",
		  pptr->path->equivs->path->pathnum, fn);
	  fprintf(fmacros, "\tld\t[%%temp1+%%lo(___path%d%s_curstate)],%%temp2 ; \\\n",
		  pptr->path->equivs->path->pathnum, fn);
	}

	if (l && l != uncached)
	  dump_logic("andn", "temp2", "temp1", l);
	if (uncached)
	  if (l && l != uncached || pptr->path->states != i)
	    dump_logic("or", "temp2", "temp1", uncached);
	  else
	    if ((unsigned int) uncached < MAX_VALUE)
	      fprintf(fmacros, "\tmov\t%d,%%temp2 ; \\\n", uncached);
	    else {
	      fprintf(fmacros, "\tmov\t%%g0,%%temp2 ; \\\n");
	      dump_logic("or", "temp2", "temp1", uncached);
	    }

	if (l && l != uncached && (unsigned int) l >= MAX_VALUE ||
	    (unsigned int) uncached >= MAX_VALUE ||
	    uncached && pptr->path->states == i) /* reload */
	  fprintf(fmacros, "\tsethi\t%%hi(___path%d%s_curstate),%%temp1 ; \\\n",
		  pptr->path->equivs->path->pathnum, fn);
	fprintf(fmacros, "\tst\t%%temp2,[%%temp1+%%lo(___path%d%s_curstate)] ; \\\n",
		pptr->path->equivs->path->pathnum, fn);
      }
#endif /*!BIT_ENCODE*/

#else /*!MERGE_STATES*/

      if (pptr->path->states > 0) {
	fprintf(fmacros, "\tld\t[%%base+___path%d%s_0_curstate-___inst_table%s0],%%temp1 ; \\\n",
		pptr->path->pathnum, fn, fn);
	fprintf(fmacros, "\tadd\t%%temp1,%%base,%%temp1 ; \\\n");
	reg = "temp1";
      }
      else
	reg = "base";
      fprintf(fmacros, "\tld\t[%%%s+___path%d%s0-___inst_table%s0],%%temp2 ; \\\n",
	      reg, pptr->path->pathnum, fn, fn);
      fprintf(fmacros, "\tinc\t%%temp2 ; \\\n");
      fprintf(fmacros, "\tst\t%%temp2,[%%%s+___path%d%s0-___inst_table%s0] ; \\\n",
	      reg, pptr->path->pathnum, fn, fn);
      /*
       * clear state: might need some more work, see clear state above
       */
      if (pptr->path->states > 0)
	fprintf(fmacros, "\tst\t%%g0,[%%base+___path%d%s_0_curstate-___inst_table%s0] ; \\\n",
		pptr->path->pathnum, fn, fn);
#endif /*!MERGE_STATES*/

#ifndef BIT_ENCODE
      /*
       * (b) collect conflicting states of the same progam line
       * [could be improved on by checking which instance can be reached]
       */
      lchange = (struct linechange_t *) NULL;
      for (iptr = pptr->path->insts; iptr; iptr = iptr->next)
	for (l = iptr->firstline; l <= iptr->lastline; l++) {
	  /*
	   * skip if not cached at end of path
	   */
	  v = vinit();
	  cache_path(pptr->path, pptr->path->in[0][0].v,
		     &v, dummy, 0, vinit(), 0);
	  k = vin(v, l + numsets);
	  vfree(v);
	  if (!k)
	    continue;
	  for (lptr = setpath[l % numsets]; lptr; lptr = lptr->next)
	    if (l / numsets == lptr->tag)
	      break;
	  for (pptr2 = lptr->list; pptr2; pptr2 = pptr2->next)
#ifdef MERGE_STATES
	    if (pptr->path->equivs != pptr2->path->equivs)
	   /*if (!inpathlist(pptr->path->equivs, pptr2->path)) slower*/
#endif /*MERGE_STATES*/
	      for (k = 0, cptr = pptr2->path->curfunc->caller; cptr;
		   k++, cptr = cptr->next)
		for (i = 0, fmptr = pptr2->path->cfptr[k]; fmptr;
		     i++, fmptr = fmptr->next)
		  if (fmptr->line == l + numsets
#ifdef TRASH
		      && vin(pptr2->path->rin[k].v, l + numsets))
#else
		      && vin(pptr2->path->in[k][0].v, l + numsets))
#endif
#ifndef MERGE_STATES
		    lchange = addlchange(lchange, pptr2->path, k,
					 sizeof(int) << i);
#else /*MERGE_STATES*/
                    for (m = 0, cfptr = pptr2->path->eqcfptr; cfptr;
			 m++, cfptr = cfptr->next)
		      if (cfptr->line == l + numsets)
			/*
			 * The check with uncached makes sure that we are
			 * looking at a line which actually is cached
			 * after executing the path.
			 */
			if (!inlchange(lchange, pptr2->path->equivs->path,
				       sizeof(int) << m))
			  lchange = addlchange(lchange,
					       pptr2->path->equivs->path, k,
					       sizeof(int) << m);
#endif /*MERGE_STATES*/
	}

      /*
       * update conflicting states of the same progam line
       * [should sort by function and instance to avoid repeated sets,
       * starting with current function and instance which is already loaded.]
       */
      last_k = 0;
      last_fn = NULL;
      temp1_set = FALSE;
      for (lchange2 = lchange; lchange2; lchange2 = lchange2->next) {
	/*
	 * only one instance and caching current path? -> done.
	 */
	if (pptr->path == lchange2->path && !pptr->path->curfunc->caller->next)
	  continue;
	k = lchange2->inst;
#ifndef MERGE_STATES
	if (last_k != k || last_fn != lchange2->path->curfunc->funcname)
	  fprintf(fmacros, "\tset\t___inst_table%s%d,%%temp1 ; \\\n",
		  last_fn = lchange2->path->curfunc->funcname, last_k = k);
#endif /*!MERGE_STATES*/
	if (pptr->path == lchange2->path ||
	    lchange2->state == ((1 << lchange2->path->states)-1) * sizeof(int)) {
	  /*
	   * same path, different instance or all flags set -> all lines cached
	   */
#ifdef MERGE_STATES
	  if (!temp1_set) {
	    temp1_set = TRUE;
	    fprintf(fmacros, "\tset\t___state_table,%%temp1 ; \\\n");
	  }
	  fprintf(fmacros, "\tst\t%%g0,[%%temp1+___path%d%s_curstate-___state_table] ; \\\n",
		  lchange2->path->equivs->path->pathnum,
		  lchange2->path->curfunc->funcname);
#else /* ifndef MERGE_STATES */
	  fprintf(fmacros, "\tst\t%%g0,[%%temp1+___path%d%s_0_curstate-___inst_table%s0] ; \\\n",
		  lchange2->path->pathnum,
		  lchange2->path->curfunc->funcname,
		  lchange2->path->curfunc->funcname);
#endif /*MERGE_STATES*/
	}
	else {
	  /*
	   * different paths -> some line replaced
	   */
#ifdef MERGE_STATES
	  if (!temp1_set) {
	    temp1_set = TRUE;
	    fprintf(fmacros, "\tset\t___state_table,%%temp1 ; \\\n");
	  }
	  fprintf(fmacros, "\tld\t[%%temp1+___path%d%s_curstate-___state_table],%%temp2 ; \\\n",
		  lchange2->path->equivs->path->pathnum,
		  lchange2->path->curfunc->funcname);
	  dump_logic("andn", "temp2", "temp1", lchange2->state);
	  reload_logic("temp1", "___state_table", lchange2->state);
	  fprintf(fmacros, "\tst\t%%temp2,[%%temp1+___path%d%s_curstate-___state_table] ; \\\n",
		  lchange2->path->equivs->path->pathnum,
		  lchange2->path->curfunc->funcname);
#else /* ifndef MERGE_STATES */
	  fprintf(fmacros, "\tld\t[%%temp1+___path%d%s_0_curstate-___inst_table%s0],%%temp2 ; \\\n",
		  lchange2->path->pathnum,
		  lchange2->path->curfunc->funcname,
		  lchange2->path->curfunc->funcname);
	  dump_logic("andn", "temp2", "temp1", lchange2->state);
	  reload_logic("temp1", "___state_table", lchange2->state);
	  fprintf(fmacros, "\tst\t%%temp2,[%%temp1+___path%d%s_0_curstate-___inst_table%s0] ; \\\n",
		  lchange2->path->pathnum,
		  lchange2->path->curfunc->funcname,
		  lchange2->path->curfunc->funcname);
#endif /*MERGE_STATES*/
	}
      }

      /*
       * (c) update other (conflicting) states
       * Notice: last instruction buffered to fill possible branch delay slot
       */
#ifdef MERGE_STATES
      pptr2 = (struct pathlist *) NULL;

      for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next)
	for (lconf = pptr->path->conf[k]; lconf; lconf = lconf->next)
	  if (!inpathlist(pptr->path->equivs, lconf->path) ||
	      !inpathinsts(pptr->path->insts, lconf->rline - numsets))
	    for (i = 0, cfptr = lconf->path->eqcfptr; cfptr;
		 i++, cfptr = cfptr->next)
	      if (cfptr->line == lconf->rline &&
		  !inpathlist(pptr2, lconf->path->equivs->path)) {
		add_pathlist(&pptr2, lconf->path->equivs->path);

		l = 0;
		m = k;
		cptr2 = cptr;
		lconf2 = lconf;
		while (cptr2) {
		  if (m != k)
		    lconf2 = pptr->path->conf[m];

		  while (lconf2) {
		    if (lconf2->path->equivs == lconf->path->equivs)
		      for (i = 0, cfptr2 = lconf2->path->eqcfptr; cfptr2;
			   i++, cfptr2 = cfptr2->next)
			/*
			 * The inlchange test excludes lines which were
			 * cleared with the "andn" previously.
			 */
			if (cfptr2->line == lconf2->rline &&
			    !inlchange(lchange, lconf->path->equivs->path,
				       sizeof(int) << i))
			  l |= sizeof(int) << i;

		    lconf2 = lconf2->next;
		  }
		  m++;
		  cptr2 = cptr2->next;
		}

		if (l) {
		  if (!temp1_set) {
		    temp1_set = TRUE;
		    fprintf(fmacros, "\tset\t___state_table,%%temp1 ; \\\n");
		  }
		  fprintf(fmacros, "\tld\t[%%temp1+___path%d%s_curstate-___state_table],%%temp2 ; \\\n",
			  lconf->path->equivs->path->pathnum,
			  lconf->path->curfunc->funcname);
		  dump_logic("or", "temp2", "temp1", l);
		  reload_logic("temp1", "___state_table", l);
		  fprintf(fmacros, "\tst\t%%temp2,[%%temp1+___path%d%s_curstate-___state_table] ; \\\n",
			  lconf->path->equivs->path->pathnum,
			  lconf->path->curfunc->funcname);
		}
	      }
      
      free_pathlist(pptr2);
#else /*!MERGE_STATES*/
      last_conflict = -1;
      for (k = 0, cptr = fptr->func->caller; cptr; k++, cptr = cptr->next) {
	for (lconf = pptr->path->conf[k]; lconf; lconf = lconf->next) {
	  for (i = 0, fmptr = lconf->path->cfptr[lconf->instance]; fmptr;
	       i++, fmptr = fmptr->next)
	    if (fmptr->line == lconf->rline)
	      break;
#ifndef DEBUG
	  if (!fmptr) {
	    fprintf(stderr, "remote conflict not found in path%d%s%d conflicting with path%d%s%d\n",
		    pptr->path->pathnum, fn, k, lconf->path->pathnum,
		    lconf->path->curfunc->funcname, lconf->instance);
	    fflush(stderr);
	  }
#endif /*DEBUG*/
	  if (last_conflict != k && fptr->func != mainfuncptr) {
	    if (last_conflict >= 0) {
	      fprintf(fmacros, "\tba\tL___path%d%s_cont ; \\\n",
		      pptr->path->pathnum, fn);
	      FLUSH_DELAYED_INST(fmacros, last_inst);
	      fprintf(fmacros, "L___path%d%s%d_cont: ; \\\n",
		      pptr->path->pathnum, fn, last_conflict);
	      last_k = 0;
	      last_fn = NULL;
	    }
	    FLUSH_DELAYED_INST(fmacros, last_inst);
	    last_conflict = k;
	    if (last_k != k || last_fn != fn)
	      fprintf(fmacros, "\tset\t___inst_table%s%d,%%temp1 ; \\\n",
		      last_fn = fn, last_k = k);
	    fprintf(fmacros, "\tcmp\t%%temp1,%%base ; \\\n");
	    fprintf(fmacros, "\tbne\tL___path%d%s%d_cont ; \\\n",
		    pptr->path->pathnum, fn, k);
	  }
	  FLUSH_DELAYED_INST(fmacros, last_inst);
	  if (last_k != lconf->instance ||
	      last_fn != lconf->path->curfunc->funcname) {
	    fprintf(fmacros, "\tsethi\t%%hi(___inst_table%s%d),%%temp1 ; \\\n",
		    last_fn = lconf->path->curfunc->funcname,
		    last_k = lconf->instance);
	    fprintf(fmacros, "\tor\t%%temp1,%%lo(___inst_table%s%d),%%temp1 ; \\\n",
		    last_fn = lconf->path->curfunc->funcname,
		    last_k = lconf->instance);
	  }
	  fprintf(fmacros, "\tld\t[%%temp1+___path%d%s_0_curstate-___inst_table%s0],%%temp2 ; \\\n",
		  lconf->path->pathnum, lconf->path->curfunc->funcname,
		  lconf->path->curfunc->funcname);
	  fprintf(fmacros, "\tor\t%%temp2,%d,%%temp2 ; \\\n", sizeof(int) << i);
	  sprintf(last_inst, "\tst\t%%temp2,[%%temp1+___path%d%s_0_curstate-___inst_table%s0] ; \\\n",
		  lconf->path->pathnum, lconf->path->curfunc->funcname,
		  lconf->path->curfunc->funcname);
	}
      }
      FLUSH_DELAYED_INST(fmacros, last_inst);
      if (last_conflict >= 0) {
	fprintf(fmacros, "L___path%d%s%d_cont: ; \\\n",
		pptr->path->pathnum, fn, last_conflict);
	fprintf(fmacros, "L___path%d%s_cont: ; \\\n", pptr->path->pathnum, fn);
      }
#endif /*!MERGE_STATES*/
      
      freelchange(lchange);

#endif /*!BIT_ENCODE*/
      
      fprintf(fmacros, "\n");
    }
  }
  
  /*
   * dump out 0 state to terminte table
   */
  fprintf(ftables, "\t.word\t0\n"); /* state/mask  */
  fprintf(ftables, "\t.word\t0\n"); /* always hit  */
  fprintf(ftables, "\t.word\t0\n"); /* all miss    */
  fprintf(ftables, "\t.word\t0\n"); /* 1st miss    */
  fprintf(ftables, "\t.word\t0\n"); /* # states    */
  
  /*
   * dump out conflict table
   */
  
  fprintf(ftables, "\t.global\t___conf_table\n");
  fprintf(ftables, "___conf_table:\n");
  
  for (i = 0; i < numsets; i++)
    for (lptr = setpath[i]; lptr; lptr = lptr->next) {
      last_conflict = FALSE;
      for (pptr = lptr->list; pptr; pptr = pptr->next)
	if (pptr->path->numinst > 0)
	  for (k = 0, cptr = pptr->path->curfunc->caller; cptr;
	       k++, cptr = cptr->next)
	    for (fmptr = pptr->path->fmptr[k]; fmptr; fmptr = fmptr->next)
	      if ((fmptr->line % numsets) == i &&
		  (fmptr->line / numsets) == lptr->tag) {
		fprintf(ftables, "\t.word\t___path%d%s_%d_curstate\n",
			pptr->path->pathnum, pptr->path->curfunc->funcname, k);
		last_conflict = TRUE;
		break;
	      }
      if (last_conflict)
	fprintf(ftables, "\t.word\t0\n"); /* end of conflict set */
    }
  fprintf(ftables, "\t.word\t0\n"); /* end of conflict table */

  /*
   * dump out stubs for dead code paths
   */
  for (pptr = dead_paths; pptr; pptr = pptr->next)
    fprintf(fmacros, "#define PATH%d%s(base, temp1, temp2) \n",
	    pptr->path->pathnum, pptr->path->curfunc->funcname);

  free_set(dummy);
}

#endif

