/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: misc.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Miscellaneous
 */

#include "scs.h"

/*
 * error - print error message and halt
 */
void error(s)
char *s;
{
   fprintf(stderr, "%s\n", s);
   exit(1);
}

/*
 * read_config - read cache configuration information
 */
void read_config(configfn)
char *configfn;
{
    unsigned int k;
    char answer[100];
    FILE *cfile;
    FILE *fopen();
    int fclose(), fflush();
    void error();

    if (!(cfile = fopen(configfn, "r")))
      error("error opening configuration file");

    if (!fgets(answer, 100, cfile))
      error("error reading title in config file");

    if (fscanf(cfile, "Number of sets: %u\n", &numsets) != 1)
      error("error reading numsets in config file");
    if (numsets > MAXNUMSETS)
      error("number of sets requested exceeds MAXNUMSETS");
    numtagwords = numsets / 4;

    if (fscanf(cfile, "Set size: %u\n", &setsize) != 1)
      error("error reading setsize in config file");
    if (setsize > MAXSETSIZE)
      error("set size requested exceeds MAXSETSIZE");

    if (fscanf(cfile, "Line size: %u\n", &linesize) != 1)
      error("error reading linesize in config file");
    if (linesize & (linesize - 1))
      error("line size must be a power of two");
    k = linesize;
    for (cline_shift = -1; k; k >>= 1)
      cline_shift++;
    cachesize = numsets*linesize;
    instonline = linesize / INSTSIZE;

    (void) fclose(cfile);
}

/*
 * addlchange - add another change to some line
 */
struct linechange_t *addlchange(head, path, inst, state)
struct linechange_t *head;
struct pathinfo *path;
int inst, state;
{
  struct linechange_t *lchange;

  for (lchange = head; lchange; lchange = lchange->next)
    if (lchange->path == path && lchange->inst == inst) {
      lchange->state |= state;
      break;
    }

  if (!lchange) {
    lchange = (struct linechange_t *) MALLOC(sizeof(*lchange));
    lchange->next = head;
#ifdef MERGE_STATES
    lchange->path = path->equivs->path;
#else
    lchange->path = path;
#endif
    lchange->inst = inst;
    lchange->state = state;
    head = lchange;
  }

  return(head);
}

#ifdef MERGE_STATES
/*
 * inlchange - test if line change already recorded for any inst of current path
 * or any other path in the equivalence class
 */
int inlchange(lchange, path, state)
struct linechange_t *lchange;
struct pathinfo *path;
int state;
{
  struct pathlist *pptr;

  for (; lchange; lchange = lchange->next)
    for (pptr = path->equivs; pptr; pptr = pptr->next)
      if (lchange->path == pptr->path && lchange->state & state)
	return(TRUE);
  return(FALSE);
}
#endif /*MERGE_STATES*/

/*
 * freelchange - free a line change list
 */
void freelchange(lchange)
struct linechange_t *lchange;
{
  if (lchange) {
    freelchange(lchange->next);
    FREE(lchange);
  }
}
    
/*
 * allocate_vectors - allocate bit vecvtors for program lines for paths in func
 */
void allocate_vectors(func, num)
struct funcinfo *func;   /* for this function */
int num;                 /* allocate so many new instances */
{
  struct veclist **in;
  struct veclist *rout;
  struct veclist *pout, **lin, *din;
  struct pathlist *pptr;
  int i, s;
  vector v;

#ifndef SLOW_ALLOC
  char *new;
  int k, numint, numvecs, numvecptrs;

  numint  = 4;
  numvecs = setsize;
  numvecptrs = 1;

  numvecs++;
  numvecs++;
  numvecs++;
  numvecs += setsize;
  numvecptrs++;

  for (k = 0, pptr = func->paths; pptr; k++, pptr = pptr->next);
  i = numint     * sizeof(int) +
      2          * sizeof(struct fmline_t *) +
      1          * sizeof(struct conf_t *) +
      numvecs    * sizeof(struct veclist) +
      numvecptrs * sizeof(struct veclist *);
  new = (char *) CALLOC(k * num, i);
#endif /*!SLOW_ALLOC*/

  for (pptr = func->paths; pptr; pptr = pptr->next) {
#ifndef SLOW_ALLOC
    pptr->path->hit = (int *) new; new += (sizeof(int) * num);
    pptr->path->miss = (int *) new; new += (sizeof(int) * num);
    pptr->path->firstmiss = (int *) new; new += (sizeof(int) * num);
    pptr->path->conflict = (int *) new; new += (sizeof(int) * num);
    pptr->path->fmptr = (struct fmline_t **) new;
      new += (num * sizeof(struct fmline_t *));
    pptr->path->cfptr = (struct fmline_t **) new;
      new += (num * sizeof(struct fmline_t *));
    pptr->path->conf = (struct conf_t **) new;
      new += (num * sizeof(struct conf_t *));
    in = (struct veclist **) new; new += (sizeof(struct veclist *) * num);
    for (i = 0; i < num; i++) {
      in[i] = (struct veclist *) new; new += (sizeof(struct veclist) * setsize);
    }
    rout = (struct veclist *) new; new += (sizeof(struct veclist) * num);
    pout = (struct veclist *) new; new += (sizeof(struct veclist) * num);
    lin = (struct veclist **) new; new += (sizeof(struct veclist *) * num);
    din = (struct veclist *) new; new += (sizeof(struct veclist) * num);
    for (i = 0; i < num; i++) {
      lin[i] = (struct veclist *) new; new += (sizeof(struct veclist) *setsize);
    }
#else /*SLOW_ALLOC*/
    pptr->path->hit = (int *) MALLOC(sizeof(int) * num);
    pptr->path->miss = (int *) MALLOC(sizeof(int) * num);
    pptr->path->firstmiss = (int *) MALLOC(sizeof(int) * num);
    pptr->path->conflict = (int *) MALLOC(sizeof(int) * num);
    pptr->path->fmptr = (struct fmline_t **)
      CALLOC(num, sizeof(struct fmline_t *));
    pptr->path->cfptr = (struct fmline_t **)
      CALLOC(num, sizeof(struct fmline_t *));
    pptr->path->conf = (struct conf_t **) CALLOC(num, sizeof(struct conf_t *));
    in = (struct veclist *) MALLOC(sizeof(struct veclist *) * num);
    for (i = 0; i < num; i++)
      in[i] = (struct veclist *) MALLOC(sizeof(struct veclist) * setsize);
    rout = (struct veclist *) MALLOC(sizeof(struct veclist) * num);
    pout = (struct veclist *) MALLOC(sizeof(struct veclist) * num);
    lin = (struct veclist **) MALLOC(sizeof(struct veclist *) * num);
    din = (struct veclist *) MALLOC(sizeof(struct veclist) * num);
    for (i = 0; i < num; i++)
      lin[i] = (struct veclist *) MALLOC(sizeof(struct veclist) * setsize);
#endif /*SLOW_ALLOC*/

    v = vallptr;
    for (i = 0; i < num; i++) {
      rout[i].v = vinit();
      pout[i].v = v;
      din[i].v = vinit();

      rout[i].changed =
	pout[i].changed =
	  din[i].changed =
	    FALSE;

      for (s = 0; s < setsize; s++) {
	in[i][s].v = vinit();
	in[i][s].changed = FALSE;
	lin[i][s].v = vinit();
	lin[i][s].changed = FALSE;
      }
    }

    pptr->path->in = in;
    pptr->path->rout = rout;
    pptr->path->pout = pout;
    pptr->path->lin = lin;
    pptr->path->din = din;
  }
}

#ifdef DEBUG
/*
 * output_conflicts - print out statistical information about conflicts
 */
void output_conflicts()
{
  register int i, k, max = 0, avg = 0;
  struct pathlistlist *pptr;
  
  for (i = 0; i < numsets; i++) {
    k = 0;
    pptr = setpath[i];
    while (pptr) {
      k++;
      pptr = pptr->next;
    }
    if (k > max)
      max = k;
    avg += k;
  }
  printf("%d sets with avg. of %f and a max. of %d conflicts found\n",
	 numsets, (float) avg / (float) numsets, max);
}
#endif /*DEBUG*/

/*
 * c_init - initialize cache
 */
void c_init()
{
  register int i;
  
  for (i = 0; i < numsets; i++) {
    currtag[i] = (unsigned char) -1;
    setpath[i] = (struct pathlistlist *) NULL;
  }   
}

/*
 * init_cacher - initialize cache trace simulation
 */
void init_cacher()
{
  char *configname, outname[100];
  FILE *fopen();
  char *get_env();
  void read_config(), c_init(), error();
  
  read_config("trace.config");
}

/*
 * init - general initializations
 */
void init()
{
  v_init();
}

