/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: scs.h,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * General data structures, constances, externals
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/times.h>
#ifdef PROFILING
#include <sun4/asm_linkage.h>
#endif

#include "misc.h"
#include "myalloc.h"
#include "vec.h"

#define MAXNAME 80
#define MAXSTRING 512*8

#define INSTSIZE 4
#define NOCALL   0
#define CALL     1
#define INVALID -1

#define MAXSETSIZE 8
#define MAXNUMSETS 8192
#define MAXPROGLINES 16384

#define MAXPREDS 30

/*
 * maximum offset for a load/store instruction is 1024 words (4kB)
 */
#define MAX_OFFSET 1024
#define MAX_VALUE (MAX_OFFSET * sizeof(int))

/*
 * maximum number of states for which we generate a counter array (size <= 2**8)
 * and maximum number of states for a bit-mask (30 bits) [ << 2 fills 32 bits ]
 * For the latter part, the lower two bits are never set. This is used to
 * recognize the beginning of a counter table for a function instance by the
 * mask below (0xffffffff)
 */
#define MAX_STATES 8
#define MAX_MASK_STATES 30
#define BEGIN_COUNTER_TABLE -1

#define FLUSH_DELAYED_INST(fd, str) \
  if (str[0] != '\0') { \
    fprintf(fd, "%s", str);\
    str[0] = '\0'; \
  }

#define SAVE_AREA_OFFSET (WINDOWSIZE+ARGPUSHSIZE)

struct statinfo_s {
  int hits;
  int misses;
  int firstmisses;
  int firsthits;
  int conflicts;
};

#ifdef BIT_ENCODE
#define MMCOST   9
#define HITCOST  1
#define MISSCOST HITCOST+MMCOST

struct fetchinfo_s {
  int cache;
  int mm;
};

struct accessinfo_s {
  int miss;
  int conflict;
  int firstmiss;
};

struct costlist_t {
  struct costlist_t *next;
  struct pathinfo *path;
  int cost;
};
#endif

struct instinfo {
  struct instinfo *next;        /* next sequential list of insts in path */
  int addr;                     /* address of first inst in sequential list */
  int num;                      /* number of insts in sequential list */
  int size;                     /* size of insts in sequential list */
  int firstline;                /* first line insts reference in the program */
  int lastline;                 /* last line insts reference in the program */
  int numlines;                 /* number of cache lines referenced */
  int startindex;               /* starting cache set index number */
  unsigned int tag;            /* tag of first cache line referenced */
};
  
struct traversalinfo {
  struct traversalinfo *next;   /* next traversal for the path */
  int inputstate;               /* input cache state */
  int outputstate;              /* output cache state */
  int *in_index, *out_index;    /* state index for input and output state */
  int hits;                     /* number of cache hits from traversal */
  int misses;                   /* number of cache misses from traversal */
};

/*
 * different cateories
 */
#define C_HIT        0
#define C_MISS       1
#define C_FIRSTMISS  2
#define C_CONFLICT   3

/*
 * Offsets in path data
 */
#define T_STATE      0
#define T_HIT        4
#define T_MISS       8
#define T_FIRSTMISS 12
#define T_NUMSTATES 16
#define T_COUNT     20

struct linechange_t {
  struct linechange_t *next;      /* link */
  struct pathinfo *path;         /* path */
  int inst;                      /* instance */
  int state;                     /* states to be merged */
};

struct veclist {
  vector v;
  int changed;
};

struct callerlist {
  struct callerlist *next;       /* link */
  struct pathinfo *caller;       /* caller's path */
  struct pathlist *succs;        /* caller's successor paths */
  struct funcinfo *func;         /* caller's function */
  int succ_pathlabel;            /* path label of the successor paths */
  int instance;                  /* caller's instance */
};

struct conf_t {
  struct conf_t *next;           /* link */
  int lline;                     /* conflicting line in current path */
  int rline;                     /* conflicting line in remote path */
  int instance;                  /* remote instance */
  struct pathinfo *path;         /* remote path */
};

struct fmline_t {
  struct fmline_t *next;
  int line;
};

struct pathinfo {
  int pathlabel;                /* number that initially identifies the path */
  int pathnum;                  /* number identifying the path */
  int loopno;                   /* number identifying loop in which path is */
  int freq;                     /* number of iterations if in loop (local) */
  struct pathlist *preds;       /* list of predecessor paths */
  int numpreds;                 /* number of predecessor paths */
  int succtype;                 /* type of successor (NOCALL or CALL) */
  char *funcname;               /* name of called function (is freed early on)*/
  struct funcinfo *func;        /* ptr to function called */
  int callnum;                  /* number of bblock func is called from */
  int ctab_offset;              /* counter table offset for callee in table */
  struct funcinfo *curfunc;     /* ptr to function path is in */
  int calltoexit;               /* indicates if exit is being called */
  struct pathlist *succs;       /* list of successor paths (one for a call) */
#ifdef MERGE_STATES
  struct pathlist *equivs;      /* list of equivalence paths (same label) */
#endif
  struct instinfo *insts;       /* list of insts that comprise the path */
  struct traversalinfo *traversal;
                                /* list of infos about traversing path */
  int numinst;                  /* total number of insts within the path */
  int fitsincache;              /* indicates if the path fits in the cache */
                                /* per function instance: */
  int *hit;                     /*   # line hits */
  int *miss;                    /*   # line misses */
  int *firstmiss;               /*   # first miss then always hit */
  int *conflict;                /*   # hits or misses due to conflict */
  struct fmline_t **fmptr;      /*   array of lists of lines w/ first misses */
  struct fmline_t **cfptr;      /*   array of lists of lines w/ conflicts */
  struct conf_t **conf;         /*   array of lists of conflicting lines */
  int states;                   /* # states in table for path */
#ifdef MERGE_STATES
  struct fmline_t *eqcfptr;     /* array of lines w/ conflicts of this state */
#endif /*MERGE_STATES*/
#ifdef BIT_ENCODE
  struct fmline_t *mmptr;       /* list of lines fetched from main memory */
  struct fmline_t *missptr;     /* list of lines with misses */
  struct fmline_t *firstmissptr;/* list of lines with first misses */
  struct fmline_t *conflictptr; /* list of lines with conflicts */
  struct costlist_t *costptr;   /* transferred cost from other paths */
#endif /*BIT_ENCODE*/
  struct veclist **in;          /* list of sets of incoming cached pgm lines */
  struct veclist *rout;         /* list of outgoing reaching program lines */
  struct veclist *pout;         /* list of outgoing post-dominating pgm lines */
  struct veclist *din;          /* list of incoming dominating pgm lines      */
  struct veclist **lin;         /* list of sets of inc. lin. cached pgm lines */
  int instance;                 /* instance of function invocation */

  int inst_offset;              /* offset in inst_table since path1 of func */
  int curloop;                  /* current loop during loop mask calculation */
  vector loopmask;              /* vector of program lines within loop */
  vector *preheaders;           /* list of vectors of all lines at loop entry */
  struct pathinfo *loopheader;  /* header of current loop for path */
  struct pathinfo *prevheader;  /* header of next outer loop for path */
  struct pathlist *exits;       /* exit paths from loop (valid for head only) */
};

struct pathlist {
  struct pathlist *next;        /* next path within the current function */
  struct pathinfo *path;        /* a path within the current function */
  union {
    int allhits;                /* indicates if traversing path will result
                                   in all hits (succs list only) */
    int relevant;               /* indicates if one line associated with
                                   will be relevant during cache simulation */
  } add;
};

struct pathlistlist {
  struct pathlistlist *next;    /* next list of pathlists (same line)    */
  unsigned int tag;            /* tag of first cache line referenced */
  struct pathlist *list;        /* a list of paths (same tag)            */
};

struct funcinfo {
  struct pathlist *paths;       /* list of all paths for the function */
  struct pathlist *entrypaths;  /* list of entry paths for the function */
  struct pathlist *exitpaths;   /* list of exit paths for the function */
  struct callerlist *caller;    /* calling paths, corresponding to vectors */
#ifdef TRASH
  int *offset;                  /* offset in bit vector for each function */
#endif
  char funcname[MAXNAME];       /* name of the function */
  int numpaths;                 /* number of paths for the function */
  int numexitpaths;             /* number of exit paths for the function */
  int funcnum;                  /* index for current function */
  vector as;                    /* bit vector of always successor paths */
  int base_on_stack;            /* set if base register on stack */
};

struct funclist {
  struct funclist *next;        /* ptr to next function */
  struct funcinfo *func;        /* ptr to information about a function */
};

struct cachestateinfo {
  int tagsum;                   /* sum of tags in cache */
  int tagptr;                   /* ptr to first elem in set of tags */
};

struct pathstackinfo {
  struct pathlist *path;        /* current position in a path list */
  struct pathlist *succs;       /* successor of caller if RETURN path */
  int state;                    /* current cache state, negative if RETURN */
  int numchglines;              /* num of tags to be restored */
};

/*
 * External definitions
 */

extern int predict;
extern int profile;
extern int verbose;
extern int warning;
extern char *basename;
extern FILE *fmacros, *ftables, *fprofile;

extern struct funclist *funcs;
extern struct funclist *funcs_tail;
extern struct pathlist *dead_paths;

extern int proglines[MAXPROGLINES];
extern int numpaths;
extern int numfuncs;
extern int highaddr;

extern struct funcinfo *mainfuncptr;

extern unsigned int currtag[MAXNUMSETS];

extern struct pathlistlist *setpath[MAXNUMSETS];

extern FILE *output_file;
extern int numsets;
extern int setsize;
extern int instonline;
extern int linesize;
extern int cachesize;
extern int numtagwords;
extern int cline_shift;
extern int vecbits;
extern vector vallptr;
