/************************************************************************/
/*                                                                      */
/* Copyright (C) 1995 by Frank Mueller.                                 */
/* This code was developed at Florida State University.                 */
/* Distributed by the author(s) under the following terms:              */
/*                                                                      */
/* This code may not be distributed further without permission          */
/*      from Frank Mueller of the Humboldt-University zu Berlin,        */
/*      mueller@informatik.hu-berlin.de or (+49) (30) 20181-276.        */
/* This code is distributed "AS IS" WITHOUT ANY WARRANTY; without even  */
/*      even the implied warranty of MERCHANTABILITY OR FITNESS FOR A   */
/*      PARTICULAR PURPOSE.  No claims are made as to whether it serves */
/*      any particular purpose or even works at all.                    */
/*                                                                      */
/************************************************************************/

/*
 * This file is part of the static cache simulator
 *
 *   $Id: vec.c,v 1.1 2004/10/27 06:32:16 smohan Exp $
 */

/*
 * Bit vector routines
 */

#include <stdio.h>
#include "scs.h"

int bitsperint;                         /* bits in an int - 1 on the host */
int intscale;                           /* log2(bitsperint), scale factor */
int int2char;                           /* scale factor to yield bytes    */
int veclen = 0;                         /* length of a vector in integers */

/*
 * v_init - initialize vector specific variables
 */
void v_init()
{
  int index, rt;
  unsigned int ui;

  ui = bitsperint = 1;
  while (ui <<= 1)
    bitsperint++;
  for (intscale = 0; bitsperint > 1; bitsperint >>= 1)
    intscale++;
  bitsperint = (bitsperint << intscale) - 1;
  index = rt = sizeof(unsigned int);
  for (int2char = 0; index > 1; index >>= 1)
    int2char++;
  if (rt > (((unsigned int) 1) << int2char))
    int2char++;
}

/*
 * vinsert - add an item to a vector
 */
void vinsert(sptr, r)
vector *sptr;
unsigned int r;
{
    register unsigned int *ptr, *end;

    if (!*sptr) {
      *sptr = (vector) (ptr = VALLOC);
      for (end = ptr + veclen; ptr != end; *ptr++ = 0);
      }
    (*sptr)[r >> intscale] |= (((unsigned) 1) << (r & bitsperint));
}

/*
 * vdelete - delete an item to a vector
 */
void vdelete(sptr, r)
vector *sptr;
unsigned int r;
{
    register unsigned int *ptr, *end;

    if (!*sptr) {
      *sptr = (vector) (ptr = VALLOC);
      for (end = ptr + veclen; ptr != end; *ptr++ = 0);
      }
    (*sptr)[r >> intscale] &= ~(((unsigned) 1) << (r & bitsperint));
}

/*
 * vin - indicate if an item line is in a vector
 */
int vin(ptr, r)
vector ptr;
unsigned int r;
{
    return(ptr && ptr[r >> intscale] & (((unsigned) 1) << (r & bitsperint)));
}

/*
 * vminus - calculates {a} - {b}
 */
void vminus(a, b)
vector *a;
vector b;
{
    register unsigned int *ptr1, *ptr2, *end;

    if ((ptr1 = (unsigned int *) *a) && (ptr2 = (unsigned int *) b))
      for (end = ptr1 + veclen; ptr1 != end; *ptr1++ &= ~(*ptr2++));
}

/*
 * vunion - calculates union of two vectors
 */
void vunion(result, vec)
vector *result;
vector vec;
{
    register unsigned int *ptr1, *ptr2;
    register end;

    if (ptr2 = (unsigned int *) vec) {
      if (!*result) {
        *result = (vector) (ptr1 = VALLOC);
        for (end = veclen; end--; *ptr1++ = *ptr2++);
        }
      else {
        ptr1 = (unsigned int *) *result;
        for (end = veclen; end--; *ptr1++ |= *ptr2++);
        }
      }
}

/*
 * vuniond - calculates {a} = {a} + ({b} - {c}), returns TRUE if result changes
 */
int vuniond(a, b, c)
vector *a;
vector b;
vector c;
{
    register unsigned int *ptr1, *ptr2, *ptr3, *end, tmp;
    register unsigned int change = FALSE;

    if (b) {
      if (!*a) {
        *a = (vector) (ptr2 = ptr1 = VALLOC);
        for (end = ptr2 + veclen; ptr2 != end; *ptr2++ = 0);
        }
      else
        end = (ptr1 = (unsigned int *) *a) + veclen;
      ptr2 = (unsigned int *) b;
      if (ptr3 = (unsigned int *) c)
        while (ptr1 != end) {
          tmp = *ptr1;
          if ((*ptr1++ |= (*ptr2++ & ~(*ptr3++))) != tmp)
            change = TRUE;
          }
      else
        while (ptr1 != end) {
          tmp = *ptr1;
          if ((*ptr1++ |= *ptr2++) != tmp)
            change = TRUE;
          }
      }
    return(change);
}

/*
 * vinter - calculate the intersection of two vector lists
 */
void vinter(res, ptr)
vector *res;
vector ptr;
{
  register unsigned int *ptr1, *ptr2, *end;
  
  if (ptr)
    if (!*res) {
      *res = (vector) (ptr1 = VALLOC);
      end = (ptr1 = (unsigned int *) *res) + veclen;
      while (ptr1 != end)
	*ptr1++ = 0;
    }
    else {
      end = (ptr1 = (unsigned int *) *res) + veclen;
      ptr2 = (unsigned int *) ptr;
      while (ptr1 != end)
	*ptr1++ &= *ptr2++;
    }
}

/*
 * vinterd - calculate the intersection of two vector lists and indicate change
 */
int vinterd(res, ptr)
vector *res;
vector ptr;
{
  register unsigned int *ptr1, *ptr2, *end, tmp;
  register unsigned int change = FALSE;

  if (ptr)
    if (!*res) {
      *res = (vector) (ptr1 = VALLOC);
      end = (ptr1 = (unsigned int *) *res) + veclen;
      while (ptr1 != end)
	*ptr1++ = 0;
    }
    else {
      end = (ptr1 = (unsigned int *) *res) + veclen;
      ptr2 = (unsigned int *) ptr;
      while (ptr1 != end) {
	tmp = *ptr1;
	if ((*ptr1++ &= *ptr2++) != tmp)
	  change = TRUE;
      }
    }

  return(change);
}

#ifdef TRASH
/*
 * vinter - calculates {x} = {a} & ({a} = {a} + ({b} U {c}))
 */
void vinter(x, a, b, c)
vector *x;
vector *a;
vector b;
vector c;
{
    register unsigned int *ptr1, *ptr2, *ptr3, *ptrx, *end, tmp;

    if ((ptr2 = (unsigned int *) b) && (ptr3 = (unsigned int *) c)) {
      if (!*a) {
        *a = (vector) (ptr1 = VALLOC);
        for (end = ptr1 + veclen; ptr1 != end; *ptr1++ = *ptr2++ & *ptr3++);
        }
      else {
        if (!(ptrx = *x)) {
          *x = (vector) (ptrx = ptr1 = VALLOC);
          for (end = ptr1 + veclen; ptr1 != end; *ptr1++ = 0);
          }
        end = (ptr1 = (unsigned int *) *a) + veclen;
        while (ptr1 != end) {
          tmp = *ptr2++ & *ptr3++;
          *ptrx++ |= *ptr1 & tmp;
          *ptr1++ |= tmp;
          }
        }
      }
}
#endif

/*
 * vcommon - return vector position of union, if there is only one
 */
int vcommon(a, b)
vector a;
vector b;
{
    register unsigned int *ptr1, *ptr2, *end, tmp;
    register int item, vpos;

    item = vpos = 0;
    if ((ptr1 = (unsigned int *) a) && (ptr2 = (unsigned int *) b)) {
      end = ptr1 + veclen;
      while (ptr1 != end) {
        if (tmp = *ptr1++ & *ptr2++) {
          if (vpos)
            return(0);
          if (tmp & (tmp - 1))
            return(0);
          while (!(tmp & 1)) {
            vpos++;
            tmp >>= 1;
            }
          vpos += item;
          }
        item += (bitsperint + 1);
        }
      }
    return(vpos);
}

/*
 * vget - returns string containing elements indicated by vector
 */
void vget(ptr, s)
vector ptr;
char *s;
{
  register int i, high = -1, last = -1;

  *s = '\0';
  if (verbose) {
    for (i = 0; i < vecbits; i++) {
      if (vin(ptr, i))
	if (i < numsets)
	  sprintf(s, "I ");
	else
	  sprintf(s, "%d ", i - numsets);
      else
	sprintf(s, "  ");
      s += strlen(s);
    }
    return;
  }

  for (i = veclen * (bitsperint + 1) - 1; i >= 0; i--)
    if (vin(ptr, i)) {
      if (high != -1 && last != i+1) {
	if (high == last)
	  sprintf(s, "%d ", high);
	else
	  sprintf(s, "%d-%d ", last, high);
	while (*s)
	  s++;
	high = i;
      }
      if (high == -1)
	high = i;
      last = i;
    }

  if (high != -1)
    if (high == last)
      sprintf(s, "%d ", high);
    else
      sprintf(s, "%d-%d ", last,  high);
}

/*
 * vall - initialize a vector list containing all elements
 */
vector vall()
{
    register unsigned int *ptr, *end;

    end = (ptr = VALLOC) + veclen;
    do {
      *(--end) = -1;
      } while (ptr != end);
    *end = -2; /* no element 0 */
    return((vector) ptr);
}

/*
 * vcpy - copy a vector list to another one
 */
void vcpy(dst, src)
vector *dst;
vector src;
{
    register unsigned int *ptr1, *ptr2, *end;

    if (src) {
      if (!*dst)
        *dst = (vector) VALLOC;
      end = (ptr1 = (unsigned int *) *dst) + veclen;
      ptr2 = (unsigned int *) src;
      while (ptr1 != end)
        *ptr1++ = *ptr2++;
      }
    else if (*dst) {
      vfree(*dst);
      *dst = vinit();
      }
}

/*
 * vcmp - compare two vector lists
 */
int vcmp(a, b)
vector a;
vector b;
{
    register unsigned int *ptr1, *ptr2, *end;

    if (a && b) {
      end = (ptr1 = (unsigned int *) a) + veclen;
      ptr2 = (unsigned int *) b;
      while (ptr1 != end)
        if (*ptr1++ != *ptr2++)
          return(FALSE);
      return(TRUE);
      }
    return(a == b);
}

